#
# System Interface Library for games
# Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
# Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
# See the file COPYING.txt for details.
#
# build/windows/build.mk: Build control file for Windows.
#

#
# Windows-specific notes:
#
# - An additional rule, "test-coverage", is provided to help in checking
#   code coverage.  Running "make test-coverage" will build the executable,
#   run the built-in tests, and output code coverage data to "coverage.dat".
#   See the "Built-in tests" section of the README file for details.
#
# - The $(SIL_COVERAGE_TESTS) variable can be set to a comma-separated list
#   of test names to select which tests are to be run for test-coverage.
#
# -------------------------------------------------------------------------
#
# To use this file, include it in your Makefile after defining values for
# configuration variables and source lists, but before defining any rules.
# For example:
#
# PROGRAM_NAME = MyProgram
# EXECUTABLE_NAME = my-program
#
# MY_SOURCES = my-program.c
# MY_OBJECTS_C = $(MY_SOURCES:$(TOPDIR)/%.c=$(OBJDIR)/%$(OBJECT_EXT))
# OBJECT_GROUPS += MY
# AUTOGEN_HEADERS += my-date.h
#
# include SIL/build/windows/build.mk
#
# .PHONY: my-date.h  # Force regeneration on every build.
# my-date.h:
# 	date +'#define DATE "%Y-%m-%d %H:%M:%S"' >$@
#

###########################################################################
################ Defaults for common configuration options ################
###########################################################################

# We don't have any Windows-specific movie playback code, so enable the
# WebM software decoder by default.
SIL_MOVIE_INCLUDE_WEBM ?= 1

# strtof() is missing in the Visual Studio runtime libraries (at least in
# VS.NET 2003), so always enable our own.
SIL_STRTOF_CUSTOM ?= 1

###########################################################################
################# Windows-specific configuration options ##################
###########################################################################

#----------------------- Program behavior options ------------------------#

# $(SIL_SDL_CENTER_WINDOW):  If set to 1, any windows opened by the program
# will be centered on the screen.  If 0 (the default), the system's default
# behavior will be used.

SIL_SDL_CENTER_WINDOW ?= 0

#---------------------- Build environment pathnames ----------------------#

# $(OPENAL32_DLL):  The path of the openal32.dll library.

OPENAL32_DLL ?= '$(SYSTEMROOT)/system32/openal32.dll'


# $(SDL_BASE):  Base path for the SDL installation.  This variable is used
# to set the default values for $(SDL_CFLAGS), $(SDL_DLL), and $(SDL_LIBS);
# if those variables are explicitly set, this variable is ignored.

SDL_BASE ?= $(SIL_DIR)/external/SDL-2.0.2/windows


# $(SDL_CFLAGS):  C compiler flags for building against the SDL library.
# If not defined, the program will build against the copy of SDL included
# in the "external" directory of the SIL distribution (or as specified by
# the $(SDL_BASE) variable).  To use a copy of SDL installed in standard
# system locations, explicitly set this variable to the empty string on
# the "make" command line or in a Makefile.

ifeq ($(origin SDL_CFLAGS),undefined)
    SDL_CFLAGS = -I'$(SDL_BASE)/include/SDL2'
endif


# $(SDL_DLL):  Path of the SDL library's DLL file.  This file will be
# copied into the executable's directory at build time.  Normally, this
# should be set (or left unset) in conjunction with $(SDL_CFLAGS).

ifeq ($(origin SDL_DLL),undefined)
    SDL_DLL = $(SDL_BASE)/bin/SDL2.dll
endif


# $(SDL_LIBS):  C compiler flags for linking against the SDL library.  If
# not defined, the program will link in the copy of SDL included in the
# "external" directory of the SIL distribution.  Normally, this should be
# set (or left unset) in conjunction with $(SDL_CFLAGS).

ifeq ($(origin SDL_LIBS),undefined)
    SDL_LIBS = -L'$(SDL_BASE)/lib' -lSDL2.dll
endif

#--------------------------- Packaging options ---------------------------#

# $(ICON_FILE):  The path of an ICO-format icon file to use as the main
# application icon for the program.  If not set, the program will not have
# a custom icon.

ICON_FILE ?=


# $(ICON_NAME):  The name of the icon as it will be referenced in the
# program.  Ignored if $(ICON_FILE) is empty.

ICON_NAME ?= icon

###########################################################################
######################### Build environment setup #########################
###########################################################################

_THISDIR := $(abspath $(dir $(lastword $(MAKEFILE_LIST))))

include $(_THISDIR)/../common/base.mk
include $(_THISDIR)/../common/config.mk


ifneq (,$(filter undefined default,$(origin CC)))
CC = mingw32-gcc
endif

OBJECT_EXT  = .o
LIBRARY_EXT = .a

# Our preferred stack boundary is 2^4 == 16 bytes for SSE instructions, but
# 32-bit Windows doesn't honor that, so tell GCC to expect 4-byte alignment.
# NOTE: GCC has a bug (in at least version 4.7.2, possibly earlier versions)
# which can destroy the first argument to a local function in certain cases;
# see http://gcc.gnu.org/bugzilla/show_bug.cgi?id=56807 for details.  If you
# run into this problem for a particular function, either reduce the stack
# space required by the function to less than 4000 bytes or attach
# __attribute__((regparm(0))) to the function.
SYS_CFLAGS    = -m32 -mincoming-stack-boundary=2
SYS_CXXFLAGS  = -m32 -mincoming-stack-boundary=2
SYS_OBJCFLAGS = -m32 -mincoming-stack-boundary=2
SYS_LDFLAGS   = -Wl,--enable-auto-import

BASE_LIBS = $(SDL_LIBS) '$(OPENAL32_DLL)' -lopengl32 \
            $(if $(filter 1,$(DEBUG)),-lpsapi) \
            -ldxguid -lole32 -luuid -lshell32

include $(_THISDIR)/../common/toolchain.mk


# Program paths used only in this file.
WINDRES ?= windres

###########################################################################
################# Source file lists and associated flags ##################
###########################################################################

SIL_SYS_CONFIG_HEADER = windows.h

SIL_SYS_SOURCES = $(SIL_OPENGL_SOURCES) \
                  sysdep/openal/openal.c \
                  sysdep/openal/sound.c \
                  sysdep/sdl/dyngl.c \
                  sysdep/sdl/graphics.c \
                  sysdep/sdl/input.c \
                  sysdep/windows/debug.c \
                  sysdep/windows/files.c \
                  sysdep/windows/log.c \
                  sysdep/windows/main.c \
                  sysdep/windows/misc.c \
                  sysdep/windows/movie.c \
                  sysdep/windows/mutex.c \
                  sysdep/windows/semaphore.c \
                  sysdep/windows/sysfont.c \
                  sysdep/windows/thread.c \
                  sysdep/windows/time.c \
                  sysdep/windows/userdata.c \
                  test/sysdep/sdl/graphics.c \
                  test/sysdep/sdl/input.c \
                  test/sysdep/windows/files.c \
                  test/sysdep/windows/misc.c \
                  test/sysdep/windows/thread.c \
                  test/sysdep/windows/time.c \
                  test/sysdep/windows/userdata.c

SIL_SYS_FLAGS += $(SDL_CFLAGS) \
                 $(call _if-define,SIL_SDL_CENTER_WINDOW)

include $(_THISDIR)/../common/sources.mk

###########################################################################
############################### Build rules ###############################
###########################################################################

SIL_DEFAULT_TARGET = all

SIL_HELP_all           = 'build the executable ($(EXECUTABLE_NAME).exe)'
SIL_HELP_test-coverage = 'build and run tests, writing coverage data to coverage.out'
SIL_HELP_clean         = 'remove all intermediate files and dependency data'
SIL_HELP_spotless      = 'remove all generated files, including the executable'

include $(_THISDIR)/../common/rules.mk

# The resource file is not an "object file" per se, but we include it on
# the link command line with everything else.
ALL_OBJECTS += $(if $(ICON_FILE),$(EXECUTABLE_NAME).res)

#-------------------------------------------------------------------------#

.PHONY: all clean spotless

all: $(EXECUTABLE_NAME).exe SDL2.dll

clean:
	$(ECHO) 'Removing object files'
	$(Q)rm -rf obj
	$(ECHO) 'Removing generated files'
	$(Q)rm -f '$(EXECUTABLE_NAME)'.{ico,rc,res}
	$(ECHO) 'Removing temporary files'
	$(Q)rm -rf .covtmp
	$(SIL_RECIPE_clean)

spotless: clean
	$(ECHO) 'Removing executable file and libraries'
	$(Q)rm -f '$(EXECUTABLE_NAME).exe' SDL2.dll
	$(ECHO) 'Removing coverage data'
	$(Q)rm -f coverage.out
	$(SIL_RECIPE_spotless)

#-------------------------------------------------------------------------#

.PHONY: test-coverage

test-coverage:
	$(Q)$(MAKE) all COVERAGE=1
	$(ECHO) 'Running tests'
	$(Q)find obj -name \*.gcda -delete
	$(Q)'./$(EXECUTABLE_NAME)' -test$(if $(SIL_COVERAGE_TESTS),=$(SIL_COVERAGE_TESTS))
	$(Q)$(MAKE) gen-coverage

#-------------------------------------------------------------------------#

# Sort objects in lexical order to make it easier to estimate build progress.
$(EXECUTABLE_NAME).exe: $(sort $(ALL_OBJECTS))
	$(ECHO) 'Linking $@'
	$(Q)mkdir -p '$(@D)'
	$(Q)$(CXX) \
	    $(BASE_LDFLAGS) $(ALL_LDFLAGS) $(LDFLAGS) \
	    $(ALL_OBJECTS) \
	    $(ALL_LIBS) $(BASE_LIBS) $(LIBS) \
	    -o '$@'

SDL2.dll: $(SDL_DLL)
	$(ECHO) 'Copying $@ into executable directory'
	$(Q)mkdir -p '$(@D)'
	$(Q)cp -pf '$<' '$@'

#-------------------------------------------------------------------------#

# We can't use the source icon filename directly in the .rc file because
# windres expects absolute pathnames to start with a drive letter, and
# $(ICON_FILE) may be a POSIX-style path starting with a "/".

$(EXECUTABLE_NAME).res: $(EXECUTABLE_NAME).rc $(EXECUTABLE_NAME).ico $(MAKEFILE_DEPS)
	$(ECHO) 'Generating $@'
	$(Q)$(WINDRES) --use-temp-file -O COFF -i '$<' -o '$@'

$(EXECUTABLE_NAME).rc: $(MAKEFILE_DEPS)
	$(ECHO) 'Generating $@'
	$(Q)echo >'$@' '$(ICON_NAME) ICON $(EXECUTABLE_NAME).ico'

ifneq ($(ICON_FILE),)
$(EXECUTABLE_NAME).ico: $(ICON_FILE) $(MAKEFILE_DEPS)
	$(ECHO) 'Copying $@ from $<'
	$(Q)cp -p '$<' '$@'
endif

###########################################################################
###########################################################################
