/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * include/SIL/math/vector.h: Vector type and function definitions for
 * generic systems.
 */

#ifndef SIL_MATH_VECTOR_H
#define SIL_MATH_VECTOR_H

/*
 * This file defines the Vector2f, Vector3f, and Vector4f types and
 * associated functions (all inline).  The code is written so as to
 * function properly and reasonably quickly in any environment, but for
 * that reason does not take advantage of any architecture-specific
 * features.  To take advantage of architectures or compilers with support
 * for faster vector operations, make a copy of this file with appropriate
 * changes for the architecture or compiler, then define the preprocessor
 * symbol SIL_SYSTEM_VECTOR_HEADER in the system's configuration header
 * to point to the new file.
 */

/*************************************************************************/
/*************************************************************************/

/* Forward-declare the data types. */

#ifdef __cplusplus
struct Vector2f;
struct Vector3f;
struct Vector4f;
#else
typedef struct Vector2f Vector2f;
typedef struct Vector3f Vector3f;
typedef struct Vector4f Vector4f;
#endif

/*-----------------------------------------------------------------------*/

/* Declare the structures themselves.  The C++ methods will need to be
 * defined later because we need all three types defined first. */

#define DECLARE_VECTOR_METHODS(type)                    \
    inline type(const Vector2f &v);                     \
    inline type(const Vector3f &v);                     \
    inline type(const Vector4f &v);                     \
                                                        \
    inline type &operator=(const float k);              \
    inline type &operator=(const Vector2f &v);          \
    inline type &operator=(const Vector3f &v);          \
    inline type &operator=(const Vector4f &v);          \
                                                        \
    inline type operator+() const;                      \
    inline type operator-() const;                      \
                                                        \
    inline type operator+(const float k) const;         \
    inline type operator+(const Vector2f &v) const;     \
    inline type operator+(const Vector3f &v) const;     \
    inline type operator+(const Vector4f &v) const;     \
                                                        \
    inline type operator-(const float k) const;         \
    inline type operator-(const Vector2f &v) const;     \
    inline type operator-(const Vector3f &v) const;     \
    inline type operator-(const Vector4f &v) const;     \
                                                        \
    inline type operator*(const float k) const;         \
    inline type operator*(const Vector2f &v) const;     \
    inline type operator*(const Vector3f &v) const;     \
    inline type operator*(const Vector4f &v) const;     \
                                                        \
    inline type operator/(const float k) const;         \
    inline type operator/(const Vector2f &v) const;     \
    inline type operator/(const Vector3f &v) const;     \
    inline type operator/(const Vector4f &v) const;     \
                                                        \
    friend inline type operator+(const float k, const type &v); \
    friend inline type operator*(const float k, const type &v); \
                                                        \
    inline type &operator+=(const float k);             \
    inline type &operator+=(const Vector2f &v);         \
    inline type &operator+=(const Vector3f &v);         \
    inline type &operator+=(const Vector4f &v);         \
                                                        \
    inline type &operator-=(const float k);             \
    inline type &operator-=(const Vector2f &v);         \
    inline type &operator-=(const Vector3f &v);         \
    inline type &operator-=(const Vector4f &v);         \
                                                        \
    inline type &operator*=(const float k);             \
    inline type &operator*=(const Vector2f &v);         \
    inline type &operator*=(const Vector3f &v);         \
    inline type &operator*=(const Vector4f &v);         \
                                                        \
    inline type &operator/=(const float k);             \
    inline type &operator/=(const Vector2f &v);         \
    inline type &operator/=(const Vector3f &v);         \
    inline type &operator/=(const Vector4f &v);         \
                                                        \
    inline bool operator==(const type &v) const;        \
    inline bool operator!=(const type &v) const;        \
                                                        \
    inline float length() const;                        \
    inline float length2() const;                       \
    inline bool is_nonzero() const;                     \
    inline bool is_length_in(const float k) const;      \
    inline void normalize();                            \
    inline void set_length(const float k);              \
    inline void cap_length(const float k);              \
    inline float dot(const type &v) const;

/*----------------------------------*/

struct Vector2f {
    float x, y;

#ifdef __cplusplus
    inline Vector2f(float _x = 0, float _y = 0)
        : x(_x), y(_y) {}
    DECLARE_VECTOR_METHODS(Vector2f)
#endif
};

struct Vector3f {
    float x, y, z;

#ifdef __cplusplus
    inline Vector3f(float _x = 0, float _y = 0, float _z = 0)
        : x(_x), y(_y), z(_z) {}
    DECLARE_VECTOR_METHODS(Vector3f)
    inline Vector3f cross(const Vector3f &v) const;
#endif
};

struct Vector4f {
    float x, y, z, w;

#ifdef __cplusplus
    inline Vector4f(float _x = 0, float _y = 0, float _z = 0, float _w = 0)
        : x(_x), y(_y), z(_z), w(_w) {}
    DECLARE_VECTOR_METHODS(Vector4f)
    inline Vector4f cross(const Vector4f &v2, const Vector4f &v3) const;
#endif  // __cplusplus
};

/*----------------------------------*/

#undef DECLARE_VECTOR_METHODS

/*************************************************************************/
/*************************************************************************/

/* Define the C operator functions. */

EXTERN_C_BEGIN

/*-----------------------------------------------------------------------*/

/**
 * vec[234]_add, vec[234]_sub, vec[234]_mul, vec[234]_div:  Add, subtract,
 * multiply, or divide two vectors, component by component.
 *
 * [Parameters]
 *     dest: Pointer to vector in which to store result.
 *     src1, src2: Pointers to vectors to add, subtract, multiply, or divide.
 */
static inline void vec2_add(Vector2f *dest,
                            const Vector2f *src1, const Vector2f *src2)
{
    dest->x = src1->x + src2->x;
    dest->y = src1->y + src2->y;
}

static inline void vec3_add(Vector3f *dest,
                            const Vector3f *src1, const Vector3f *src2)
{
    dest->x = src1->x + src2->x;
    dest->y = src1->y + src2->y;
    dest->z = src1->z + src2->z;
}

static inline void vec4_add(Vector4f *dest,
                            const Vector4f *src1, const Vector4f *src2)
{
    dest->x = src1->x + src2->x;
    dest->y = src1->y + src2->y;
    dest->z = src1->z + src2->z;
    dest->w = src1->w + src2->w;
}

/*----------------------------------*/

static inline void vec2_sub(Vector2f *dest,
                            const Vector2f *src1, const Vector2f *src2)
{
    dest->x = src1->x - src2->x;
    dest->y = src1->y - src2->y;
}

static inline void vec3_sub(Vector3f *dest,
                            const Vector3f *src1, const Vector3f *src2)
{
    dest->x = src1->x - src2->x;
    dest->y = src1->y - src2->y;
    dest->z = src1->z - src2->z;
}

static inline void vec4_sub(Vector4f *dest,
                            const Vector4f *src1, const Vector4f *src2)
{
    dest->x = src1->x - src2->x;
    dest->y = src1->y - src2->y;
    dest->z = src1->z - src2->z;
    dest->w = src1->w - src2->w;
}

/*----------------------------------*/

static inline void vec2_mul(Vector2f *dest,
                            const Vector2f *src1, const Vector2f *src2)
{
    dest->x = src1->x * src2->x;
    dest->y = src1->y * src2->y;
}

static inline void vec3_mul(Vector3f *dest,
                            const Vector3f *src1, const Vector3f *src2)
{
    dest->x = src1->x * src2->x;
    dest->y = src1->y * src2->y;
    dest->z = src1->z * src2->z;
}

static inline void vec4_mul(Vector4f *dest,
                            const Vector4f *src1, const Vector4f *src2)
{
    dest->x = src1->x * src2->x;
    dest->y = src1->y * src2->y;
    dest->z = src1->z * src2->z;
    dest->w = src1->w * src2->w;
}

/*----------------------------------*/

static inline void vec2_div(Vector2f *dest,
                            const Vector2f *src1, const Vector2f *src2)
{
    dest->x = src1->x / src2->x;
    dest->y = src1->y / src2->y;
}

static inline void vec3_div(Vector3f *dest,
                            const Vector3f *src1, const Vector3f *src2)
{
    dest->x = src1->x / src2->x;
    dest->y = src1->y / src2->y;
    dest->z = src1->z / src2->z;
}

static inline void vec4_div(Vector4f *dest,
                            const Vector4f *src1, const Vector4f *src2)
{
    dest->x = src1->x / src2->x;
    dest->y = src1->y / src2->y;
    dest->z = src1->z / src2->z;
    dest->w = src1->w / src2->w;
}

/*-----------------------------------------------------------------------*/

/**
 * vec[234]_add_scalar, vec[234]_scale:  Add or multiply a scalar value to
 * each component of a vector.
 *
 * [Parameters]
 *     dest: Pointer to vector in which to store result.
 *     src: Pointer to vector to operate on.
 *     k: Scalar value to add or multiply.
 */
static inline void vec2_add_scalar(Vector2f *dest,
                                   const Vector2f *src, const float k)
{
    dest->x = src->x + k;
    dest->y = src->y + k;
}
static inline void vec3_add_scalar(Vector3f *dest,
                                   const Vector3f *src, const float k)
{
    dest->x = src->x + k;
    dest->y = src->y + k;
    dest->z = src->z + k;
}
static inline void vec4_add_scalar(Vector4f *dest,
                                   const Vector4f *src, const float k)
{
    dest->x = src->x + k;
    dest->y = src->y + k;
    dest->z = src->z + k;
    dest->w = src->w + k;
}

/*----------------------------------*/

static inline void vec2_scale(Vector2f *dest,
                              const Vector2f *src, const float k)
{
    dest->x = src->x * k;
    dest->y = src->y * k;
}
static inline void vec3_scale(Vector3f *dest,
                              const Vector3f *src, const float k)
{
    dest->x = src->x * k;
    dest->y = src->y * k;
    dest->z = src->z * k;
}
static inline void vec4_scale(Vector4f *dest,
                              const Vector4f *src, const float k)
{
    dest->x = src->x * k;
    dest->y = src->y * k;
    dest->z = src->z * k;
    dest->w = src->w * k;
}

/*-----------------------------------------------------------------------*/

/**
 * vec[234]_dot:  Calculate the dot product of two vectors.
 *
 * [Parameters]
 *     a, b: Pointers to vectors to compute the dot product of.
 * [Return value]
 *     dot(a, b)
 */
static PURE_FUNCTION inline float vec2_dot(const Vector2f *a, const Vector2f *b)
{
    return a->x*b->x + a->y*b->y;
}

static PURE_FUNCTION inline float vec3_dot(const Vector3f *a, const Vector3f *b)
{
    return a->x*b->x + a->y*b->y + a->z*b->z;
}

static PURE_FUNCTION inline float vec4_dot(const Vector4f *a, const Vector4f *b)
{
    return a->x*b->x + a->y*b->y + a->z*b->z + a->w*b->w;
}

/*-----------------------------------------------------------------------*/

/**
 * vec[234]_length2:  Return the squared length of the given vector.
 *
 * [Parameters]
 *     v: Pointer to vector to compute the squared length of.
 * [Return value]
 *     ||v||^2
 */
static PURE_FUNCTION inline float vec2_length2(const Vector2f *v)
{
    return vec2_dot(v, v);
}

static PURE_FUNCTION inline float vec3_length2(const Vector3f *v)
{
    return vec3_dot(v, v);
}

static PURE_FUNCTION inline float vec4_length2(const Vector4f *v)
{
    return vec4_dot(v, v);
}

/*----------------------------------*/

/**
 * vec[234]_length:  Return the length of the given vector.
 *
 * [Parameters]
 *     v: Pointer to vector to compute the length of.
 * [Return value]
 *     ||v||
 */
static PURE_FUNCTION inline float vec2_length(const Vector2f *v)
{
    return sqrtf(vec2_length2(v));
}

static PURE_FUNCTION inline float vec3_length(const Vector3f *v)
{
    return sqrtf(vec3_length2(v));
}

static PURE_FUNCTION inline float vec4_length(const Vector4f *v)
{
    return sqrtf(vec4_length2(v));
}

/*----------------------------------*/

/**
 * vec[234]_is_nonzero:  Return whether the given vector is nonzero, i.e.
 * has at least one nonzero component.
 *
 * IMPORTANT: Even if this function returns false (not zero), the
 * vec*_length() and vec*_length2() functions may still return zero due to
 * floating-point underflow.  To catch division by zero when dividing by
 * the length of a vector, use "vec*_length(v) != 0" (or, equivalently,
 * "vec*_length2(v) != 0") rather than calling vec*_is_nonzero().
 *
 * [Parameters]
 *     v: Pointer to vector to check for zeroness.
 * [Return value]
 *     True if the vector is nonzero, false otherwise.
 */
static PURE_FUNCTION inline int vec2_is_nonzero(const Vector2f *v)
{
    return v->x != 0 || v->y != 0;
}

static PURE_FUNCTION inline int vec3_is_nonzero(const Vector3f *v)
{
    return v->x != 0 || v->y != 0 || v->z != 0;
}

static PURE_FUNCTION inline int vec4_is_nonzero(const Vector4f *v)
{
    return v->x != 0 || v->y != 0 || v->z != 0 || v->w != 0;
}

/*----------------------------------*/

/**
 * vec[234]_is_length_in:  Return whether the given vector's length is
 * within (less than or equal to) the given limit.
 *
 * [Parameters]
 *     v: Pointer to vector to check.
 *     k: Length limit to check against.
 * [Return value]
 *     True if the vector's length is within the limit, false otherwise.
 */
static PURE_FUNCTION inline int vec2_is_length_in(const Vector2f *v, float k)
{
    return vec2_length2(v) <= k*k;
}

static PURE_FUNCTION inline int vec3_is_length_in(const Vector3f *v, float k)
{
    return vec3_length2(v) <= k*k;
}

static PURE_FUNCTION inline int vec4_is_length_in(const Vector4f *v, float k)
{
    return vec4_length2(v) <= k*k;
}

/*----------------------------------*/

/**
 * vec[234]_normalize:  Normalize (adjust to unit length) the given vector.
 *
 * [Parameters]
 *     dest: Pointer to vector in which to store result.
 *     src: Pointer to vector to normalize.
 */

static inline void vec2_normalize(Vector2f *dest, const Vector2f *src)
{
    const float length2 = vec2_length2(src);
    if (LIKELY(length2 > 0)) {
        vec2_scale(dest, src, 1/sqrtf(length2));
    } else {
        dest->x = dest->y = 0;
    }
}

static inline void vec3_normalize(Vector3f *dest, const Vector3f *src)
{
    const float length2 = vec3_length2(src);
    if (LIKELY(length2 > 0)) {
        vec3_scale(dest, src, 1/sqrtf(length2));
    } else {
        dest->x = dest->y = dest->z = 0;
    }
}

static inline void vec4_normalize(Vector4f *dest, const Vector4f *src)
{
    const float length2 = vec4_length2(src);
    if (LIKELY(length2 > 0)) {
        vec4_scale(dest, src, 1/sqrtf(length2));
    } else {
        dest->x = dest->y = dest->z = dest->w = 0;
    }
}

/*----------------------------------*/

/**
 * vec[234]_set_length:  Scale the given vector to have the given length.
 *
 * [Parameters]
 *     dest: Pointer to vector in which to store result.
 *     src: Pointer to vector to scale.
 *     k: New length for vector.
 */

static inline void vec2_set_length(Vector2f *dest,
                                   const Vector2f *src, const float k)
{
    const float length2 = vec2_length2(src);
    if (LIKELY(length2 > 0)) {
        vec2_scale(dest, src, k/sqrtf(length2));
    } else {
        dest->x = dest->y = 0;
    }
}

static inline void vec3_set_length(Vector3f *dest,
                                   const Vector3f *src, const float k)
{
    const float length2 = vec3_length2(src);
    if (LIKELY(length2 > 0)) {
        vec3_scale(dest, src, k/sqrtf(length2));
    } else {
        dest->x = dest->y = dest->z = 0;
    }
}

static inline void vec4_set_length(Vector4f *dest,
                                   const Vector4f *src, const float k)
{
    const float length2 = vec4_length2(src);
    if (LIKELY(length2 > 0)) {
        vec4_scale(dest, src, k/sqrtf(length2));
    } else {
        dest->x = dest->y = dest->z = dest->w = 0;
    }
}

/*-----------------------------------------------------------------------*/

/**
 * vec[234]_cap_length:  Scale the given vector if necessary to have a
 * length no greater than the given length.
 *
 * [Parameters]
 *     dest: Pointer to vector in which to store result.
 *     src: Pointer to vector to scale.
 *     k: Maximum length for vector.
 */

static inline void vec2_cap_length(Vector2f *dest,
                                   const Vector2f *src, const float k)
{
    const float length2 = vec2_length2(src);
    if (length2 > k*k) {
        vec2_scale(dest, src, k/sqrtf(length2));
    } else {
        *dest = *src;
    }
}

static inline void vec3_cap_length(Vector3f *dest,
                                   const Vector3f *src, const float k)
{
    const float length2 = vec3_length2(src);
    if (length2 > k*k) {
        vec3_scale(dest, src, k/sqrtf(length2));
    } else {
        *dest = *src;
    }
}

static inline void vec4_cap_length(Vector4f *dest,
                                   const Vector4f *src, const float k)
{
    const float length2 = vec4_length2(src);
    if (length2 > k*k) {
        vec4_scale(dest, src, k/sqrtf(length2));
    } else {
        *dest = *src;
    }
}

/*-----------------------------------------------------------------------*/

/**
 * vec3_cross:  Calculate the cross product of two 3-component vectors.
 *
 * [Parameters]
 *     dest: Pointer to vector in which to store result.
 *     src1, src2: Pointers to vectors to take the cross product of.
 */
static inline void vec3_cross(Vector3f *dest,
                              const Vector3f *src1, const Vector3f *src2)
{
    /* We use temporary values here so that if dest is the same as src1
     * or src2, we don't overwrite the input values before we're finished
     * using them. */
    const float x = src1->y * src2->z - src1->z * src2->y;
    const float y = src1->z * src2->x - src1->x * src2->z;
    const float z = src1->x * src2->y - src1->y * src2->x;
    dest->x = x;
    dest->y = y;
    dest->z = z;
}

/*----------------------------------*/

/**
 * vec4_cross:  Calculate the cross product of three 4-component vectors.
 *
 * [Parameters]
 *     dest: Pointer to vector in which to store result.
 *     src1, src2, src3: Pointers to vectors to take the cross product of.
 */
static inline void vec4_cross(Vector4f *dest, const Vector4f *src1,
                              const Vector4f *src2, const Vector4f *src3)
{
    const float x =   src1->y * (src2->z * src3->w - src2->w * src3->z)
                    + src1->z * (src2->w * src3->y - src2->y * src3->w)
                    + src1->w * (src2->y * src3->z - src2->z * src3->y);
    const float y = -(src1->x * (src2->z * src3->w - src2->w * src3->z)
                    + src1->z * (src2->w * src3->x - src2->x * src3->w)
                    + src1->w * (src2->x * src3->z - src2->z * src3->x));
    const float z =   src1->x * (src2->y * src3->w - src2->w * src3->y)
                    + src1->y * (src2->w * src3->x - src2->x * src3->w)
                    + src1->w * (src2->x * src3->y - src2->y * src3->x);
    const float w = -(src1->x * (src2->y * src3->z - src2->z * src3->y)
                    + src1->y * (src2->z * src3->x - src2->x * src3->z)
                    + src1->z * (src2->x * src3->y - src2->y * src3->x));
    dest->x = x;
    dest->y = y;
    dest->z = z;
    dest->w = w;
}

/*-----------------------------------------------------------------------*/

EXTERN_C_END

/*************************************************************************/
/*************************************************************************/

/* Define the C++ operator functions. */

#ifdef __cplusplus

/*-----------------------------------------------------------------------*/

Vector2f::Vector2f(const Vector2f &v) : x(v.x), y(v.y) {}
Vector2f::Vector2f(const Vector3f &v) : x(v.x), y(v.y) {}
Vector2f::Vector2f(const Vector4f &v) : x(v.x), y(v.y) {}

Vector2f &Vector2f::operator=(const float k)
    {x = k;   y = 0;   return *this;}
Vector2f &Vector2f::operator=(const Vector2f &v)
    {x = v.x; y = v.y; return *this;}
Vector2f &Vector2f::operator=(const Vector3f &v)
    {x = v.x; y = v.y; return *this;}
Vector2f &Vector2f::operator=(const Vector4f &v)
    {x = v.x; y = v.y; return *this;}

Vector2f Vector2f::operator+() const
    {return *this;}
Vector2f Vector2f::operator-() const
    {return Vector2f(-x, -y);}

Vector2f Vector2f::operator+(const float k) const
    {return Vector2f(x+k, y+k);}
Vector2f Vector2f::operator+(const Vector2f &v) const
    {return Vector2f(x+v.x, y+v.y);}
Vector2f Vector2f::operator+(const Vector3f &v) const
    {return Vector2f(x+v.x, y+v.y);}
Vector2f Vector2f::operator+(const Vector4f &v) const
    {return Vector2f(x+v.x, y+v.y);}

Vector2f Vector2f::operator-(const float k) const
    {return Vector2f(x-k, y-k);}
Vector2f Vector2f::operator-(const Vector2f &v) const
    {return Vector2f(x-v.x, y-v.y);}
Vector2f Vector2f::operator-(const Vector3f &v) const
    {return Vector2f(x-v.x, y-v.y);}
Vector2f Vector2f::operator-(const Vector4f &v) const
    {return Vector2f(x-v.x, y-v.y);}

Vector2f Vector2f::operator*(const float k) const
    {return Vector2f(x*k, y*k);}
Vector2f Vector2f::operator*(const Vector2f &v) const
    {return Vector2f(x*v.x, y*v.y);}
Vector2f Vector2f::operator*(const Vector3f &v) const
    {return Vector2f(x*v.x, y*v.y);}
Vector2f Vector2f::operator*(const Vector4f &v) const
    {return Vector2f(x*v.x, y*v.y);}

Vector2f Vector2f::operator/(const float k) const
    {return Vector2f(x/k, y/k);}
Vector2f Vector2f::operator/(const Vector2f &v) const
    {return Vector2f(x/v.x, y/v.y);}
Vector2f Vector2f::operator/(const Vector3f &v) const
    {return Vector2f(x/v.x, y/v.y);}
Vector2f Vector2f::operator/(const Vector4f &v) const
    {return Vector2f(x/v.x, y/v.y);}

Vector2f operator+(const float k, const Vector2f &v) {return v+k;}
Vector2f operator*(const float k, const Vector2f &v) {return v*k;}

Vector2f &Vector2f::operator+=(const float k)
    {x += k;   y += k;   return *this;}
Vector2f &Vector2f::operator+=(const Vector2f &v)
    {x += v.x; y += v.y; return *this;}
Vector2f &Vector2f::operator+=(const Vector3f &v)
    {x += v.x; y += v.y; return *this;}
Vector2f &Vector2f::operator+=(const Vector4f &v)
    {x += v.x; y += v.y; return *this;}

Vector2f &Vector2f::operator-=(const float k)
    {x -= k;   y -= k;   return *this;}
Vector2f &Vector2f::operator-=(const Vector2f &v)
    {x -= v.x; y -= v.y; return *this;}
Vector2f &Vector2f::operator-=(const Vector3f &v)
    {x -= v.x; y -= v.y; return *this;}
Vector2f &Vector2f::operator-=(const Vector4f &v)
    {x -= v.x; y -= v.y; return *this;}

Vector2f &Vector2f::operator*=(const float k)
    {x *= k;   y *= k;   return *this;}
Vector2f &Vector2f::operator*=(const Vector2f &v)
    {x *= v.x; y *= v.y; return *this;}
Vector2f &Vector2f::operator*=(const Vector3f &v)
    {x *= v.x; y *= v.y; return *this;}
Vector2f &Vector2f::operator*=(const Vector4f &v)
    {x *= v.x; y *= v.y; return *this;}

Vector2f &Vector2f::operator/=(const float k)
    {x /= k;   y /= k;   return *this;}
Vector2f &Vector2f::operator/=(const Vector2f &v)
    {x /= v.x; y /= v.y; return *this;}
Vector2f &Vector2f::operator/=(const Vector3f &v)
    {x /= v.x; y /= v.y; return *this;}
Vector2f &Vector2f::operator/=(const Vector4f &v)
    {x /= v.x; y /= v.y; return *this;}

bool Vector2f::operator==(const Vector2f &v) const
    {return x == v.x && y == v.y;}
bool Vector2f::operator!=(const Vector2f &v) const
    {return x != v.x || y != v.y;}

float Vector2f::length() const           {return sqrtf(length2());}
float Vector2f::length2() const          {return x*x + y*y;}
bool Vector2f::is_nonzero() const        {return x!=0 || y!=0;}
bool Vector2f::is_length_in(const float k) const {return length2() <= k*k;}
void Vector2f::normalize() {
    if (LIKELY(length2() != 0)) {
        *this /= length();
    } else {
        x = y = 0;
    }
}
void Vector2f::set_length(const float k) {
    if (LIKELY(length2() != 0)) {
        *this *= k/length();
    } else {
        x = y = 0;
    }
}
void Vector2f::cap_length(const float k)
    {if (length2() > k*k) *this *= k/length();}
float Vector2f::dot(const Vector2f &v) const
    {return x*v.x + y*v.y;}

/*-----------------------------------------------------------------------*/

Vector3f::Vector3f(const Vector2f &v) : x(v.x), y(v.y), z(0) {}
Vector3f::Vector3f(const Vector3f &v) : x(v.x), y(v.y), z(v.z) {}
Vector3f::Vector3f(const Vector4f &v) : x(v.x), y(v.y), z(v.z) {}

Vector3f &Vector3f::operator=(const float k)
    {x = k;   y = 0;   z = 0;   return *this;}
Vector3f &Vector3f::operator=(const Vector2f &v)
    {x = v.x; y = v.y; z = 0;   return *this;}
Vector3f &Vector3f::operator=(const Vector3f &v)
    {x = v.x; y = v.y; z = v.z; return *this;}
Vector3f &Vector3f::operator=(const Vector4f &v)
    {x = v.x; y = v.y; z = v.z; return *this;}

Vector3f Vector3f::operator+() const
    {return *this;}
Vector3f Vector3f::operator-() const
    {return Vector3f(-x, -y, -z);}

Vector3f Vector3f::operator+(const float k) const
    {return Vector3f(x+k, y+k, z+k);}
Vector3f Vector3f::operator+(const Vector2f &v) const
    {return Vector3f(x+v.x, y+v.y, z);}
Vector3f Vector3f::operator+(const Vector3f &v) const
    {return Vector3f(x+v.x, y+v.y, z+v.z);}
Vector3f Vector3f::operator+(const Vector4f &v) const
    {return Vector3f(x+v.x, y+v.y, z+v.z);}

Vector3f Vector3f::operator-(const float k) const
    {return Vector3f(x-k, y-k, z-k);}
Vector3f Vector3f::operator-(const Vector2f &v) const
    {return Vector3f(x-v.x, y-v.y, z);}
Vector3f Vector3f::operator-(const Vector3f &v) const
    {return Vector3f(x-v.x, y-v.y, z-v.z);}
Vector3f Vector3f::operator-(const Vector4f &v) const
    {return Vector3f(x-v.x, y-v.y, z-v.z);}

Vector3f Vector3f::operator*(const float k) const
    {return Vector3f(x*k, y*k, z*k);}
Vector3f Vector3f::operator*(const Vector2f &v) const
    {return Vector3f(x*v.x, y*v.y, z);}
Vector3f Vector3f::operator*(const Vector3f &v) const
    {return Vector3f(x*v.x, y*v.y, z*v.z);}
Vector3f Vector3f::operator*(const Vector4f &v) const
    {return Vector3f(x*v.x, y*v.y, z*v.z);}

Vector3f Vector3f::operator/(const float k) const
    {return Vector3f(x/k, y/k, z/k);}
Vector3f Vector3f::operator/(const Vector2f &v) const
    {return Vector3f(x/v.x, y/v.y, z);}
Vector3f Vector3f::operator/(const Vector3f &v) const
    {return Vector3f(x/v.x, y/v.y, z/v.z);}
Vector3f Vector3f::operator/(const Vector4f &v) const
    {return Vector3f(x/v.x, y/v.y, z/v.z);}

Vector3f operator+(const float k, const Vector3f &v) {return v+k;}
Vector3f operator*(const float k, const Vector3f &v) {return v*k;}

Vector3f &Vector3f::operator+=(const float k)
    {x += k;   y += k;   z += k;   return *this;}
Vector3f &Vector3f::operator+=(const Vector2f &v)
    {x += v.x; y += v.y;           return *this;}
Vector3f &Vector3f::operator+=(const Vector3f &v)
    {x += v.x; y += v.y; z += v.z; return *this;}
Vector3f &Vector3f::operator+=(const Vector4f &v)
    {x += v.x; y += v.y; z += v.z; return *this;}

Vector3f &Vector3f::operator-=(const float k)
    {x -= k;   y -= k;   z -= k;   return *this;}
Vector3f &Vector3f::operator-=(const Vector2f &v)
    {x -= v.x; y -= v.y;           return *this;}
Vector3f &Vector3f::operator-=(const Vector3f &v)
    {x -= v.x; y -= v.y; z -= v.z; return *this;}
Vector3f &Vector3f::operator-=(const Vector4f &v)
    {x -= v.x; y -= v.y; z -= v.z; return *this;}

Vector3f &Vector3f::operator*=(const float k)
    {x *= k;   y *= k;   z *= k;   return *this;}
Vector3f &Vector3f::operator*=(const Vector2f &v)
    {x *= v.x; y *= v.y;           return *this;}
Vector3f &Vector3f::operator*=(const Vector3f &v)
    {x *= v.x; y *= v.y; z *= v.z; return *this;}
Vector3f &Vector3f::operator*=(const Vector4f &v)
    {x *= v.x; y *= v.y; z *= v.z; return *this;}

Vector3f &Vector3f::operator/=(const float k)
    {x /= k;   y /= k;   z /= k;   return *this;}
Vector3f &Vector3f::operator/=(const Vector2f &v)
    {x /= v.x; y /= v.y;           return *this;}
Vector3f &Vector3f::operator/=(const Vector3f &v)
    {x /= v.x; y /= v.y; z /= v.z; return *this;}
Vector3f &Vector3f::operator/=(const Vector4f &v)
    {x /= v.x; y /= v.y; z /= v.z; return *this;}

bool Vector3f::operator==(const Vector3f &v) const
    {return x == v.x && y == v.y && z == v.z;}
bool Vector3f::operator!=(const Vector3f &v) const
    {return x != v.x || y != v.y || z != v.z;}

float Vector3f::length() const           {return sqrtf(length2());}
float Vector3f::length2() const          {return x*x + y*y + z*z;}
bool Vector3f::is_nonzero() const        {return x!=0 || y!=0 || z!=0;}
bool Vector3f::is_length_in(const float k) const {return length2() <= k*k;}
void Vector3f::normalize() {
    if (LIKELY(length2() != 0)) {
        *this /= length();
    } else {
        x = y = z = 0;
    }
}
void Vector3f::set_length(const float k) {
    if (LIKELY(length2() != 0)) {
        *this *= k/length();
    } else {
        x = y = z = 0;
    }
}
void Vector3f::cap_length(const float k)
    {if (length2() > k*k) *this *= k/length();}
float Vector3f::dot(const Vector3f &v) const
    {return x*v.x + y*v.y + z*v.z;}
Vector3f Vector3f::cross(const Vector3f &v) const
    {return Vector3f(y*v.z - z*v.y,
                     z*v.x - x*v.z,
                     x*v.y - y*v.x);}

/*-----------------------------------------------------------------------*/

Vector4f::Vector4f(const Vector2f &v) : x(v.x), y(v.y), z(0), w(0) {}
Vector4f::Vector4f(const Vector3f &v) : x(v.x), y(v.y), z(v.z), w(0) {}
Vector4f::Vector4f(const Vector4f &v) : x(v.x), y(v.y), z(v.z), w(v.w) {}

Vector4f &Vector4f::operator=(const float k)
    {x = k;   y = 0;   z = 0;   w = 0;   return *this;}
Vector4f &Vector4f::operator=(const Vector2f &v)
    {x = v.x; y = v.y; z = 0;   w = 0;   return *this;}
Vector4f &Vector4f::operator=(const Vector3f &v)
    {x = v.x; y = v.y; z = v.z; w = 0;   return *this;}
Vector4f &Vector4f::operator=(const Vector4f &v)
    {x = v.x; y = v.y; z = v.z; w = v.w; return *this;}

Vector4f Vector4f::operator+() const
    {return *this;}
Vector4f Vector4f::operator-() const
    {return Vector4f(-x, -y, -z, -w);}

Vector4f Vector4f::operator+(const float k) const
    {return Vector4f(x+k, y+k, z+k, w+k);}
Vector4f Vector4f::operator+(const Vector2f &v) const
    {return Vector4f(x+v.x, y+v.y, z, w);}
Vector4f Vector4f::operator+(const Vector3f &v) const
    {return Vector4f(x+v.x, y+v.y, z+v.z, w);}
Vector4f Vector4f::operator+(const Vector4f &v) const
    {return Vector4f(x+v.x, y+v.y, z+v.z, w+v.w);}

Vector4f Vector4f::operator-(const float k) const
    {return Vector4f(x-k, y-k, z-k, w-k);}
Vector4f Vector4f::operator-(const Vector2f &v) const
    {return Vector4f(x-v.x, y-v.y, z, w);}
Vector4f Vector4f::operator-(const Vector3f &v) const
    {return Vector4f(x-v.x, y-v.y, z-v.z, w);}
Vector4f Vector4f::operator-(const Vector4f &v) const
    {return Vector4f(x-v.x, y-v.y, z-v.z, w-v.w);}

Vector4f Vector4f::operator*(const float k) const
    {return Vector4f(x*k, y*k, z*k, w*k);}
Vector4f Vector4f::operator*(const Vector2f &v) const
    {return Vector4f(x*v.x, y*v.y, z, w);}
Vector4f Vector4f::operator*(const Vector3f &v) const
    {return Vector4f(x*v.x, y*v.y, z*v.z, w);}
Vector4f Vector4f::operator*(const Vector4f &v) const
    {return Vector4f(x*v.x, y*v.y, z*v.z, w*v.w);}

Vector4f Vector4f::operator/(const float k) const
    {return Vector4f(x/k, y/k, z/k, w/k);}
Vector4f Vector4f::operator/(const Vector2f &v) const
    {return Vector4f(x/v.x, y/v.y, z, w);}
Vector4f Vector4f::operator/(const Vector3f &v) const
    {return Vector4f(x/v.x, y/v.y, z/v.z, w);}
Vector4f Vector4f::operator/(const Vector4f &v) const
    {return Vector4f(x/v.x, y/v.y, z/v.z, w/v.w);}

Vector4f operator+(const float k, const Vector4f &v) {return v+k;}
Vector4f operator*(const float k, const Vector4f &v) {return v*k;}

Vector4f &Vector4f::operator+=(const float k)
    {x += k;   y += k;   z += k;   w += k;   return *this;}
Vector4f &Vector4f::operator+=(const Vector2f &v)
    {x += v.x; y += v.y;                     return *this;}
Vector4f &Vector4f::operator+=(const Vector3f &v)
    {x += v.x; y += v.y; z += v.z;           return *this;}
Vector4f &Vector4f::operator+=(const Vector4f &v)
    {x += v.x; y += v.y; z += v.z; w += v.w; return *this;}

Vector4f &Vector4f::operator-=(const float k)
    {x -= k;   y -= k;   z -= k;   w -= k;   return *this;}
Vector4f &Vector4f::operator-=(const Vector2f &v)
    {x -= v.x; y -= v.y;                     return *this;}
Vector4f &Vector4f::operator-=(const Vector3f &v)
    {x -= v.x; y -= v.y; z -= v.z;           return *this;}
Vector4f &Vector4f::operator-=(const Vector4f &v)
    {x -= v.x; y -= v.y; z -= v.z; w -= v.w; return *this;}

Vector4f &Vector4f::operator*=(const float k)
    {x *= k;   y *= k;   z *= k;   w *= k;   return *this;}
Vector4f &Vector4f::operator*=(const Vector2f &v)
    {x *= v.x; y *= v.y;                     return *this;}
Vector4f &Vector4f::operator*=(const Vector3f &v)
    {x *= v.x; y *= v.y; z *= v.z;           return *this;}
Vector4f &Vector4f::operator*=(const Vector4f &v)
    {x *= v.x; y *= v.y; z *= v.z; w *= v.w; return *this;}

Vector4f &Vector4f::operator/=(const float k)
    {x /= k;   y /= k;   z /= k;   w /= k;   return *this;}
Vector4f &Vector4f::operator/=(const Vector2f &v)
    {x /= v.x; y /= v.y;                     return *this;}
Vector4f &Vector4f::operator/=(const Vector3f &v)
    {x /= v.x; y /= v.y; z /= v.z;           return *this;}
Vector4f &Vector4f::operator/=(const Vector4f &v)
    {x /= v.x; y /= v.y; z /= v.z; w /= v.w; return *this;}

bool Vector4f::operator==(const Vector4f &v) const
    {return x == v.x && y == v.y && z == v.z && w == v.w;}
bool Vector4f::operator!=(const Vector4f &v) const
    {return x != v.x || y != v.y || z != v.z || w != v.w;}

float Vector4f::length() const           {return sqrtf(length2());}
float Vector4f::length2() const          {return x*x + y*y + z*z + w*w;}
bool Vector4f::is_nonzero() const        {return x!=0 || y!=0 || z!=0 || w!=0;}
bool Vector4f::is_length_in(const float k) const {return length2() <= k*k;}
void Vector4f::normalize() {
    if (LIKELY(length2() != 0)) {
        *this /= length();
    } else {
        x = y = z = w = 0;
    }
}
void Vector4f::set_length(const float k) {
    if (LIKELY(length2() != 0)) {
        *this *= k/length();
    } else {
        x = y = z = w = 0;
    }
}
void Vector4f::cap_length(const float k)
    {if (length2() > k*k) *this *= k/length();}
float Vector4f::dot(const Vector4f &v) const
    {return x*v.x + y*v.y + z*v.z + w*v.w;}
Vector4f Vector4f::cross(const Vector4f &v2, const Vector4f &v3) const
    {return Vector4f(+(  y * (v2.z*v3.w - v2.w*v3.z)
                       + z * (v2.w*v3.y - v2.y*v3.w)
                       + w * (v2.y*v3.z - v2.z*v3.y)),
                     -(  x * (v2.z*v3.w - v2.w*v3.z)
                       + z * (v2.w*v3.x - v2.x*v3.w)
                       + w * (v2.x*v3.z - v2.z*v3.x)),
                     +(  x * (v2.y*v3.w - v2.w*v3.y)
                       + y * (v2.w*v3.x - v2.x*v3.w)
                       + w * (v2.x*v3.y - v2.y*v3.x)),
                     -(  x * (v2.y*v3.z - v2.z*v3.y)
                       + y * (v2.z*v3.x - v2.x*v3.z)
                       + z * (v2.x*v3.y - v2.y*v3.x)));}

/*----------------------------------*/

#endif  // __cplusplus

/*************************************************************************/
/*************************************************************************/

#endif  // SIL_MATH_VECTOR_H

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
