/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/graphics/framebuffer.c: Framebuffer management.
 */

#include "src/base.h"
#include "src/graphics.h"
#include "src/graphics/internal.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/utility/id-array.h"

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Simple data structure for managing framebuffer data. */
typedef struct Framebuffer Framebuffer;
struct Framebuffer {
    SysFramebuffer *sysfb;
    int width, height;
};

/* Array of allocated framebuffers. */
static IDArray framebuffers = ID_ARRAY_INITIALIZER(10);

/*-----------------------------------------------------------------------*/

/**
 * VALIDATE_FRAMEBUFFER:  Validate the framebuffer ID passed to a
 * framebuffer management routine, and store the corresponding pointer in
 * the variable "framebuffer".  If the framebuffer pointer is invalid, the
 * "error_return" statement is executed; this may consist of multiple
 * statements, but must include a "return" to exit the function.
 */
#define VALIDATE_FRAMEBUFFER(id,framebuffer,error_return) \
    ID_ARRAY_VALIDATE(&framebuffers, (id), Framebuffer *, framebuffer, \
                      DLOG("Framebuffer ID %d is invalid", __id); error_return)

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

int graphics_framebuffer_supported(void)
{
    return sys_framebuffer_supported();
}

/*-----------------------------------------------------------------------*/

int graphics_create_framebuffer(int width, int height)
{
    if (UNLIKELY(width <= 0) || UNLIKELY(height <= 0)) {
        DLOG("Invalid framebuffer size: %dx%d", width, height);
        goto error_return;
    }

    Framebuffer *framebuffer = mem_alloc(sizeof(*framebuffer), 0, 0);
    if (!framebuffer) {
        DLOG("No memory for %dx%d framebuffer", width, height);
        goto error_return;
    }

    framebuffer->sysfb = sys_framebuffer_create(width, height);
    if (!framebuffer->sysfb) {
        goto error_free_framebuffer;
    }
    framebuffer->width = width;
    framebuffer->height = height;

    const int id = id_array_register(&framebuffers, framebuffer);
    if (!id) {
        DLOG("Failed to register framebuffer");
        goto error_destroy_sysfb;
    }

    return id;

  error_destroy_sysfb:
    sys_framebuffer_destroy(framebuffer->sysfb);
  error_free_framebuffer:
    mem_free(framebuffer);
  error_return:
    return 0;
}

/*-----------------------------------------------------------------------*/

void graphics_destroy_framebuffer(int framebuffer_id)
{
    if (framebuffer_id) {
        Framebuffer *framebuffer;
        VALIDATE_FRAMEBUFFER(framebuffer_id, framebuffer, return);
        sys_framebuffer_destroy(framebuffer->sysfb);
        mem_free(framebuffer);
        id_array_release(&framebuffers, framebuffer_id);
    }
}

/*-----------------------------------------------------------------------*/

int graphics_framebuffer_width(int framebuffer_id)
{
    Framebuffer *framebuffer;
    VALIDATE_FRAMEBUFFER(framebuffer_id, framebuffer, return 0);
    return framebuffer->width;
}

/*-----------------------------------------------------------------------*/

int graphics_framebuffer_height(int framebuffer_id)
{
    Framebuffer *framebuffer;
    VALIDATE_FRAMEBUFFER(framebuffer_id, framebuffer, return 0);
    return framebuffer->height;
}

/*-----------------------------------------------------------------------*/

void graphics_bind_framebuffer(int framebuffer_id)
{
    if (framebuffer_id) {
        Framebuffer *framebuffer;
        VALIDATE_FRAMEBUFFER(framebuffer_id, framebuffer, return);
        sys_framebuffer_bind(framebuffer->sysfb);
    } else {
        sys_framebuffer_bind(NULL);
    }
}

/*-----------------------------------------------------------------------*/

void graphics_apply_framebuffer(int framebuffer_id)
{
    Framebuffer *framebuffer;
    VALIDATE_FRAMEBUFFER(framebuffer_id, framebuffer, return);
    sys_framebuffer_apply(framebuffer->sysfb);
}

/*-----------------------------------------------------------------------*/

void graphics_set_framebuffer_antialias(int framebuffer_id, int on)
{
    Framebuffer *framebuffer;
    VALIDATE_FRAMEBUFFER(framebuffer_id, framebuffer, return);
    sys_framebuffer_set_antialias(framebuffer->sysfb, on);
}

/*-----------------------------------------------------------------------*/

void graphics_discard_framebuffer_data(int framebuffer_id)
{
    Framebuffer *framebuffer;
    VALIDATE_FRAMEBUFFER(framebuffer_id, framebuffer, return);
    sys_framebuffer_discard_data(framebuffer->sysfb);
}

/*************************************************************************/
/*********************** Library-internal routines ***********************/
/*************************************************************************/

void framebuffer_cleanup(void)
{
    for (int i = 1; i < id_array_size(&framebuffers); i++) {
        if (id_array_get(&framebuffers, i)) {
            graphics_destroy_framebuffer(i);
        }
    }
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
