/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/graphics/state.c: Render state manipulation.
 */

#include "src/base.h"
#include "src/graphics.h"
#include "src/graphics/internal.h"
#include "src/math.h"
#include "src/sysdep.h"

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Current size of the viewport, or zero if the viewport has never been set. */
static int viewport_width, viewport_height;

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

void graphics_set_viewport(int left, int bottom, int width, int height)
{
    if (UNLIKELY(left < 0)
     || UNLIKELY(bottom < 0)
     || UNLIKELY(width <= 0)
     || UNLIKELY(height <= 0)) {
        DLOG("Invalid parameters: %d %d %d %d", left, bottom, width, height);
        return;
    }

    sys_graphics_set_viewport(left, bottom, width, height);
    viewport_width = width;
    viewport_height = height;
}

/*-----------------------------------------------------------------------*/

int graphics_viewport_width(void)
{
    return viewport_width;
}

/*-----------------------------------------------------------------------*/

int graphics_viewport_height(void)
{
    return viewport_height;
}

/*-----------------------------------------------------------------------*/

void graphics_set_clip_region(int left, int bottom, int width, int height)
{
    if (UNLIKELY(left < 0)
     || UNLIKELY(bottom < 0)
     || UNLIKELY(width < 0)
     || UNLIKELY(height < 0)) {
        DLOG("Invalid parameters: %d %d %d %d", left, bottom, width, height);
        return;
    }

    if (width == 0 && height == 0) {
        sys_graphics_set_int_param(SYS_GRAPHICS_PARAM_CLIP, 0);
    } else if (width == 0 || height == 0) {
        DLOG("Invalid parameters: %d %d %d %d", left, bottom, width, height);
        return;
    } else {
        sys_graphics_set_int_param(SYS_GRAPHICS_PARAM_CLIP, 1);
        sys_graphics_set_clip_region(left, bottom, width, height);
    }
}

/*-----------------------------------------------------------------------*/

void graphics_set_depth_range(float near, float far)
{
    if (UNLIKELY(near < 0)
     || UNLIKELY(near > 1)
     || UNLIKELY(far < 0)
     || UNLIKELY(far > 1)) {
        DLOG("Invalid parameters: %g %g", near, far);
        return;
    }

    sys_graphics_set_depth_range(near, far);
}

/*-----------------------------------------------------------------------*/

void graphics_set_blend(GraphicsBlendOperation operation,
                        GraphicsBlendFactor src_factor,
                        GraphicsBlendFactor dest_factor)
{
    if (UNLIKELY((int)operation < 1)
     || UNLIKELY(operation > GRAPHICS_BLEND_RSUB)
     || UNLIKELY((int)src_factor < 0)
     || UNLIKELY(src_factor > GRAPHICS_BLEND_INV_CONSTANT)
     || UNLIKELY((int)dest_factor < 0)
     || UNLIKELY(dest_factor > GRAPHICS_BLEND_INV_CONSTANT)) {
        DLOG("Invalid parameters: %d %d %d",
             operation, src_factor, dest_factor);
        return;
    }

    sys_graphics_set_blend(operation, src_factor, dest_factor);
}

/*-----------------------------------------------------------------------*/

void graphics_set_projection_matrix(const Matrix4f *matrix)
{
    if (UNLIKELY(!matrix)) {
        DLOG("matrix == NULL");
        return;
    }

    sys_graphics_set_matrix_param(SYS_GRAPHICS_PARAM_PROJECTION_MATRIX, matrix);
}

/*-----------------------------------------------------------------------*/

void graphics_set_view_matrix(const Matrix4f *matrix)
{
    if (UNLIKELY(!matrix)) {
        DLOG("matrix == NULL");
        return;
    }

    sys_graphics_set_matrix_param(SYS_GRAPHICS_PARAM_VIEW_MATRIX, matrix);
}

/*-----------------------------------------------------------------------*/

void graphics_set_model_matrix(const Matrix4f *matrix)
{
    if (UNLIKELY(!matrix)) {
        DLOG("matrix == NULL");
        return;
    }

    sys_graphics_set_matrix_param(SYS_GRAPHICS_PARAM_MODEL_MATRIX, matrix);
}

/*-----------------------------------------------------------------------*/

void graphics_get_projection_matrix(Matrix4f *matrix_ret)
{
    if (UNLIKELY(!matrix_ret)) {
        DLOG("matrix_ret == NULL");
        return;
    }

    sys_graphics_get_matrix_param(
        SYS_GRAPHICS_PARAM_PROJECTION_MATRIX, matrix_ret);
}

/*-----------------------------------------------------------------------*/

void graphics_get_view_matrix(Matrix4f *matrix_ret)
{
    if (UNLIKELY(!matrix_ret)) {
        DLOG("matrix_ret == NULL");
        return;
    }

    sys_graphics_get_matrix_param(SYS_GRAPHICS_PARAM_VIEW_MATRIX, matrix_ret);
}

/*-----------------------------------------------------------------------*/

void graphics_get_model_matrix(Matrix4f *matrix_ret)
{
    if (UNLIKELY(!matrix_ret)) {
        DLOG("matrix_ret == NULL");
        return;
    }

    sys_graphics_get_matrix_param(SYS_GRAPHICS_PARAM_MODEL_MATRIX, matrix_ret);
}

/*-----------------------------------------------------------------------*/

void graphics_set_parallel_projection(
    float left, float right, float bottom, float top, float near, float far)
{
    const float dx = right - left;
    const float dy = top - bottom;
    const float dz = far - near;

    Matrix4f m = mat4_identity;
    m._11 = 2.0f / dx;
    m._22 = 2.0f / dy;
    m._33 = 2.0f / dz;
    m._41 = -(right + left) / dx;
    m._42 = -(top + bottom) / dy;
    m._43 = -(far + near) / dz;
    sys_graphics_set_matrix_param(SYS_GRAPHICS_PARAM_PROJECTION_MATRIX, &m);
}

/*-----------------------------------------------------------------------*/

void graphics_set_perspective_projection(
    float y_fov, float aspect, float near, float far)
{
    const float cotangent = dcosf(y_fov/2) / dsinf(y_fov/2);
    const float dz = far - near;

    Matrix4f m = mat4_identity;
    m._11 = cotangent / aspect;
    m._22 = cotangent;
    m._33 = (far + near) / dz;
    m._34 = 1;
    m._43 = (-2 * far * near) / dz;
    m._44 = 0;
    sys_graphics_set_matrix_param(SYS_GRAPHICS_PARAM_PROJECTION_MATRIX, &m);
}

/*-----------------------------------------------------------------------*/

void graphics_enable_alpha_test(int on)
{
    sys_graphics_set_int_param(SYS_GRAPHICS_PARAM_ALPHA_TEST, (on != 0));
}

/*-----------------------------------------------------------------------*/

void graphics_set_alpha_test_comparison(GraphicsComparisonType type)
{
    switch (type) {
      case GRAPHICS_COMPARISON_LESS:
      case GRAPHICS_COMPARISON_LESS_EQUAL:
      case GRAPHICS_COMPARISON_GREATER_EQUAL:
      case GRAPHICS_COMPARISON_GREATER:
        sys_graphics_set_int_param(
            SYS_GRAPHICS_PARAM_ALPHA_TEST_COMPARISON, type);
        return;
      case GRAPHICS_COMPARISON_FALSE:
      case GRAPHICS_COMPARISON_TRUE:
      case GRAPHICS_COMPARISON_EQUAL:
      case GRAPHICS_COMPARISON_NOT_EQUAL:
        break;  // Fall through to error message.
    }
    DLOG("Invalid comparison type: %d", type);
}

/*-----------------------------------------------------------------------*/

void graphics_set_alpha_test_reference(float alpha)
{
    sys_graphics_set_float_param(SYS_GRAPHICS_PARAM_ALPHA_REFERENCE, alpha);
}

/*-----------------------------------------------------------------------*/

void graphics_set_blend_color(const Vector4f *color)
{
    if (UNLIKELY(!color)) {
        DLOG("color == NULL");
        return;
    }

    sys_graphics_set_vec4_param(SYS_GRAPHICS_PARAM_BLEND_COLOR, color);
}

/*-----------------------------------------------------------------------*/

void graphics_enable_color_write(int red, int green, int blue)
{
    sys_graphics_set_int_param(
        SYS_GRAPHICS_PARAM_COLOR_WRITE,
        (red ? 1<<0 : 0) | (green ? 1<<1 : 0) | (blue ? 1<<2 : 0));
}

/*-----------------------------------------------------------------------*/

void graphics_enable_depth_test(int on)
{
    sys_graphics_set_int_param(SYS_GRAPHICS_PARAM_DEPTH_TEST, (on != 0));
}

/*-----------------------------------------------------------------------*/

void graphics_set_depth_test_comparison(GraphicsComparisonType type)
{
    switch (type) {
      case GRAPHICS_COMPARISON_LESS:
      case GRAPHICS_COMPARISON_LESS_EQUAL:
      case GRAPHICS_COMPARISON_GREATER_EQUAL:
      case GRAPHICS_COMPARISON_GREATER:
        sys_graphics_set_int_param(
            SYS_GRAPHICS_PARAM_DEPTH_TEST_COMPARISON, type);
        return;
      case GRAPHICS_COMPARISON_FALSE:
      case GRAPHICS_COMPARISON_TRUE:
      case GRAPHICS_COMPARISON_EQUAL:
      case GRAPHICS_COMPARISON_NOT_EQUAL:
        break;  // Fall through to error message.
    }
    DLOG("Invalid comparison type: %d", type);
}

/*-----------------------------------------------------------------------*/

void graphics_enable_depth_write(int on)
{
    sys_graphics_set_int_param(SYS_GRAPHICS_PARAM_DEPTH_WRITE, (on != 0));
}

/*-----------------------------------------------------------------------*/

void graphics_set_face_cull(GraphicsFaceCullMode mode)
{
    sys_graphics_set_int_param(SYS_GRAPHICS_PARAM_FACE_CULL,
                               (mode != GRAPHICS_FACE_CULL_NONE));
    sys_graphics_set_int_param(SYS_GRAPHICS_PARAM_FACE_CULL_CW,
                               (mode == GRAPHICS_FACE_CULL_CW));
}

/*-----------------------------------------------------------------------*/

void graphics_set_fixed_color(const Vector4f *color)
{
    if (UNLIKELY(!color)) {
        DLOG("color == NULL");
        return;
    }

    sys_graphics_set_vec4_param(SYS_GRAPHICS_PARAM_FIXED_COLOR, color);
}

/*-----------------------------------------------------------------------*/

void graphics_enable_fog(int on)
{
    sys_graphics_set_int_param(SYS_GRAPHICS_PARAM_FOG, (on != 0));
}

/*-----------------------------------------------------------------------*/

void graphics_set_fog_start(float distance)
{
    sys_graphics_set_float_param(SYS_GRAPHICS_PARAM_FOG_START, distance);
}

/*-----------------------------------------------------------------------*/

void graphics_set_fog_end(float distance)
{
    sys_graphics_set_float_param(SYS_GRAPHICS_PARAM_FOG_END, distance);
}

/*-----------------------------------------------------------------------*/

void graphics_set_fog_color(const Vector4f *color)
{
    if (UNLIKELY(!color)) {
        DLOG("color == NULL");
        return;
    }

    sys_graphics_set_vec4_param(SYS_GRAPHICS_PARAM_FOG_COLOR, color);
}

/*-----------------------------------------------------------------------*/

void graphics_set_line_width(float width)
{
    sys_graphics_set_float_param(SYS_GRAPHICS_PARAM_LINE_WIDTH, width);
}

/*-----------------------------------------------------------------------*/

void graphics_set_point_size(float size)
{
    sys_graphics_set_float_param(SYS_GRAPHICS_PARAM_POINT_SIZE, size);
}

/*-----------------------------------------------------------------------*/

void graphics_enable_stencil_test(int on)
{
    sys_graphics_set_int_param(SYS_GRAPHICS_PARAM_STENCIL_TEST, (on != 0));
}

/*-----------------------------------------------------------------------*/

void graphics_set_stencil_comparison(
    GraphicsComparisonType type, unsigned int reference, unsigned int mask)
{
    /* This switch statement is just a check for validity. */
    switch (type) {
      case GRAPHICS_COMPARISON_FALSE:
      case GRAPHICS_COMPARISON_TRUE:
      case GRAPHICS_COMPARISON_EQUAL:
      case GRAPHICS_COMPARISON_NOT_EQUAL:
      case GRAPHICS_COMPARISON_LESS:
      case GRAPHICS_COMPARISON_LESS_EQUAL:
      case GRAPHICS_COMPARISON_GREATER_EQUAL:
      case GRAPHICS_COMPARISON_GREATER:
        sys_graphics_set_int_param(
            SYS_GRAPHICS_PARAM_STENCIL_COMPARISON, type);
        sys_graphics_set_int_param(
            SYS_GRAPHICS_PARAM_STENCIL_REFERENCE, reference);
        sys_graphics_set_int_param(
            SYS_GRAPHICS_PARAM_STENCIL_MASK, mask);
        return;
    }
    DLOG("Invalid comparison type: %d", type);
}

/*-----------------------------------------------------------------------*/

void graphics_set_stencil_operations(
    GraphicsStencilOp sfail, GraphicsStencilOp dfail, GraphicsStencilOp dpass)
{
    /* These switch statements are just checks for validity. */
    switch (sfail) {
      case GRAPHICS_STENCIL_KEEP:
      case GRAPHICS_STENCIL_CLEAR:
      case GRAPHICS_STENCIL_REPLACE:
      case GRAPHICS_STENCIL_INCR:
      case GRAPHICS_STENCIL_DECR:
      case GRAPHICS_STENCIL_INVERT:
        switch (dfail) {
          case GRAPHICS_STENCIL_KEEP:
          case GRAPHICS_STENCIL_CLEAR:
          case GRAPHICS_STENCIL_REPLACE:
          case GRAPHICS_STENCIL_INCR:
          case GRAPHICS_STENCIL_DECR:
          case GRAPHICS_STENCIL_INVERT:
            switch (dpass) {
              case GRAPHICS_STENCIL_KEEP:
              case GRAPHICS_STENCIL_CLEAR:
              case GRAPHICS_STENCIL_REPLACE:
              case GRAPHICS_STENCIL_INCR:
              case GRAPHICS_STENCIL_DECR:
              case GRAPHICS_STENCIL_INVERT:
                sys_graphics_set_int_param(
                    SYS_GRAPHICS_PARAM_STENCIL_OP_SFAIL, sfail);
                sys_graphics_set_int_param(
                    SYS_GRAPHICS_PARAM_STENCIL_OP_DFAIL, dfail);
                sys_graphics_set_int_param(
                    SYS_GRAPHICS_PARAM_STENCIL_OP_DPASS, dpass);
                return;
            }
        }
    }
    DLOG("Invalid operations: %d %d %d", sfail, dfail, dpass);
}

/*-----------------------------------------------------------------------*/

void graphics_set_texture_offset(const Vector2f *offset)
{
    if (UNLIKELY(!offset)) {
        DLOG("offset == NULL");
        return;
    }

    sys_graphics_set_vec2_param(SYS_GRAPHICS_PARAM_TEXTURE_OFFSET, offset);
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
