/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sound/decode-ogg.c: Audio decoder for Ogg Vorbis data.
 */

#ifdef SIL_SOUND_INCLUDE_OGG  // To the end of the file.

#include "src/base.h"
#include "src/memory.h"
#include "src/sound.h"
#include "src/sound/decode.h"
#include "src/sound/decode-ogg.h"

/*************************************************************************/
/************************* Configuration options *************************/
/*************************************************************************/

/**
 * SIL_SOUND_USE_TREMOR:  If defined, use the integer-based Tremor decoder
 * rather than the standard (double precision floating-point) decoder from
 * libvorbis.
 */
// #define SIL_SOUND_USE_TREMOR

/*-----------------------------------------------------------------------*/

#ifdef SIL_SOUND_USE_TREMOR
# include <ivorbisfile.h>
#else
# define OV_EXCLUDE_STATIC_CALLBACKS
# include <vorbis/vorbisfile.h>
#endif

#include <errno.h>

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* libvorbisfile callback routine declarations. */

/**
 * ogg_read:  Read callback for libvorbisfile.
 *
 * [Parameters]
 *     ptr: Read buffer.
 *     size, nmemb: Number of bytes to read (size * nmemb).
 *     datasource: Data source (a SoundDecodeHandle pointer).
 * [Return value]
 *     Number of "size" byte units read, or zero (with errno set
 *     appropriately) on error.
 */
static size_t ogg_read(void *ptr, size_t size, size_t nmemb, void *datasource);

/**
 * ogg_seek:  Seek callback for libvorbisfile.
 *
 * [Parameters]
 *     datasource: Data source (a SoundDecodeHandle pointer).
 *     offset: New read position, in bytes.
 *     whence: Meaning of "offset" (a stdio SEEK_* constant).
 * [Return value]
 *     0 on success, -1 on error.
 */
static int ogg_seek(void *datasource, ogg_int64_t offset, int whence);

/**
 * ogg_tell:  Tell callback for libvorbisfile.
 *
 * [Parameters]
 *     datasource: Data source (a SoundDecodeHandle pointer).
 * [Return value]
 *     Current read position, in bytes, or -1 on error.
 */
static long ogg_tell(void *datasource);

/**
 * ogg_close:  Close callback for libvorbisfile.  Does nothing, but is
 * required for older versions (pre-1.2.0) of libvorbisfile, which crash
 * if NULL is specified for the close callback.
 *
 * [Parameters]
 *     datasource: Data source (a SoundDecodeHandle pointer).
 * [Return value]
 *     Always zero.
 */
static int ogg_close(void *datasource);

static const ov_callbacks ogg_callbacks = {
    .read_func  = ogg_read,
    .seek_func  = ogg_seek,
    .close_func = ogg_close,
    .tell_func  = ogg_tell,
};

#ifdef INCLUDE_TESTS
/* Set true to force a one-shot seek error. */
static uint8_t test_fail_ogg_seek;
#endif

/*-----------------------------------------------------------------------*/

/* Private data structure. */

struct SoundDecodePrivate {
    OggVorbis_File vf;
    int filepos;       // Read position for libvorbisfile.
    uint8_t error;     // True if a fatal decoding error has occurred.
};

/*-----------------------------------------------------------------------*/

/* Method implementation declarations. */

static void decode_ogg_reset(SoundDecodeHandle *this);
static int decode_ogg_get_pcm(
    SoundDecodeHandle *this, int16_t *pcm_buffer, int pcm_len,
    int *loop_offset_ret);
static void decode_ogg_close(SoundDecodeHandle *this);

/*************************************************************************/
/************************ Method implementations *************************/
/*************************************************************************/

int decode_ogg_open(SoundDecodeHandle *this)
{
    this->reset   = decode_ogg_reset;
    this->get_pcm = decode_ogg_get_pcm;
    this->close   = decode_ogg_close;

    this->private = mem_alloc(sizeof(*this->private), 0,
                              MEM_ALLOC_TEMP | MEM_ALLOC_CLEAR);
    if (!this->private) {
        DLOG("Out of memory");
        goto error_return;
    }

    if (ov_open_callbacks(this, &this->private->vf, NULL, 0,
                          ogg_callbacks) != 0) {
        DLOG("ov_open() failed");
        goto error_free_private;
    }

    vorbis_info *info = ov_info(&this->private->vf, -1);
    /* This can never fail in current libvorbis/tremor. */
    ASSERT(info != NULL, goto error_clear_vorbis);

    if (info->channels == 1) {
        this->stereo = 0;
    } else if (info->channels == 2) {
        this->stereo = 1;
    } else {
        DLOG("Bad channel count %d", info->channels);
        goto error_clear_vorbis;
    }
    this->native_freq = info->rate;

    return 1;

  error_clear_vorbis:
    ov_clear(&this->private->vf);
  error_free_private:
    mem_free(this->private);
  error_return:
    return 0;
}

/*-----------------------------------------------------------------------*/

static void decode_ogg_reset(SoundDecodeHandle *this)
{
    const int error = ov_pcm_seek(&this->private->vf, 0);
    if (error < 0) {
        DLOG("Seek error: %d", error);
        this->private->error = 1;
    } else {
        this->private->error = 0;
    }
}

/*-----------------------------------------------------------------------*/

static int decode_ogg_get_pcm(
    SoundDecodeHandle *this, int16_t *pcm_buffer, int pcm_len,
    int *loop_offset_ret)
{
    SoundDecodePrivate *private = this->private;
    const int channels = this->stereo ? 2 : 1;
    /* loopend is unused if looplen < 0. */
    const int loopend = this->internal->loopstart + this->internal->looplen;
    *loop_offset_ret = 0;

    if (private->error) {
        return 0;
    }

    int allow_loop = 1;  // Safety valve to avoid infinite loops.
    int copied = 0;
    while (copied < pcm_len) {

        int64_t curpos = ov_pcm_tell(&private->vf);
        ASSERT(curpos >= 0, curpos = 0);
        int toread = pcm_len - copied;
        if (this->internal->looplen > 0 && curpos < loopend) {
            toread = min(toread, loopend - curpos);
        }

        int bitstream_unused;
        const int thisread = ov_read(
            &private->vf, (char *)(pcm_buffer + (copied * channels)),
            toread * (2*channels),
#ifndef SIL_SOUND_USE_TREMOR
            /*bigendianp*/ 0, /*word*/ 2, /*sgned*/ 1,
#endif
            &bitstream_unused);
        if (thisread == 0) {
            /* Reached the end of the audio stream. */
            if (this->internal->loop && this->internal->looplen != 0
             && (this->internal->looplen < 0 || curpos <= loopend)) {
                if ((int64_t)this->internal->loopstart < curpos) {
                    /* allow_loop will only be false if ov_read() returns EOF
                     * despite having seeked backwards on the previous loop,
                     * which can't happen without a bug in libogg/libvorbis. */
                    ASSERT(allow_loop, break);
                    if (this->internal->looplen > 0) {
                        DLOG("Warning: loopend %u > pcmlen %lld, looping"
                             " early", loopend, (long long)curpos);
                    }
                    const int error = ov_pcm_seek(
                        &private->vf, this->internal->loopstart);
                    if (error < 0) {
                        DLOG("Seek error: %d", error);
                        private->error = 1;
                        break;
                    } else {
                        *loop_offset_ret += curpos - this->internal->loopstart;
                        curpos = this->internal->loopstart;
                        allow_loop = 0;
                    }
                } else {
                    DLOG("Tried to loop but no data available, bailing");
                    break;
                }
            } else {
                break;
            }
        } else if (thisread == OV_HOLE) {
            DLOG("WARNING: decompression error, data dropped!");
            /* Resync the position counter since the decoder may have
             * skipped past the hole. */
            curpos = ov_pcm_tell(&private->vf);
        } else if (thisread < 0) {
            DLOG("Decompression error: %d", thisread);
            private->error = 1;
            break;
        } else {
            copied += thisread / (2*channels);
            curpos += thisread / (2*channels);
        }

        const int64_t newpos = ov_pcm_tell(&private->vf);
        if (UNLIKELY(newpos != curpos)) {
            DLOG("WARNING: decode position mismatch: current position is"
                 " %lld but should be %lld!  Audio data may be corrupt.",
                 (long long)newpos, (long long)curpos);
            curpos = newpos;
        }

        if (this->internal->loop && this->internal->looplen > 0
         && curpos == loopend)  {
            const int error = ov_pcm_seek(
                &private->vf, this->internal->loopstart);
            if (error < 0) {
                DLOG("Seek error: %d", error);
                private->error = 1;
                break;
            } else {
                *loop_offset_ret += curpos - this->internal->loopstart;
            }
        }

    }  // while (copied < pcm_len)

    return copied;
}

/*-----------------------------------------------------------------------*/

static void decode_ogg_close(SoundDecodeHandle *this)
{
    ov_clear(&this->private->vf);
    mem_free(this->private);
}

/*************************************************************************/
/************************* Test control routines *************************/
/*************************************************************************/

#ifdef INCLUDE_TESTS

void sound_decode_ogg_test_fail_next_seek(void)
{
    test_fail_ogg_seek = 1;
}

#endif

/*************************************************************************/
/******************** I/O routines for libvorbisfile *********************/
/*************************************************************************/

static size_t ogg_read(void *ptr, size_t size, size_t nmemb, void *datasource)
{
    ASSERT(ptr != NULL, errno = EINVAL; return 0);
    ASSERT(datasource != NULL, errno = EINVAL; return 0);
    SoundDecodeHandle *handle = (SoundDecodeHandle *)datasource;

    const uint8_t *data;
    const int nread =
        decode_get_data(handle, handle->private->filepos, size * nmemb, &data);
    memcpy(ptr, data, nread);
    handle->private->filepos += nread;
    errno = 0;
    return nread / size;
}

/*-----------------------------------------------------------------------*/

static int ogg_seek(void *datasource, ogg_int64_t offset, int whence)
{
    ASSERT(datasource != NULL, errno = EINVAL; return -1);
    SoundDecodeHandle *handle = (SoundDecodeHandle *)datasource;

#ifdef INCLUDE_TESTS
    if (test_fail_ogg_seek) {
        test_fail_ogg_seek = 0;
        errno = EIO;
        return -1;
    }
#endif

    if (whence == SEEK_CUR) {
        offset += handle->private->filepos;
    } else if (whence == SEEK_END) {
        offset += handle->internal->datalen;
    }
    handle->private->filepos = bound(offset, 0, handle->internal->datalen);
    return 0;
}

/*-----------------------------------------------------------------------*/

static long ogg_tell(void *datasource)
{
    ASSERT(datasource != NULL, errno = EINVAL; return -1);
    SoundDecodeHandle *handle = (SoundDecodeHandle *)datasource;

    return handle->private->filepos;
}

/*-----------------------------------------------------------------------*/

static int ogg_close(UNUSED void *datasource)
{
    return 0;
}

/*************************************************************************/
/*************************************************************************/

#endif  // SIL_SOUND_INCLUDE_OGG

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
