/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/linux/sound.c: Linux audio output interface.  Uses the ALSA
 * library.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/sound/mixer.h"
#include "src/sysdep.h"
#include "src/thread.h"
#include "src/time.h"

#include <alloca.h>
#include <alsa/asoundlib.h>

/*************************************************************************/
/************************* Configuration options *************************/
/*************************************************************************/

/**
 * SOUND_BUFLEN:  Number of samples to send to the hardware in a single
 * output call.  Should be the same as the mixer buffer length (in
 * sound/mixer.c) for best performance.
 */
#define SOUND_BUFLEN  1024

/**
 * SOUND_RATE:  Output sampling rate to use.
 */
#define SOUND_RATE  48000

/**
 * PLAYBACK_THREAD_PRIORITY:  Thread priority for the playback thread.
 */
#define PLAYBACK_THREAD_PRIORITY  5

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Have we been initialized? */
static int sound_initted;

/* ALSA sound device handle. */
static snd_pcm_t *pcm_handle = NULL;
/* Number of samples in the internal audio buffer. */
static unsigned int hw_buffer_size;
/* Number of samples in the optimum transfer block size. */
static unsigned int hw_period_size;

/* Thread ID of playback thread, and flag used to tell it to stop. */
static int playback_thread_id;
static uint8_t playback_thread_stop;

/* Sample output counter and counter of silent buffers for ALSA bug
 * workaround (see playback_thread()). */
static uint32_t samples_played;
static uint32_t silent_buffers_played;

/*-----------------------------------------------------------------------*/

/* Local routine declarations. */

/**
 * playback_thread:  Thread which mixes audio data and sends it to the
 * audio output device.
 *
 * [Parameters]
 *     userdata: Opaque user data pointer (unused).
 * [Return value]
 *     0
 */
static int playback_thread(void *userdata);

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

int sys_sound_init(const char *device_name, int num_channels,
                   int *channels_ret)
{
    PRECOND(device_name != NULL, goto error_return);
    PRECOND(num_channels > 0, goto error_return);
    PRECOND(channels_ret != NULL, goto error_return);
    if (sound_initted) {
        DLOG("Tried to init twice!");
        goto error_return;
    }

    /* Open the ALSA device. */

    /* Convenience macro for checking ALSA function return codes. */
    #define CHECK_ALSA(call)  do {                       \
        int result;                                      \
        if ((result = (call)) < 0) {                     \
            DLOG("%s: %s", #call, snd_strerror(result)); \
            goto error_close_device;                     \
        }                                                \
    } while (0)

    snd_pcm_hw_params_t *hwparams;
    snd_pcm_hw_params_alloca(&hwparams);
    snd_pcm_sw_params_t *swparams;
    snd_pcm_sw_params_alloca(&swparams);

    if (!device_name || !*device_name) {
        device_name = "default";
    }
    pcm_handle = NULL;
    CHECK_ALSA(snd_pcm_open(
                   &pcm_handle, device_name, SND_PCM_STREAM_PLAYBACK, 0));
    CHECK_ALSA(snd_pcm_hw_params_any(pcm_handle, hwparams));
    CHECK_ALSA(snd_pcm_hw_params_set_access(
                   pcm_handle, hwparams, SND_PCM_ACCESS_RW_INTERLEAVED));
    CHECK_ALSA(snd_pcm_hw_params_set_format(
                   pcm_handle, hwparams, SND_PCM_FORMAT_S16_LE));
    CHECK_ALSA(snd_pcm_hw_params_set_channels(pcm_handle, hwparams, 2));
    CHECK_ALSA(snd_pcm_hw_params_set_rate(pcm_handle, hwparams, SOUND_RATE, 0));
    CHECK_ALSA(snd_pcm_hw_params_set_buffer_size_near(
             pcm_handle, hwparams, (snd_pcm_uframes_t[1]){SOUND_BUFLEN}));
    CHECK_ALSA(snd_pcm_hw_params(pcm_handle, hwparams));

    snd_pcm_uframes_t buffer_size, period_size;
    CHECK_ALSA(snd_pcm_hw_params_get_buffer_size(hwparams, &buffer_size));
    CHECK_ALSA(snd_pcm_hw_params_get_period_size(hwparams, &period_size, NULL));
    hw_buffer_size = (unsigned int)buffer_size;
    hw_period_size = (unsigned int)period_size;
    DLOG("Audio output buffer size: %u, period: %u",
         hw_buffer_size, hw_period_size);

    CHECK_ALSA(snd_pcm_sw_params_current(pcm_handle, swparams));
    CHECK_ALSA(snd_pcm_sw_params_set_avail_min(
                   pcm_handle, swparams, period_size));
    CHECK_ALSA(snd_pcm_sw_params(pcm_handle, swparams));

    /* Set up the software mixer. */

    int retval = sound_mixer_init(num_channels, SOUND_RATE);
    if (!retval) {
        DLOG("Mixer initialization failed");
        goto error_close_device;
    }
    for (int i = 0; i < num_channels; i++) {
        channels_ret[i] = i+1;
    }

    /* Start playback. */

    playback_thread_stop = 0;
    playback_thread_id = thread_create_with_priority(
        PLAYBACK_THREAD_PRIORITY, playback_thread, NULL);
    if (!playback_thread_id) {
        DLOG("Failed to create playback thread for mixer output: %s",
             sys_last_errstr());
        goto error_close_device_mixer;
    }

    /* All done. */

    samples_played = 0;
    silent_buffers_played = 0;
    sound_initted = 1;
    return num_channels;


  error_close_device_mixer:
    sound_mixer_cleanup();
  error_close_device:
    if (pcm_handle) {
        snd_pcm_close(pcm_handle);
        pcm_handle = NULL;
    }
  error_return:
    return 0;

    #undef CHECK_ALSA
}

/*-----------------------------------------------------------------------*/

int sys_sound_playback_rate(void)
{
    return SOUND_RATE;
}

/*-----------------------------------------------------------------------*/

int sys_sound_check_headphone_disconnect(void)
{
    return 0;  // Not supported.
}

/*-----------------------------------------------------------------------*/

void sys_sound_acknowledge_headphone_disconnect(void)
{
    /* Not supported. */
}

/*-----------------------------------------------------------------------*/

void sys_sound_cleanup(void)
{
    if (sound_initted) {
        playback_thread_stop = 1;
        thread_wait(playback_thread_id);
        playback_thread_id = 0;

        sound_mixer_cleanup();
        snd_pcm_close(pcm_handle);
        pcm_handle = NULL;

        sound_initted = 0;
    }
}

/*-----------------------------------------------------------------------*/

/*
 * The remaining sys_sound_*() functions are all redirected to the software
 * mixer functions, which implement the actual channel management.
 */

int sys_sound_setdata(int channel, SoundDecodeCallback decode_func,
                      void *handle, int is_stereo, UNUSED int sample_rate)
{
    return sound_mixer_setdata(channel, decode_func, handle, is_stereo);
}

void sys_sound_setvol(int channel, float volume)
{
    return sound_mixer_setvol(channel, volume);
}

void sys_sound_setpan(int channel, float pan)
{
    return sound_mixer_setpan(channel, pan);
}

void sys_sound_setfade(int channel, float target, float length, int cut)
{
    return sound_mixer_setfade(channel, target, length, cut);
}

void sys_sound_start(int channel)
{
    return sound_mixer_start(channel);
}

void sys_sound_stop(int channel)
{
    return sound_mixer_stop(channel);
}

void sys_sound_reset(int channel)
{
    return sound_mixer_reset(channel);
}

int sys_sound_status(int channel)
{
    return sound_mixer_status(channel);
}

/*************************************************************************/
/**************************** Local routines *****************************/
/*************************************************************************/

static int playback_thread(UNUSED void *userdata)
{
    while (!playback_thread_stop)
    {
        /* Wait for the device to be ready to receive data. */
        {
            int result;
            do {
                result = snd_pcm_wait(pcm_handle, -1);
            } while (result < 0 && snd_pcm_recover(pcm_handle, result, 1) == 0);
            if (result < 0) {
                DLOG("Failed waiting for audio device: %s",
                     snd_strerror(errno));
                /* Wait a little while just so we don't spend lots of CPU
                 * time spinning on failing write calls. */
                nanosleep(&(struct timespec){
                              .tv_sec = 0,
                              .tv_nsec = (1000000000/SOUND_BUFLEN) / 2},
                          NULL);
            }
        }

        /* Get some data from the mixer. */
        static int16_t pcm_buffer[SOUND_BUFLEN*2];
        sound_mixer_get_pcm(pcm_buffer, SOUND_BUFLEN);

        /* Work around an ALSA bug on 32-bit systems. */
        if (sizeof(long) == 4) {
            /*
             * ALSA keeps an internal counter of samples played, which
             * wraps around at 2^(long_bits-2) -- on 32-bit systems, this
             * is 0x40000000, or 6+ hours at 48kHz.  Due to ALSA bug 5190,
             * the library does not compute the available buffer space
             * properly when the counter is close to the wraparound point;
             * depending on the ratio of playback to hardware sample rate
             * or of internal buffer sizes, this can cause the audio stream
             * to start skipping or looping on a single hardware buffer,
             * and the condition will not be resolved until a reset
             * operation is performed on the device.
             *
             * To work around this, SIL will (when running on a 32-bit
             * system) monitor the output audio stream for periods of
             * extended silence.  When we find such a period, we trigger a
             * reset on the audio device, under the assumption the silence
             * will continue for a while longer; this allows us to reset
             * ALSA's internal counters without any audible impact.  If we
             * get close to 2^30 samples without seeing such a period of
             * silence, we likewise trigger a reset, accepting the
             * momentary audio glitch in order to prevent extended problems
             * later on.
             *
             * Of course, this problem can theoretically happen on 64-bit
             * systems as well, but since it would take about three million
             * years to occur, we don't worry about it.  If this decision
             * causes difficulties in Y3M, I'll be happy to take the blame.
             */
            int is_silent = 1;
            const uint32_t *ptr32 = (const uint32_t *)(void *)pcm_buffer;
            for (unsigned int i = 0; i < SOUND_BUFLEN; i++) {
                if (ptr32[i]) {
                    is_silent = 0;
                    break;
                }
            }
            int do_reset = 0;
            if (is_silent) {
                silent_buffers_played++;
                /* Reset after 1 second of silence. */
                do_reset =
                    (silent_buffers_played * SOUND_BUFLEN >= SOUND_RATE);
            } else {
                silent_buffers_played = 0;
            }
            samples_played += SOUND_BUFLEN;
            if (samples_played >= 0x3FF00000 && !do_reset) {
                DLOG("Close to wraparound, resetting audio for ALSA bug 5190");
                do_reset = 1;
            }
            if (do_reset) {
                int result;
                if ((result = snd_pcm_drain(pcm_handle)) < 0) {
                    DLOG("snd_pcm_drain() failed: %s", snd_strerror(result));
                }
                if ((result = snd_pcm_prepare(pcm_handle)) < 0) {
                    DLOG("snd_pcm_prepare() failed: %s", snd_strerror(result));
                }
                silent_buffers_played = 0;
                samples_played = 0;
            }
        }

        /* Write the data to the device. */
        unsigned int count = SOUND_BUFLEN;
        const uint8_t *ptr = (const uint8_t *)pcm_buffer;
        while (count > 0) {
            unsigned int this_count =
                (unsigned int)ubound(count, hw_buffer_size);
            ssize_t out;
            while ((out = snd_pcm_writei(pcm_handle, ptr, this_count)) < 0) {
                int result = snd_pcm_recover(pcm_handle, out, 1);
                if (result < 0) {
                    DLOG("snd_pcm_writei(): %s", snd_strerror(result));
                    out = 0;
                    count = 0;
                    break;
                }
            }
            ptr += out*4;
            count -= out;
        }

        BARRIER();
    }  // while (!playback_thread_stop)

    return 0;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
