/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/macosx/filedata.m: Implementation of the SILFileData URL
 * protocol.
 */

#import "src/base.h"
#import "src/memory.h"
#import "src/sysdep.h"
#import "src/sysdep/macosx/filedata.h"

#import <Foundation/NSData.h>
#import <Foundation/NSError.h>
#import <Foundation/NSURL.h>
#import <Foundation/NSURLError.h>
#import <Foundation/NSURLProtocol.h>
#import <Foundation/NSURLRequest.h>
#import <Foundation/NSURLResponse.h>

#import <errno.h>

/*************************************************************************/
/*********************** SILFileDataProtocol class ***********************/
/*************************************************************************/

@interface SILFileDataProtocol: NSURLProtocol
#if defined(__GNUC__) && !defined(__clang__)
{  // These are private, but GCC complains if they're not in the interface.
    SysFile *file;
    int64_t start, length;
    int cancel;
}
#endif
@end

@implementation SILFileDataProtocol {
    /* File handle from which to read. */
    SysFile *file;
    /* File offset at which the data starts. */
    int64_t start;
    /* Length of the data, in bytes. */
    int64_t length;
    /* Set to true if a -[stopLoading] call is received. */
    int cancel;
}

/*-----------------------------------------------------------------------*/

/* Quick helper to portably parse a 64-bit value. */
static inline int64_t strtoi64(const char *s, char **end, int base)
{
    return sizeof(long)>=8
        ? (int64_t)strtol(s, end, base)
        : (int64_t)strtoll(s, end, base);
}

/**
 * parseURL:  Internal method which parses a URL's path into its pointer,
 * start, and length parts.
 *
 * [Parameters]
 *     url: URL whose path is to be parsed.
 *     file_ret: Pointer to variable to receive SysFile pointer.
 *     start_ret: Pointer to variable to receive data start position.
 *     length_ret: Pointer to variable to receive data length.
 * [Return value]
 *     True if the path was successfully parsed, false if not.
 */
+ (BOOL)parseURL:(NSURL *)url file_ret:(SysFile **)file_ret
      start_ret:(int64_t *)start_ret length_ret:(int64_t *)length_ret
{
    PRECOND(url != NULL, return FALSE);
    PRECOND(file_ret != NULL, return FALSE);
    PRECOND(start_ret != NULL, return FALSE);
    PRECOND(length_ret != NULL, return FALSE);

    char path[100];
    if (![[url path] getCString:path maxLength:sizeof(path)
                     encoding:NSASCIIStringEncoding]) {
        return FALSE;
    }
    char *s = path;
    errno = 0;

    if (*s++ != '/') {
        return FALSE;
    }
    ASSERT(sizeof(SysFile *) <= sizeof(long), return FALSE);
    *file_ret = (SysFile *)strtoul(s, &s, 16);
    if (errno == ERANGE) {
        return FALSE;
    }

    if (*s++ != '/') {
        return FALSE;
    }
    *start_ret = strtoi64(s, &s, 10);
    if (errno == ERANGE || *start_ret < 0) {
        return NO;
    }

    if (*s++ != '/') {
        return FALSE;
    }
    *length_ret = strtoi64(s, &s, 10);
    if (errno == ERANGE || *length_ret < 0) {
        return TRUE;
    }

    if (*s != '\0') {
        return FALSE;
    }

    return TRUE;
}

/*-----------------------------------------------------------------------*/

/**
 * canInitWithRequest:  NSURLProtocol interface method called to check
 * whether this class can process the given request.
 */
+ (BOOL)canInitWithRequest:(NSURLRequest *)request
{
    NSString *scheme = [[request URL] scheme];
    if ([scheme caseInsensitiveCompare:@"silfiledata"] != NSOrderedSame) {
        return NO;
    }

    NSString *host = [[request URL] host];
    if ([host compare:@""] != NSOrderedSame) {
        return NO;
    }

    SysFile *dummy_file;
    int64_t dummy_start, dummy_length;
    return [self parseURL:[request URL] file_ret:&dummy_file
                 start_ret:&dummy_start length_ret:&dummy_length];
}

/*-----------------------------------------------------------------------*/

/**
 * canonicalRequestForRequest:  NSURLProtocol interface method called to
 * canonicalize a request so it can be used as a cache key.  Not relevant
 * for us, so we just return the incoming request as is.
 */
+ (NSURLRequest *)canonicalRequestForRequest:(NSURLRequest *)request
{
    return request;
}

/*-----------------------------------------------------------------------*/

/**
 * startLoading:  NSURLProtocol interface method called to start loading a
 * URL.  We perform the entire load here, passing it off one block at a
 * time to the request client.
 */
- (void)startLoading
{
    NSURLRequest *request = [self request];
    id<NSURLProtocolClient> client = [self client];

    ASSERT([[self class] parseURL:[request URL] file_ret:&self->file
                         start_ret:&self->start length_ret:&self->length],
           goto send_error);
    cancel = 0;

    NSURLResponse *response = [[NSURLResponse alloc]
                                  initWithURL:[request URL]
                                  MIMEType:@"application/octet-stream"
                                  expectedContentLength:self->length
                                  textEncodingName:nil];
    [client URLProtocol:self didReceiveResponse:response
            cacheStoragePolicy:NSURLCacheStorageNotAllowed];
    [response release];

    const int block_size = 4096;  // Arbitrary block size.
    sys_file_seek(self->file, self->start, FILE_SEEK_SET);
    int64_t bytes_left = length;
    while (!cancel && bytes_left > 0) {
        const int read_size = ubound(bytes_left, block_size);
        /* Note that the client might not consume the data before it returns
         * from -[URLProtocol:didLoadData:], so we need to allocate a new
         * buffer for each block we read. */
        NSMutableData *data = [[NSMutableData alloc] initWithLength:read_size];
        if (UNLIKELY(!data)) {
            DLOG("Failed to allocate data buffer with %d bytes", read_size);
            goto send_error;
        }
        const int nread =
            sys_file_read(self->file, [data mutableBytes], read_size);
        if (UNLIKELY(nread != read_size)) {
            if (nread < 0) {
                DLOG("Read error: %s", sys_last_errstr());
            } else {
                DLOG("Short read: only got %d of %d bytes", nread, read_size);
            }
            [data release];
            goto send_error;
        }
        [client URLProtocol:self didLoadData:data];
        [data release];
        bytes_left -= nread;
    }

    if (!cancel) {
        [client URLProtocolDidFinishLoading:self];
    }
    return;

  send_error:;
    NSError *error = [[NSError alloc] initWithDomain:NSURLErrorDomain
                                      code:NSURLErrorResourceUnavailable
                                      userInfo:nil];
    [client URLProtocol:self didFailWithError:error];
    [error release];
}

/*-----------------------------------------------------------------------*/

/**
 * stopLoading:  NSURLProtocol interface method called to stop loading a URL.
 */
- (void)stopLoading
{
    self->cancel = 1;
}

/*-----------------------------------------------------------------------*/

@end  // SILFileDataProtocol

/*************************************************************************/
/******************* Registration function (exported) ********************/
/*************************************************************************/

int osx_filedata_register(void)
{
    return [NSURLProtocol registerClass:[SILFileDataProtocol class]];
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
