/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/macosx/util.c: Miscellaneous utility functions specific to
 * Mac OS X.
 */

/*
 * Note that we use the standard library's memory management functions
 * (malloc() and friends) rather than mem_*() in macosx_get_application_name()
 * below, since it may be called before the memory management subsystem has
 * been initialized.
 */
#undef SIL_MEMORY_FORBID_MALLOC

#include "src/base.h"
#include "src/sysdep/macosx/internal.h"

#import <AppKit/AppKit.h>

/*************************************************************************/
/*********************** Exported utility routines ***********************/
/*************************************************************************/

/*
 * Gestalt() was deprecated in OS X 10.8, but according to one Apple
 * engineer (see https://twitter.com/Catfish_Man/status/373277120408997889),
 * the recommended way as of August 2013 to retrieve the OSX version is to
 * call Gestalt() and use a #pragma to suppress the deprecation warning:
 *
 * <Catfish_Man> just use gestalt. It's fine for that key
 * <Catfish_Man> use #pragma to turn that [compiler warning] off for that
 * one spot. I'm completely serious.
 *
 * Bug 15543376 has been filed with Apple to request a non-deprecated
 * replacement for this usage of Gestalt().
 */

#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdeprecated-declarations"

int macosx_version_major(void)
{
    static SInt32 version = -1;
    if (version == -1) {
        OSErr error = Gestalt(gestaltSystemVersionMajor, &version);
        if (error != 0) {
            DLOG("Failed to get version: %d", error);
            return 0;
        }
        ASSERT(version != -1, version = 0);
    }
    return version;
}

int macosx_version_minor(void)
{
    static SInt32 version = -1;
    if (version == -1) {
        OSErr error = Gestalt(gestaltSystemVersionMinor, &version);
        if (error != 0) {
            DLOG("Failed to get version: %d", error);
            return 0;
        }
        ASSERT(version != -1, version = 0);
    }
    return version;
}

int macosx_version_bugfix(void)
{
    static SInt32 version = -1;
    if (version == -1) {
        OSErr error = Gestalt(gestaltSystemVersionBugFix, &version);
        if (error != 0) {
            DLOG("Failed to get version: %d", error);
            return 0;
        }
        ASSERT(version != -1, version = 0);
    }
    return version;
}

#pragma clang diagnostic pop

/*-----------------------------------------------------------------------*/

int macosx_version_is_at_least(int major, int minor, int bugfix)
{
    const int os_major  = macosx_version_major();
    const int os_minor  = macosx_version_minor();
    const int os_bugfix = macosx_version_bugfix();
    return (os_major > major
            || (os_major == major
                && (os_minor > minor
                    || (os_minor == minor
                        && os_bugfix >= bugfix))));
}

/*************************************************************************/
/*********************** Internal utility routines ***********************/
/*************************************************************************/

const char *macosx_get_application_name(void)
{
    static char *saved_name;

    /* If we already looked it up, return the looked-up value. */
    if (saved_name) {
        return saved_name;
    }

    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

    const NSDictionary *dict =
        (const NSDictionary *)CFBundleGetInfoDictionary(CFBundleGetMainBundle());
    if (dict) {
        NSString *name_string = [dict objectForKey: @"CFBundleName"];
        if (![name_string length]) {
            name_string = [[NSProcessInfo processInfo] processName];
        }
        const char *name = [name_string UTF8String];
        if (name) {
            saved_name = strdup(name);
        }
    }

    [pool release];

    return saved_name ? saved_name : "The application";
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
