/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/openal/openal.c: OpenAL audio output interface.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/math.h"
#include "src/memory.h"
#include "src/sound.h"
#include "src/sysdep.h"
#include "src/sysdep/openal/openal.h"
#include "src/thread.h"
#include "src/time.h"

#if defined(SIL_PLATFORM_MACOSX) || defined(SIL_PLATFORM_IOS)
# include <OpenAL/al.h>
# include <OpenAL/alc.h>
#else
# include <AL/al.h>
# include <AL/alc.h>
#endif

/*************************************************************************/
/************************* Configuration options *************************/
/*************************************************************************/

/**
 * SIL_OPENAL_SINGLE_THREAD:  If defined, channel processing will be
 * performed by a single thread for all channels, rather than one thread
 * per channel.  This can improve performance on single-CPU / single-core
 * systems with expensive thread switches.
 */
// #define SIL_OPENAL_SINGLE_THREAD

/**
 * SIL_OPENAL_THREAD_PRIORITY:  Relative thread priority for playback
 * processing threads, passed to thread_create_with_priority().  Defaults
 * to zero (same priority as the main thread) if not defined.
 */
#ifndef SIL_OPENAL_THREAD_PRIORITY
# define SIL_OPENAL_THREAD_PRIORITY  0
#endif

/**
 * SIL_OPENAL_USE_INTERNAL_MIXER:  If defined, the internal software mixer
 * will be used to mix PCM data, and the mixed data will be continuously
 * fed to a single stereo OpenAL source.  Typically only useful for testing
 * the mixer, or on systems such as Windows where the default OpenAL
 * library is broken.
 */
// #define SIL_OPENAL_USE_INTERNAL_MIXER

/**
 * SIL_OPENAL_INTERNAL_MIXER_RATE:  If SIL_OPENAL_USE_INTERNAL_MIXER is
 * defined, this sets the sampling rate at which mixing will be performed
 * (mandatory).
 */
// #define SIL_OPENAL_INTERNAL_MIXER_RATE  48000

/*-----------------------------------------------------------------------*/

/* Options that generally don't need to be changed: */

/**
 * SOUND_BUFFERS:  The number of playback buffers to allocate per channel.
 */
#define SOUND_BUFFERS  4

/**
 * SOUND_BUFLEN:  The length of a single playback buffer, in samples.
 */
#define SOUND_BUFLEN  1024

/**
 * MAX_VOLUME:  The maximum channel volume to support.  (This is required
 * because OpenAL seems to limit source volumes to the range [0,1], so we
 * need to scale on the listener side.)
 */
#define MAX_VOLUME  1024.0f  // Use a power of 2 to avoid losing precision.

/*************************************************************************/

#ifdef SIL_OPENAL_USE_INTERNAL_MIXER
# include "src/sound/mixer.h"
#endif

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Have we been initialized? */
static int sound_initted;

/* OpenAL device and context. */
static ALCdevice *device;
static ALCcontext *context;

/*-----------------------------------------------------------------------*/

/* Number of channels allocated. */
static int num_channels;

/* Data for each sound channel. */
typedef struct Channel Channel;
struct Channel {
    ALuint source_id;       // OpenAL source ID

    int thread;             // Decode thread for this channel
    uint8_t stop_thread;    // True = thread should stop

    SoundDecodeCallback decode_func;
    void *decode_handle;
    uint32_t sample_rate;   // Sampling rate of data from decode function
    uint8_t stereo;         // True = stereo, false = monaural
    uint8_t playing;        // Flag: Is the channel playing?
    uint8_t out_of_data;    // Flag: Has the channel hit the end of the data?
    uint8_t fade_cut;       // True = stop playback when fade volume hits zero
    float volume;           // Current volume (0...MAX_VOLUME)
    float pan;              // Current pan value (-1...+1)
    int32_t stereo_pan_l;   // Pan multipler for left stereo channel
                            //    (0x10000 = full volume)
    int32_t stereo_pan_r;   // Pan multipler for right stereo channel
    float fade_rate;        // Amount to add to volume per sample played
    float fade_target;      // Fade target volume

    /* OpenAL buffers allocated for the currently playing sound. */
    ALuint buffers[SOUND_BUFFERS];
    /* In-use (playing) flags for each buffer. */
    uint8_t buf_used[SOUND_BUFFERS];
    /* Queue age counters for each buffer.  Whenever a new buffer is
     * queued, the age counters for all other buffers are incremented; if
     * the age of a buffer gets too high (see logic in process_channel()),
     * the buffer is assumed to have been lost by OpenAL and is reused. */
    uint8_t buf_age[SOUND_BUFFERS];

    /* Temporary buffer for decoding audio data. */
    int16_t pcm_buffer[2 * SOUND_BUFLEN];
};
static Channel *channels;

#ifdef SIL_OPENAL_SINGLE_THREAD
/* Master playback thread, and dummy Channel structure passed to it. */
int master_thread;
static Channel master_thread_dummy_channel;
#endif

/*-----------------------------------------------------------------------*/

/* Convenience macros for error-checking OpenAL calls.  ALWRAP() calls an
 * OpenAL function and returns its error code; ALCHECK() calls a function
 * and logs a message if an error occurs, returning the success or failure
 * of the call as a boolean value.  Both macros take a function call
 * expression as their parameter. */

#define ALWRAP(call) ( \
    alGetError(),  /* Clear error state from any previous calls. */     \
    (call),        /* Call the requested function. */                   \
    alGetError()   /* Retrieve and return the error code. */            \
)

#define ALCHECK(call)  (check_error(ALWRAP(call), #call))

/*-----------------------------------------------------------------------*/

/* Local function declarations. */

/**
 * check_error:  If the given error code indicates an error, log a debug
 * message and return false; otherwise return true.
 *
 * [Parameters]
 *     error_code: OpenAL error code.
 *     call_text: Text describing the OpenAL call (for debug messages).
 * [Return value]
 *     True if the error code was AL_NO_ERROR, indicating a successful
 *     call; false otherwise.
 */
static int check_error(int error_code, const char *call_text);

/**
 * playback_thread:  Playback thread for a single channel (or for all
 * channels, if SIL_OPENAL_SINGLE_THREAD is defined).
 *
 * [Parameters]
 *     channel: Channel data structure pointer.
 */
static void playback_thread(Channel *channel);

/**
 * process_channel:  Handle periodic processing for the given channel.
 * Called from the playback thread.
 *
 * [Parameters]
 *     channel: Channel data structure pointer.
 * [Return value]
 *     True if any processing was performed on the channel, false if not.
 */
static inline int process_channel(Channel *channel);

/**
 * queue_buffer:  Decode and queue a buffer of PCM data to be played on
 * the given channel.
 *
 * [Parameters]
 *     channel: Channel data structure pointer.
 *     buffer: OpenAL buffer ID to queue.
 * [Return value]
 *     True on success; false on error, end of stream, or fade cut.
 */
static int queue_buffer(Channel *channel, ALuint buffer);

#ifdef SIL_OPENAL_USE_INTERNAL_MIXER
/**
 * mix_callback:  Callback function for passing mixer output to the system.
 *
 * [Parameters]
 *     handle: Audio data handle (unused).
 *     pcm_buffer: Output buffer for PCM samples.
 *     pcm_len: Number of samples to retrieve.
 * [Return value]
 *     Always true.
 */
static int mix_callback(void *handle, void *pcm_buffer, int pcm_len);
#endif

/*************************************************************************/
/***************** Interface: General control functions ******************/
/*************************************************************************/

int openal_init(const char *device_name, int num_channels_, int *channels_ret)
{
    PRECOND(device_name != NULL, goto error_return);
    PRECOND(num_channels_ > 0, goto error_return);
    PRECOND(channels_ret != NULL, goto error_return);
    if (sound_initted) {
        DLOG("Tried to init twice!");
        goto error_return;
    }

#ifdef SIL_OPENAL_USE_INTERNAL_MIXER
    const int mixer_channels = num_channels_;
    if (!sound_mixer_init(mixer_channels, SIL_OPENAL_INTERNAL_MIXER_RATE)) {
        DLOG("Failed to init software mixer");
        goto error_return;
    }
    num_channels = 1;
#else
    num_channels = num_channels_;
#endif

    device = alcOpenDevice(*device_name ? device_name : NULL);
    if (!device) {
        DLOG("Failed to open %s%s",
             *device_name ? "device " : "default device",
             *device_name ? device_name : "");
#ifdef SIL_OPENAL_USE_INTERNAL_MIXER
        goto error_close_mixer;
#else
        goto error_return;
#endif
    }

    /* We don't make a distinction between mono and stereo sources, so
     * be safe and allocate a full contingent of each type of channel. */
    ALCint requested_attributes[5];
    int attr_index = 0;
#ifndef SIL_OPENAL_USE_INTERNAL_MIXER
    requested_attributes[attr_index++] = ALC_MONO_SOURCES;
    requested_attributes[attr_index++] = num_channels;
#endif
    requested_attributes[attr_index++] = ALC_STEREO_SOURCES;
    requested_attributes[attr_index++] = num_channels;
    requested_attributes[attr_index++] = 0;
    context = alcCreateContext(device, requested_attributes);
    if (!context) {
        DLOG("Failed to create OpenAL context");
        goto error_close_device;
    }
    if (!alcMakeContextCurrent(context)) {
        DLOG("Failed to select OpenAL context");
        goto error_destroy_context;
    }

    ALCint num_attributes = 0;
    alcGetIntegerv(device, ALC_ATTRIBUTES_SIZE, 1, &num_attributes);
    if (num_attributes > 0) {
        ALCint *attributes = mem_alloc(sizeof(*attributes) * num_attributes, 0,
                                       MEM_ALLOC_TEMP | MEM_ALLOC_CLEAR);
        if (UNLIKELY(!attributes)) {
            DLOG("Failed to allocate an attribute buffer (%d bytes)",
                 (int)(num_attributes * sizeof(ALCint)));
            goto error_destroy_context;
        }
        alcGetIntegerv(device, ALC_ALL_ATTRIBUTES, num_attributes, attributes);
        for (int i = 0; i < num_attributes; i += 2) {
            if (attributes[i] == ALC_STEREO_SOURCES) {
                if (attributes[i+1] <= 0) {
                    DLOG("Couldn't get any stereo sources, aborting");
                    goto error_destroy_context;
                } else if (attributes[i+1] < num_channels) {
                    DLOG("Only got %d of %d stereo sources",
                         attributes[i+1], num_channels);
                    num_channels = attributes[i+1];
                }
#ifndef SIL_OPENAL_USE_INTERNAL_MIXER
# ifndef SIL_PLATFORM_MACOSX  // The OSX OpenAL library is broken and reports
                               // 0 mono sources even though it provides them.
            } else if (attributes[i] == ALC_MONO_SOURCES) {
                if (attributes[i+1] <= 0) {
                    DLOG("Couldn't get any mono sources, aborting");
                    goto error_destroy_context;
                } else if (attributes[i+1] < num_channels) {
                    DLOG("Only got %d of %d mono sources",
                         attributes[i+1], num_channels);
                    num_channels = attributes[i+1];
                }
# endif
#endif
            }
        }
        mem_free(attributes);
    } else {
        DLOG("WARNING: couldn't get device attributes!");
    }

    /* As noted in the MAX_VOLUME documentation, we scale volumes up here
     * and down on the source side because OpenAL doesn't seem to allow
     * channel (source) volumes greater than 1. */
    ALCHECK(alListenerf(AL_GAIN, MAX_VOLUME));
    /* The (rather sparse) OpenAL documentation doesn't seem to specify
     * the default values for listener properties, so set the rest of
     * them explicitly. */
    ALCHECK(alListenerfv(AL_POSITION,    (ALfloat[]){0,0,0}));
    ALCHECK(alListenerfv(AL_VELOCITY,    (ALfloat[]){0,0,0}));
    ALCHECK(alListenerfv(AL_ORIENTATION, (ALfloat[]){0,0,1, 0,1,0}));

    /* We don't make use of distance (only angle, for panning). */
    ALCHECK(alDistanceModel(AL_NONE));

    channels = mem_alloc(sizeof(*channels) * num_channels, 0, 0);
    if (!channels) {
        DLOG("No memory for %d channel structures (size %u)", num_channels,
             (int)(sizeof(*channels) * num_channels));
        goto error_destroy_context;
    }

    for (int i = 0; i < num_channels; i++) {
        channels[i].decode_func   = NULL;
        channels[i].decode_handle = NULL;
        channels[i].stereo        = 0;
        channels[i].playing       = 0;
        channels[i].fade_cut      = 0;
        channels[i].volume        = 1;
        channels[i].pan           = 0;
        channels[i].fade_rate     = 0;

        if (!ALCHECK(alGenSources(1, &channels[i].source_id))) {
            DLOG("Failed to create source for channel %d", i);
            num_channels = i;
            break;
        }

        if (!ALCHECK(alGenBuffers(lenof(channels[i].buffers),
                                  channels[i].buffers))) {
            DLOG("Failed to create buffers for channel %d", i);
            alDeleteSources(1, &channels[i].source_id);
            num_channels = i;
            break;
        }

        channels_ret[i] = i;
        const ALuint source_id = channels[i].source_id;
        /* Make sure the volume defaults to a sane value (just in case). */
        ALCHECK(alSourcef (source_id, AL_GAIN,    1/MAX_VOLUME));
        /* Again, set the other parameters explicitly to be safe. */
        ALCHECK(alSourcef (source_id, AL_PITCH,              1));
        ALCHECK(alSourcef (source_id, AL_MIN_GAIN,           0));
        ALCHECK(alSourcef (source_id, AL_MAX_GAIN,           1));
        ALCHECK(alSourcef (source_id, AL_CONE_OUTER_GAIN,    1));
        ALCHECK(alSourcef (source_id, AL_CONE_INNER_ANGLE, 360));
        ALCHECK(alSourcef (source_id, AL_CONE_OUTER_ANGLE, 360));
        ALCHECK(alSourcef (source_id, AL_ROLLOFF_FACTOR,     1));
        ALCHECK(alSourcefv(source_id, AL_POSITION,  (ALfloat[]){0,0,0}));
        ALCHECK(alSourcefv(source_id, AL_VELOCITY,  (ALfloat[]){0,0,0}));
        ALCHECK(alSourcefv(source_id, AL_DIRECTION, (ALfloat[]){0,0,0}));
    }

    if (num_channels == 0) {
        DLOG("Couldn't allocate any channels!  Aborting.");
        goto error_destroy_context;
    }

#ifdef SIL_OPENAL_USE_INTERNAL_MIXER
    /* sys_sound_setdata() will redirect to sound_mixer_setdata(), so set
     * up the channel structure ourselves. */
    channels[0].decode_func   = mix_callback;
    channels[0].decode_handle = NULL;
    channels[0].sample_rate   = SIL_OPENAL_INTERNAL_MIXER_RATE;
    channels[0].stereo        = 1;
    channels[0].out_of_data   = 0;
    for (int i = 0; i < lenof(channels[0].buffers); i++) {
        channels[0].buf_used[i] = 0;
        channels[0].buf_age[i] = 0;
    }
    channels[0].stop_thread   = 0;
    channels[0].playing       = 1;
# ifdef SIL_OPENAL_SINGLE_THREAD
    channels[0].thread = -1;  // Arbitrary nonzero value.
# else
    channels[0].thread = thread_create_with_priority(
        SIL_OPENAL_THREAD_PRIORITY,
        (int (*)(void *))playback_thread, &channels[0]);
    if (!channels[0].thread) {
        DLOG("Failed to create playback thread for mixer output: %s",
             sys_last_errstr());
        channels[0].decode_func = NULL;
        goto error_destroy_context;
    }
# endif
#endif

#ifdef SIL_OPENAL_SINGLE_THREAD
    master_thread_dummy_channel.stop_thread = 0;
    master_thread = thread_create_with_priority(
        SIL_OPENAL_THREAD_PRIORITY,
        (int (*)(void *))playback_thread, &master_thread_dummy_channel);
    if (!master_thread) {
        DLOG("Failed to create playback thread: %s", sys_last_errstr());
        goto error_destroy_context;
    }
#endif

    alcProcessContext(context);

    sound_initted = 1;

#ifdef SIL_OPENAL_USE_INTERNAL_MIXER
    for (int i = 0; i < mixer_channels; i++) {
        channels_ret[i] = i+1;
    }
    return mixer_channels;
#else
    return num_channels;
#endif


  error_destroy_context:
    alcMakeContextCurrent(NULL);
    alcDestroyContext(context);
    context = NULL;
  error_close_device:
    alcCloseDevice(device);
    device = NULL;
#ifdef SIL_OPENAL_USE_INTERNAL_MIXER
  error_close_mixer:
    sound_mixer_cleanup();
#endif
  error_return:
    return 0;
}

/*-----------------------------------------------------------------------*/

int sys_sound_playback_rate(void)
{
#ifdef SIL_OPENAL_USE_INTERNAL_MIXER
    return SIL_OPENAL_INTERNAL_MIXER_RATE;
#else
    return 0;  // OpenAL will resample for us.
#endif
}

/*-----------------------------------------------------------------------*/

void openal_cleanup(void)
{
    if (!sound_initted) {
        return;
    }

#ifdef SIL_OPENAL_USE_INTERNAL_MIXER
# ifndef SIL_OPENAL_SINGLE_THREAD
    channels[0].stop_thread = 1;
    thread_wait(channels[0].thread);
    alSourceStop(channels[0].source_id);
# endif
#else
    for (int channel = 0; channel < num_channels; channel++) {
        if (channels[channel].decode_func) {
            sys_sound_reset(channel);
        }
    }
#endif

#ifdef SIL_OPENAL_SINGLE_THREAD
    master_thread_dummy_channel.stop_thread = 1;
    thread_wait(master_thread);
    master_thread = 0;
#endif

    if (UNLIKELY(!alcMakeContextCurrent(NULL))) {
        DLOG("Warning: alcMakeContextCurrent(NULL) failed");
    }
    alcDestroyContext(context);
    context = NULL;

    if (UNLIKELY(!alcCloseDevice(device))) {
        DLOG("Warning: alcCloseDevice() failed");
    }
    device = NULL;

    mem_free(channels);
    channels = NULL;
    num_channels = 0;

#ifdef SIL_OPENAL_USE_INTERNAL_MIXER
    sound_mixer_cleanup();
#endif

    sound_initted = 0;
}

/*-----------------------------------------------------------------------*/

void openal_suspend(void)
{
#ifdef SIL_PLATFORM_IOS
    /* Perhaps because the documentation of alcSuspendContext() and
     * alcProcessContext() isn't clear, iOS handles suspend/resume by
     * changing the current context.  See, e.g.:
     * http://opensource.creative.com/pipermail/openal/2009-July/011766.html */
    alcMakeContextCurrent(NULL);
#else
    alcSuspendContext(context);
#endif
}

/*-----------------------------------------------------------------------*/

void openal_resume(void)
{
#ifdef SIL_PLATFORM_IOS
    /* See note in openal_suspend(). */
    alcMakeContextCurrent(context);
#else
    alcProcessContext(context);
#endif
}

/*************************************************************************/
/*********** Interface: Sound playback/manipulation functions ************/
/*************************************************************************/

int sys_sound_setdata(int channel, SoundDecodeCallback decode_func,
                      void *handle, int is_stereo, int sample_rate)
{
#ifdef SIL_OPENAL_USE_INTERNAL_MIXER
    return sound_mixer_setdata(channel, decode_func, handle, is_stereo);
#endif

    if (UNLIKELY(channel < 0) || UNLIKELY(channel >= num_channels)
     || UNLIKELY(!decode_func)
     || UNLIKELY(sample_rate <= 0)) {
        DLOG("Invalid parameters: %d %p %p %d", channel, decode_func,
             handle, is_stereo);
        return 0;
    }
    if (channels[channel].decode_func) {
        DLOG("Channel %d is already in use", channel);
        return 0;
    }

    channels[channel].decode_func   = decode_func;
    channels[channel].decode_handle = handle;
    channels[channel].sample_rate   = sample_rate;
    channels[channel].stereo        = is_stereo ? 1 : 0;
    channels[channel].out_of_data   = 0;

    for (int i = 0; i < lenof(channels[channel].buffers); i++) {
        channels[channel].buf_used[i] = 0;
        channels[channel].buf_age[i] = 0;
    }

    /*
     * NOTE: The lack of ALCHECK() here and below is intentional -- since
     * alGetError() writes to a global buffer and is thus not thread-safe,
     * we can't use it except in cases like sys_sound_init() where we know
     * no other threads can be calling OpenAL functions.  In cases like
     * alSourceUnqueueBuffers() here, we rely on the specification that
     * failing OpenAL calls do not modify return parameters to detect
     * failure; for functions that don't return values, we have no choice
     * but to hope for the best.  If we tried to use ALCHECK() to check for
     * errors, we'd introduce a race condition where an error generated by
     * one thread could be picked up by another thread's alGetError() call,
     * causing both threads to act incorrectly.
     *
     * From a multithreading point of view, the insistence on using a
     * global error buffer instead of returning success/failure or error
     * codes from functions is a remarkably poor design decision.  Not that
     * a mere library user has much choice except to live with it...
     */

    alSourceStop(channels[channel].source_id);
    for (;;) {
        ALuint buffer = 0;
        alSourceUnqueueBuffers(channels[channel].source_id, 1, &buffer);
        if (buffer == 0) {
            break;
        }
    }

    channels[channel].stop_thread = 0;
#ifdef SIL_OPENAL_SINGLE_THREAD
    channels[channel].thread = -1;  // Signal that it's active.
    BARRIER();
#else
    channels[channel].thread = thread_create_with_priority(
        SIL_OPENAL_THREAD_PRIORITY,
        (int (*)(void *))playback_thread, &channels[channel]);
    if (!channels[channel].thread) {
        DLOG("Failed to create playback thread for channel %d: %s",
             channel, sys_last_errstr());
        channels[channel].decode_func = NULL;
        return 0;
    }
#endif

    return 1;
}

/*-----------------------------------------------------------------------*/

void sys_sound_setvol(int channel, float volume)
{
#ifdef SIL_OPENAL_USE_INTERNAL_MIXER
    return sound_mixer_setvol(channel, volume);
#endif

    if (UNLIKELY(channel < 0) || UNLIKELY(channel >= num_channels)) {
        DLOG("Invalid parameters: %d %g", channel, volume);
        return;
    }

    channels[channel].volume = bound(volume, 0, MAX_VOLUME);
    channels[channel].fade_rate = 0;
    channels[channel].fade_cut = 0;
    BARRIER();
    alSourcef(channels[channel].source_id, AL_GAIN, volume/MAX_VOLUME);
}

/*-----------------------------------------------------------------------*/

void sys_sound_setpan(int channel, float pan)
{
#ifdef SIL_OPENAL_USE_INTERNAL_MIXER
    return sound_mixer_setpan(channel, pan);
#endif

    if (UNLIKELY(channel < 0) || UNLIKELY(channel >= num_channels)) {
        DLOG("Invalid parameters: %d %g", channel, pan);
        return;
    }

    channels[channel].pan = bound(pan, -1, 1);

    const float left = 1 - channels[channel].pan;
    const float right = 1 + channels[channel].pan;
    if (left < right) {
        channels[channel].stereo_pan_l = (int32_t)roundf(0x10000 * left/right);
        channels[channel].stereo_pan_r = 0x10000;
    } else {
        channels[channel].stereo_pan_l = 0x10000;
        channels[channel].stereo_pan_r = (int32_t)roundf(0x10000 * right/left);
    }

    /* Channel position is documented to be ignored for stereo samples, so
     * we go ahead and set it unconditionally. */
    const float pan_angle = channels[channel].pan * -90;
    float position[3];
    dsincosf(pan_angle, &position[0], &position[2]);
    position[1] = 0;
    alSourcefv(channels[channel].source_id, AL_POSITION, position);
}

/*-----------------------------------------------------------------------*/

void sys_sound_setfade(int channel, float target, float length, int cut)
{
#ifdef SIL_OPENAL_USE_INTERNAL_MIXER
    return sound_mixer_setfade(channel, target, length, cut);
#endif

    if (UNLIKELY(channel < 0) || UNLIKELY(channel >= num_channels)
     || UNLIKELY(length < 0)) {
        DLOG("Invalid parameters: %d %g %g", channel, target, length);
        return;
    }
    if (!channels[channel].decode_func) {
        DLOG("Channel %d has no data", channel);
        return;
    }

    /* Disable any current fade first to avoid racing with the playback
     * thread. */
    channels[channel].fade_rate = 0;
    BARRIER();

    if (length == 0) {
        channels[channel].fade_cut = 0;
    } else {
        target = bound(target, 0, MAX_VOLUME);
        const float delta_volume = target - channels[channel].volume;
        if (delta_volume == 0) {
            if (cut) {
                channels[channel].playing = 0;
            }
        } else {
            const int rate = channels[channel].sample_rate;
            const uint32_t samples = lbound((uint32_t)roundf(length * rate), 1);
            channels[channel].fade_target = target;
            channels[channel].fade_cut = (cut != 0);
            BARRIER();  // Ensure fade_rate is set last, again to avoid racing.
            channels[channel].fade_rate = delta_volume / samples;
        }
    }
    BARRIER();
}

/*-----------------------------------------------------------------------*/

void sys_sound_start(int channel)
{
#ifdef SIL_OPENAL_USE_INTERNAL_MIXER
    return sound_mixer_start(channel);
#endif

    if (UNLIKELY(channel < 0) || UNLIKELY(channel >= num_channels)) {
        DLOG("Invalid parameters: %d", channel);
        return;
    }

    if (channels[channel].decode_func != NULL) {
        BARRIER();  // Make sure all pending updates go out first.
        channels[channel].playing = 1;
        BARRIER();
    }
}

/*-----------------------------------------------------------------------*/

void sys_sound_stop(int channel)
{
#ifdef SIL_OPENAL_USE_INTERNAL_MIXER
    return sound_mixer_stop(channel);
#endif

    if (UNLIKELY(channel < 0) || UNLIKELY(channel >= num_channels)) {
        DLOG("Invalid parameters: %d", channel);
        return;
    }

    BARRIER();
    channels[channel].playing = 0;
    BARRIER();
}

/*-----------------------------------------------------------------------*/

void sys_sound_reset(int channel)
{
#ifdef SIL_OPENAL_USE_INTERNAL_MIXER
    return sound_mixer_reset(channel);
#endif

    if (UNLIKELY(channel < 0) || UNLIKELY(channel >= num_channels)) {
        DLOG("Invalid parameters: %d", channel);
        return;
    }

    if (channels[channel].decode_func) {
        channels[channel].stop_thread = 1;
#ifdef SIL_OPENAL_SINGLE_THREAD
        BARRIER();
        while (channels[channel].thread != 0) {
            time_delay(0.0001);
        }
#else
        thread_wait(channels[channel].thread);
        channels[channel].thread = 0;
#endif

        alSourceStop(channels[channel].source_id);
        for (;;) {
            ALuint buffer = 0;
            alSourceUnqueueBuffers(channels[channel].source_id, 1, &buffer);
            if (buffer == 0) {
                break;
            }
        }

        channels[channel].decode_func   = NULL;
        channels[channel].decode_handle = NULL;
        channels[channel].stereo        = 0;
        channels[channel].fade_cut      = 0;
        channels[channel].volume        = 1;
        channels[channel].pan           = 0;
        channels[channel].fade_rate     = 0;
    }
}

/*-----------------------------------------------------------------------*/

int sys_sound_status(int channel)
{
#ifdef SIL_OPENAL_USE_INTERNAL_MIXER
    return sound_mixer_status(channel);
#endif

    if (UNLIKELY(channel < 0) || UNLIKELY(channel >= num_channels)) {
        DLOG("Invalid parameters: %d", channel);
        return 0;
    }

    return channels[channel].playing;
}

/*************************************************************************/
/**************************** Local routines *****************************/
/*************************************************************************/

static int check_error(int error_code, DEBUG_USED const char *call_text)
{
    if (error_code == AL_NO_ERROR) {

        return 1;

    } else {

        if (error_code == AL_INVALID_NAME) {
            DLOG("%s: Invalid ID", call_text);
        } else if (error_code == AL_INVALID_ENUM) {
            DLOG("%s: Invalid enumeration code", call_text);
        } else if (error_code == AL_INVALID_VALUE) {
            DLOG("%s: Invalid value", call_text);
        } else if (error_code == AL_INVALID_OPERATION) {
            DLOG("%s: Invalid operation", call_text);
        } else if (error_code == AL_OUT_OF_MEMORY) {
            DLOG("%s: Out of memory", call_text);
        } else {
            DLOG("%s: Unknown error 0x%X", call_text, error_code);
        }

        return 0;

    }
}

/*-----------------------------------------------------------------------*/

static void playback_thread(Channel *channel)
{
    while (!channel->stop_thread)
    {
        /* Wait a short period of time each iteration so we don't spin on
         * the CPU when there's nothing to do. */
        time_delay(0.001);

#ifdef SIL_OPENAL_SINGLE_THREAD

        int did_something;
        do {
            did_something = 0;
            for (int i = 0; i < num_channels; i++) {
                Channel *this_channel = &channels[i];
                if (this_channel->thread) {
                    if (this_channel->stop_thread) {
                        this_channel->thread = 0;
                        BARRIER();
                    } else {
                        if (process_channel(this_channel)) {
                            BARRIER();
                            did_something = 1;
                        }
                    }
                }
            }
        } while (did_something);

#else  // !SIL_OPENAL_SINGLE_THREAD

        if (process_channel(channel)) {
            BARRIER();
        }

#endif

    }  // while (!channel->stop_thread)
}

/*-----------------------------------------------------------------------*/

static inline int process_channel(Channel *channel)
{
    int retval = 0;

    /* If we're not playing, stop the OpenAL channel and wait. */
    if (!channel->playing) {
        alSourceStop(channel->source_id);
        return 0;
    }

    /* Pull off any completed buffers. */
    ALint processed = 0;
    alGetSourcei(channel->source_id, AL_BUFFERS_PROCESSED, &processed);
    for (int i = 0; i < processed; i++) {
        ALuint buffer = 0;
        alSourceUnqueueBuffers(channel->source_id, 1, &buffer);
        if (buffer) {
            for (int index = 0; index < lenof(channel->buffers); index++) {
                if (channel->buffers[index] == buffer) {
                    channel->buf_used[index] = 0;
                    channel->buf_age[index] = 0;
                    break;
                }
            }
            retval = 1;
        }
    }

    /* Check for buffers which have not been reused recently (whether due
     * to a local bug, an OpenAL implementation bug, or whatever), and
     * forcibly reset them to the unused state.  If playback proceeds
     * normally, every buffer should be reused in sequence, so the age of
     * a single buffer should never exceed the number of buffers available
     * (SOUND_BUFFERS); we check against 2*SOUND_BUFFERS just to provide a
     * bit of leeway for unforeseen circumstances. */
    for (int i = 0; i < lenof(channel->buffers); i++) {
        if (channel->buf_used[i] && channel->buf_age[i] > 2*SOUND_BUFFERS) {
            DLOG("WARNING: channel %u buffer %d is too old (%u > %u),"
                 " resetting...", (int)(channel - channels), i,
                 channel->buf_age[i], 2*SOUND_BUFFERS);
            ALuint new_buffer = 0;
            alGenBuffers(1, &new_buffer);
            if (new_buffer == 0) {
                DLOG("...Failed to generate new buffer!  Will reuse old buffer"
                     " %d (this may cause bugs).", channel->buffers[i]);
            } else {
                alDeleteBuffers(1, &channel->buffers[i]);
                channel->buffers[i] = new_buffer;
                DLOG("...Generated new buffer %d.", channel->buffers[i]);
            }
            channel->buf_used[i] = 0;
            channel->buf_age[i] = 0;
            retval = 1;
        }
    }

    /* Fill and queue any unused buffers. */
    if (!channel->out_of_data) {
        for (int i = 0; i < lenof(channel->buffers); i++) {
            if (!channel->buf_used[i]) {
                if (!queue_buffer(channel, channel->buffers[i])) {
                    channel->out_of_data = 1;
                    break;
                }
                channel->buf_used[i] = 1;
                channel->buf_age[i] = 0;
                for (int j = 0; j < lenof(channel->buf_age); j++) {
                    if (j != i && channel->buf_used[i]) {
                        channel->buf_age[j]++;
                    }
                }
                retval = 1;
            }
        }
    }

    /* Start the channel playing if it wasn't already (and if we didn't
     * hit the end of the stream). */
    ALenum state = 0;
    alGetSourcei(channel->source_id, AL_SOURCE_STATE, &state);
    if (state != AL_PLAYING) {
        ALint queued = 0;
        alGetSourcei(channel->source_id, AL_BUFFERS_QUEUED, &queued);
        if (queued == 0) {
            /* Work around a bug (race condition?) in which a sound started
             * immediately after a previous sound has stopped on the same
             * channel can end up with no buffers queued but all buffers full.
             * Does OpenAL dump queued buffers under some circumstances? */
            int all_full = 1;
            for (int i = 0; i < lenof(channel->buffers); i++) {
                if (!channel->buf_used[i]) {
                    all_full = 0;
                    break;
                }
            }
            if (all_full) {
                alSourceQueueBuffers(channel->source_id,
                                     lenof(channel->buffers),
                                     channel->buffers);
            } else {
                channel->playing = 0;
            }
        }
        if (channel->playing) {
            alSourcePlay(channel->source_id);
        }
        retval = 1;
    }

    return retval;
}

/*-----------------------------------------------------------------------*/

static int queue_buffer(Channel *channel, ALuint buffer)
{
    const ALenum format =
        (channel->stereo ? AL_FORMAT_STEREO16 : AL_FORMAT_MONO16);
    const int rate = channel->sample_rate;
    const int chans = (channel->stereo ? 2 : 1);
    const int pcm_size = SOUND_BUFLEN * chans * 2;

    if (channel->fade_rate != 0) {
        const float new_volume =
            channel->volume + channel->fade_rate * SOUND_BUFLEN;
        if ((channel->fade_rate < 0 && new_volume < channel->fade_target)
         || (channel->fade_rate > 0 && new_volume > channel->fade_target)) {
            channel->volume = channel->fade_target;
            channel->fade_rate = 0;
            if (channel->fade_cut) {
                return 0;
            }
        } else {
            channel->volume = new_volume;
        }
        alSourcef(channel->source_id, AL_GAIN, channel->volume / MAX_VOLUME);
    }

    if (!(*channel->decode_func)(channel->decode_handle,
                                 channel->pcm_buffer, SOUND_BUFLEN)) {
        return 0;
    }

    if (channel->stereo && channel->pan != 0) {
        const int32_t stereo_pan_l = channel->stereo_pan_l;
        const int32_t stereo_pan_r = channel->stereo_pan_r;
        int16_t *buffer16 = channel->pcm_buffer;
        for (int i = 0; i < SOUND_BUFLEN; i++) {
            buffer16[i*2+0] = (buffer16[i*2+0] * stereo_pan_l) >> 16;
            buffer16[i*2+1] = (buffer16[i*2+1] * stereo_pan_r) >> 16;
        }
    }

    alBufferData(buffer, format, channel->pcm_buffer, pcm_size, rate);
    alSourceQueueBuffers(channel->source_id, 1, &buffer);

    return 1;
}

/*-----------------------------------------------------------------------*/

#ifdef SIL_OPENAL_USE_INTERNAL_MIXER

static int mix_callback(UNUSED void *handle, void *pcm_buffer, int pcm_len)
{
    sound_mixer_get_pcm(pcm_buffer, pcm_len);
    return 1;
}

#endif

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
