/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/opengl/framebuffer.c: Framebuffer management functionality
 * for OpenGL-based ports.
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/opengl/internal.h"
#include "src/sysdep/opengl/opengl.h"

#ifdef SIL_PLATFORM_ANDROID
/* Workaround for https://code.google.com/p/android/issues/detail?id=62903 */
#include <dlfcn.h>
static __typeof__(glDiscardFramebufferEXT) *p_glDiscardFramebufferEXT;
#define glDiscardFramebufferEXT (*p_glDiscardFramebufferEXT)
#endif

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Global framebuffer list (used when restoring after device loss). */
static SysFramebuffer *all_framebuffers;

/* Default framebuffer ID.  This is applied when rendering to the display. */
static GLuint default_framebuffer = 0;

/* Currently bound framebuffer, or NULL if none. */
static SysFramebuffer *current_framebuffer = NULL;

/* Bit depths for depth and stencil buffers. */
static int depth_bits = 16, stencil_bits = 0;

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

int sys_framebuffer_supported(void)
{
    return opengl_has_features(OPENGL_FEATURE_FRAMEBUFFERS);
}

/*-----------------------------------------------------------------------*/

SysFramebuffer *sys_framebuffer_create(int width, int height)
{
    /* If the OpenGL framebuffer extension is not available, we can't do
     * anything. */

    if (!opengl_has_features(OPENGL_FEATURE_FRAMEBUFFERS)) {
        return NULL;
    }

    /* Determine the proper depth and (if applicable) stencil buffer format
     * based on the requested bit depths.  We prefer a packed depth/stencil
     * buffer if such a format is available. */

    GLenum depth_format = 0, stencil_format = 0;
    const int has_packed_24_8 = (
        opengl_major_version() >= 3  // Mandatory in GL 3.0 and GLES 3.0.
        || opengl_has_extension("GL_EXT_packed_depth_stencil")
        || opengl_has_extension("GL_OES_packed_depth_stencil"));
    const int has_packed_32_8 = (opengl_major_version() >= 3);
    if (depth_bits <= 16 && stencil_bits == 0) {
        depth_format = GL_DEPTH_COMPONENT16;
    } else if (depth_bits <= 24 && stencil_bits == 0) {
        depth_format = GL_DEPTH_COMPONENT24;
    } else if (depth_bits <= 24 && stencil_bits <= 8 && has_packed_24_8) {
        depth_format = GL_DEPTH24_STENCIL8;
    } else if (depth_bits <= 32 && stencil_bits == 0) {
        depth_format = GL_DEPTH_COMPONENT32;
    } else if (depth_bits <= 32 && stencil_bits <= 8 && has_packed_32_8) {
        depth_format = GL_DEPTH32F_STENCIL8;
    } else if (depth_bits <= 16 && stencil_bits <= 8) {
        depth_format = GL_DEPTH_COMPONENT16;
        stencil_format = GL_STENCIL_INDEX8;
    } else if (depth_bits <= 24 && stencil_bits <= 8) {
        depth_format = GL_DEPTH_COMPONENT24;
        stencil_format = GL_STENCIL_INDEX8;
    } else if (depth_bits <= 32 && stencil_bits <= 8) {
        depth_format = GL_DEPTH_COMPONENT32;
        stencil_format = GL_STENCIL_INDEX8;
    }
    if (!depth_format) {
        DLOG("Unsupported depth/stencil bit depths: %d/%d",
             depth_bits, stencil_bits);
        return NULL;
    }

    /* Allocate and initialize the SysFramebuffer structure. */

    SysFramebuffer *framebuffer = mem_alloc(sizeof(*framebuffer), 0, 0);
    if (UNLIKELY(!framebuffer)) {
        DLOG("Failed to allocate SysFramebuffer");
        goto error_return;
    }

    framebuffer->width            = width;
    framebuffer->height           = height;
    framebuffer->depth_format     = depth_format;
    framebuffer->stencil_format   = stencil_format;
    framebuffer->has_stencil      = (stencil_bits > 0);
    framebuffer->separate_stencil = (stencil_format != 0);
    framebuffer->antialias        = 1;

    /* Allocate OpenGL resources.  We try to avoid allocating a depth
     * if none is requested, but when using a packed depth/stencil format,
     * we need to allocate a depth buffer even if only a stencil buffer is
     * requested. */

    framebuffer->framebuffer = 0;
    framebuffer->depth_buffer = 0;
    framebuffer->stencil_buffer = 0;
    framebuffer->texture = 0;
    glGenFramebuffers(1, &framebuffer->framebuffer);
    if (UNLIKELY(!framebuffer->framebuffer)) {
        DLOG("Failed to generate an OpenGL framebuffer ID (0x%X)",
             glGetError());
        goto error_free_framebuffer;
    }
    if (depth_bits != 0
     || (stencil_bits != 0 && !framebuffer->separate_stencil)) {
        glGenRenderbuffers(1, &framebuffer->depth_buffer);
        if (UNLIKELY(!framebuffer->depth_buffer)) {
            DLOG("Failed to generate an OpenGL renderbuffer ID (0x%X)",
                 glGetError());
            goto error_delete_gl_framebuffer;
        }
    }
    if (stencil_format != 0) {
        glGenRenderbuffers(1, &framebuffer->stencil_buffer);
        if (UNLIKELY(!framebuffer->stencil_buffer)) {
            DLOG("Failed to generate an OpenGL renderbuffer ID (0x%X)",
                 glGetError());
            goto error_delete_gl_depth_buffer;
        }
    }
    glGenTextures(1, &framebuffer->texture);
    if (UNLIKELY(!framebuffer->texture)) {
        DLOG("Failed to generate an OpenGL texture ID (0x%X)", glGetError());
        goto error_delete_gl_stencil_buffer;
    }

    /* Allocate internal graphics memory for the framebuffer. */

    opengl_clear_error();
    if (framebuffer->depth_buffer) {
        glBindRenderbuffer(GL_RENDERBUFFER, framebuffer->depth_buffer);
        glRenderbufferStorage(GL_RENDERBUFFER, depth_format, width, height);
    }
    if (framebuffer->separate_stencil) {
        glBindRenderbuffer(GL_RENDERBUFFER, framebuffer->stencil_buffer);
        glRenderbufferStorage(GL_RENDERBUFFER, stencil_format, width, height);
    }
    glBindTexture(GL_TEXTURE_2D, framebuffer->texture);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0,
                 GL_RGBA, GL_UNSIGNED_BYTE, NULL);
    if (UNLIKELY(glGetError() != GL_NO_ERROR)) {
        DLOG("Failed to initialize framebuffer resources (%dx%d)",
             width, height);
        goto error_delete_gl_texture;
    }

    /* Bind the texture and depth buffer objects to the framebuffer, and
     * make sure the system accepts the result. */

    glBindFramebuffer(GL_FRAMEBUFFER, framebuffer->framebuffer);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0,
                           GL_TEXTURE_2D, framebuffer->texture, 0);
    if (framebuffer->depth_buffer) {
        glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT,
                                  GL_RENDERBUFFER, framebuffer->depth_buffer);
    }
    if (stencil_bits > 0) {
        glFramebufferRenderbuffer(
            GL_FRAMEBUFFER, GL_STENCIL_ATTACHMENT, GL_RENDERBUFFER,
            framebuffer->separate_stencil ? framebuffer->stencil_buffer
                                          : framebuffer->depth_buffer);
    }
    const int status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
    if (UNLIKELY(status != GL_FRAMEBUFFER_COMPLETE)) {
        if (status == GL_FRAMEBUFFER_UNSUPPORTED) {
            DLOG("Framebuffer not supported by system (size %dx%d)",
                 width, height);
        } else if (status == GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT) {
            DLOG("Framebuffer reported to be incomplete");
        } else {
            DLOG("Unknown error 0x%X while setting up framebuffer (size"
                 " %dx%d)", status, width, height);
        }
        goto error_delete_gl_texture;
    }

    /* Reset the rendering state to normal. */

    glBindTexture(GL_TEXTURE_2D, opengl_current_texture_id);
    glBindRenderbuffer(GL_RENDERBUFFER, 0);
    if (current_framebuffer) {
        glBindFramebuffer(GL_FRAMEBUFFER, current_framebuffer->framebuffer);
    } else {
        glBindFramebuffer(GL_FRAMEBUFFER, default_framebuffer);
    }
    opengl_apply_viewport();
    opengl_apply_clip_region();

    /* Take care of other initialization and return the new framebuffer. */

    framebuffer->next = all_framebuffers;
    if (all_framebuffers) {
        all_framebuffers->prev = framebuffer;
    }
    framebuffer->prev = NULL;
    all_framebuffers = framebuffer;

    return framebuffer;

    /* Error handling. */

  error_delete_gl_texture:
    opengl_delete_texture(framebuffer->texture);
  error_delete_gl_stencil_buffer:
    if (stencil_format != 0) {
        opengl_delete_renderbuffer(framebuffer->stencil_buffer);
    }
  error_delete_gl_depth_buffer:
    opengl_delete_renderbuffer(framebuffer->depth_buffer);
  error_delete_gl_framebuffer:
    opengl_delete_framebuffer(framebuffer->framebuffer);
  error_free_framebuffer:
    mem_free(framebuffer);
  error_return:
    return NULL;
}

/*-----------------------------------------------------------------------*/

void sys_framebuffer_destroy(SysFramebuffer *framebuffer)
{
    if (framebuffer->next) {
        framebuffer->next->prev = framebuffer->prev;
    }
    if (framebuffer->prev) {
        framebuffer->prev->next = framebuffer->next;
    } else {
        all_framebuffers = framebuffer->next;
    }

    if (current_framebuffer == framebuffer) {
        sys_framebuffer_bind(NULL);
    }

    opengl_delete_texture(framebuffer->texture);
    if (framebuffer->separate_stencil) {
        opengl_delete_renderbuffer(framebuffer->stencil_buffer);
    }
    opengl_delete_renderbuffer(framebuffer->depth_buffer);
    opengl_delete_framebuffer(framebuffer->framebuffer);
    mem_free(framebuffer);
}

/*-----------------------------------------------------------------------*/

void sys_framebuffer_bind(SysFramebuffer *framebuffer)
{
    if (opengl_has_features(OPENGL_FEATURE_FRAMEBUFFERS)) {
        if (framebuffer) {
            glBindFramebuffer(GL_FRAMEBUFFER, framebuffer->framebuffer);
        } else {
            glBindFramebuffer(GL_FRAMEBUFFER, default_framebuffer);
        }
        current_framebuffer = framebuffer;
        opengl_framebuffer_changed = 1;
        opengl_apply_viewport();
    }
}

/*-----------------------------------------------------------------------*/

void sys_framebuffer_apply(SysFramebuffer *framebuffer)
{
    opengl_apply_texture(GL_TEXTURE_2D, framebuffer->texture);
    opengl_texture_is_framebuffer = 1;
}

/*-----------------------------------------------------------------------*/

void sys_framebuffer_set_antialias(SysFramebuffer *framebuffer, int on)
{
    if (on && !framebuffer->antialias) {
        glBindTexture(GL_TEXTURE_2D, framebuffer->texture);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glBindTexture(GL_TEXTURE_2D, opengl_current_texture_id);
        framebuffer->antialias = 1;
    } else if (!on && framebuffer->antialias) {
        glBindTexture(GL_TEXTURE_2D, framebuffer->texture);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glBindTexture(GL_TEXTURE_2D, opengl_current_texture_id);
        framebuffer->antialias = 0;
    }
}

/*-----------------------------------------------------------------------*/

void sys_framebuffer_discard_data(
#ifndef SIL_OPENGL_ES
    UNUSED
#endif
    SysFramebuffer *framebuffer)
{
#ifdef SIL_OPENGL_ES  // glInvalidateFramebuffer() not currently supported.
    if (!opengl_has_features(OPENGL_FEATURE_DISCARD_FRAMEBUFFER)) {
        return;
    }

#ifdef SIL_PLATFORM_ANDROID
/* Workaround for https://code.google.com/p/android/issues/detail?id=62903 */
    if (!p_glDiscardFramebufferEXT) {
        p_glDiscardFramebufferEXT =
            dlsym(RTLD_DEFAULT, "glDiscardFramebufferEXT");
        ASSERT(p_glDiscardFramebufferEXT != NULL, return);
    }
#endif

    if (current_framebuffer != framebuffer) {
        glBindFramebuffer(GL_FRAMEBUFFER, framebuffer->framebuffer);
    }

    static const GLenum attachments[] = {
        GL_COLOR_ATTACHMENT0,
        GL_DEPTH_ATTACHMENT,
        GL_STENCIL_ATTACHMENT,
    };
    glDiscardFramebufferEXT(GL_FRAMEBUFFER, lenof(attachments), attachments);

    if (current_framebuffer != framebuffer) {
        if (current_framebuffer) {
            glBindFramebuffer(GL_FRAMEBUFFER, current_framebuffer->framebuffer);
        } else {
            glBindFramebuffer(GL_FRAMEBUFFER, default_framebuffer);
        }
    }
#endif
}

/*************************************************************************/
/*********************** Exported utility routines ***********************/
/*************************************************************************/

void opengl_set_default_framebuffer(GLuint default_fb)
{
    default_framebuffer = default_fb;
}

/*-----------------------------------------------------------------------*/

GLuint opengl_get_default_framebuffer(void)
{
    return default_framebuffer;
}

/*-----------------------------------------------------------------------*/

void opengl_set_depth_bits(int bits)
{
    depth_bits = bits;
}

/*-----------------------------------------------------------------------*/

void opengl_set_stencil_bits(int bits)
{
    stencil_bits = bits;
}

/*************************************************************************/
/******************* Library-internal utility routines *******************/
/*************************************************************************/

SysFramebuffer *opengl_current_framebuffer(void)
{
    return current_framebuffer;
}

/*-----------------------------------------------------------------------*/

void opengl_framebuffer_clear_all(void)
{
    for (SysFramebuffer *framebuffer = all_framebuffers; framebuffer;
         framebuffer = framebuffer->next)
    {
        opengl_delete_texture(framebuffer->texture);
        framebuffer->texture = 0;
        if (framebuffer->separate_stencil) {
            opengl_delete_renderbuffer(framebuffer->stencil_buffer);
            framebuffer->stencil_buffer = 0;
        }
        opengl_delete_renderbuffer(framebuffer->depth_buffer);
        framebuffer->depth_buffer = 0;
        opengl_delete_framebuffer(framebuffer->framebuffer);
        framebuffer->framebuffer = 0;
    }
}

/*-----------------------------------------------------------------------*/

void opengl_framebuffer_restore_all(void)
{
    for (SysFramebuffer *framebuffer = all_framebuffers; framebuffer;
         framebuffer = framebuffer->next)
    {
        opengl_clear_error();

        if (!framebuffer->framebuffer) {
            glGenFramebuffers(1, &framebuffer->framebuffer);
        }
        if (!framebuffer->depth_buffer) {
            glGenRenderbuffers(1, &framebuffer->depth_buffer);
        }
        if (framebuffer->separate_stencil && !framebuffer->stencil_buffer) {
            glGenRenderbuffers(1, &framebuffer->stencil_buffer);
        }
        if (!framebuffer->texture) {
            glGenTextures(1, &framebuffer->texture);
        }

        glBindRenderbuffer(GL_RENDERBUFFER, framebuffer->depth_buffer);
        glRenderbufferStorage(GL_RENDERBUFFER, framebuffer->depth_format,
                              framebuffer->width, framebuffer->height);
        if (framebuffer->separate_stencil) {
            glBindRenderbuffer(GL_RENDERBUFFER, framebuffer->stencil_buffer);
            glRenderbufferStorage(GL_RENDERBUFFER, framebuffer->stencil_format,
                                  framebuffer->width, framebuffer->height);
        }
        glBindTexture(GL_TEXTURE_2D, framebuffer->texture);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexImage2D(GL_TEXTURE_2D, 0,
                     GL_RGBA, framebuffer->width, framebuffer->height, 0,
                     GL_RGBA, GL_UNSIGNED_BYTE, NULL);
        glBindFramebuffer(GL_FRAMEBUFFER, framebuffer->framebuffer);
        glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0,
                               GL_TEXTURE_2D, framebuffer->texture, 0);
        glFramebufferRenderbuffer(
            GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT,
            GL_RENDERBUFFER, framebuffer->depth_buffer);
        if (framebuffer->has_stencil) {
            glFramebufferRenderbuffer(
                GL_FRAMEBUFFER, GL_STENCIL_ATTACHMENT, GL_RENDERBUFFER,
                framebuffer->separate_stencil ? framebuffer->stencil_buffer
                                              : framebuffer->depth_buffer);
        }

        const int error = glGetError();
        const int status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
        int ok = 0;
        if (UNLIKELY(error != GL_NO_ERROR)) {
            DLOG("WARNING: failed to restore %dx%d framebuffer %p"
                 " (error 0x%X)", framebuffer->width, framebuffer->height,
                 framebuffer, error);
        } else if (UNLIKELY(status != GL_FRAMEBUFFER_COMPLETE)) {
            DLOG("WARNING: failed to restore %dx%d framebuffer %p"
                 " (status 0x%X)",framebuffer->width, framebuffer->height,
                 framebuffer, status);
        } else {
            ok = 1;
        }
        if (!ok) {
            opengl_delete_texture(framebuffer->texture);
            framebuffer->texture = 0;
            if (framebuffer->stencil_buffer) {
                opengl_delete_renderbuffer(framebuffer->stencil_buffer);
                framebuffer->stencil_buffer = 0;
            }
            opengl_delete_renderbuffer(framebuffer->depth_buffer);
            framebuffer->depth_buffer = 0;
            opengl_delete_framebuffer(framebuffer->framebuffer);
            framebuffer->framebuffer = 0;
        }

        glBindFramebuffer(GL_FRAMEBUFFER, 0);
        glBindRenderbuffer(GL_RENDERBUFFER, 0);
        glBindTexture(GL_TEXTURE_2D, opengl_current_texture_id);
    }

    sys_framebuffer_bind(current_framebuffer);
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
