/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/opengl/gl-headers.h: Header which wraps the individual OpenGL
 * headers, defining compatibility constants if necessary.
 */

#ifndef SIL_SRC_SYSDEP_OPENGL_GL_HEADERS_H
#define SIL_SRC_SYSDEP_OPENGL_GL_HEADERS_H

/*************************************************************************/
/*************************************************************************/

#define GL_GLEXT_PROTOTYPES

#ifdef SIL_OPENGL_ES

# if defined(SIL_PLATFORM_IOS)
#  include <OpenGLES/ES2/gl.h>
#  include <OpenGLES/ES2/glext.h>
# else
#  include <GLES2/gl2.h>
#  include <GLES2/gl2ext.h>
# endif

/* OpenGL ES 2.0 and later does away with the fixed function pipeline. */
# define SIL_OPENGL_SUPPRESS_FFP_FUNCTIONS

/* The VertexArray functions are only defined with the OES extension, and
 * then only on iOS. */
# ifdef SIL_PLATFORM_IOS
#  define glBindVertexArray     glBindVertexArrayOES
#  define glDeleteVertexArrays  glDeleteVertexArraysOES
#  define glGenVertexArrays     glGenVertexArraysOES
# endif

#else  // !SIL_OPENGL_ES

/* On Windows, gl.h includes windows.h, so we need the near/far hack from
 * sysdep/windows/internal.h here. */
# if defined(SIL_PLATFORM_WINDOWS)
#  undef near
#  undef far
#  include <windef.h>
# endif

# if defined(SIL_PLATFORM_MACOSX)
#  include <OpenGL/gl3.h>
#  include <OpenGL/gl3ext.h>
#  ifndef GLAPIENTRY
#   define GLAPIENTRY APIENTRY
#  endif
#  define SIL_OPENGL_SUPPRESS_FFP_FUNCTIONS
# else
#  include <GL/gl.h>
#  include <GL/glext.h>
# endif

# if defined(SIL_PLATFORM_WINDOWS)
#  undef near
#  undef far
#  define near near_
#  define far far_
# endif

/* Map glDepthRangef() to glDepthRange() for the benefit of old OpenGL
 * implementations that don't have the former. */
# define glDepthRangef  glDepthRange

#endif  // SIL_OPENGL_ES

/* Define functions missing from system headers to invalid glEnable() calls
 * to (1) avoid compiler warnings and (2) trigger errors if the functions
 * are in fact called. */
#define _INVALID_GL_CALL(name)  (DLOG("Invalid GL call: %s",name), glEnable(0))
#ifdef SIL_OPENGL_SUPPRESS_FFP_FUNCTIONS
# define glAlphaFunc(...)            (_INVALID_GL_CALL("glAlphaFunc"))
# define glColor4f(...)              (_INVALID_GL_CALL("glColor4f"))
# define glColorPointer(...)         (_INVALID_GL_CALL("glColorPointer"))
# define glDisableClientState(...)   (_INVALID_GL_CALL("glDisableClientState"))
# define glEnableClientState(...)    (_INVALID_GL_CALL("glEnableClientState"))
# define glFogf(...)                 (_INVALID_GL_CALL("glFogf"))
# define glFogfv(...)                (_INVALID_GL_CALL("glFogfv"))
# define glFogi(...)                 (_INVALID_GL_CALL("glFogi"))
# define glFogx(...)                 (_INVALID_GL_CALL("glFogx"))
# define glLoadIdentity(...)         (_INVALID_GL_CALL("glLoadIdentity"))
# define glLoadMatrixf(...)          (_INVALID_GL_CALL("glLoadMatrixf"))
# define glMatrixMode(...)           (_INVALID_GL_CALL("glMatrixMode"))
# define glMultMatrixf(...)          (_INVALID_GL_CALL("glMultMatrixf"))
# define glTexCoordPointer(...)      (_INVALID_GL_CALL("glTexCoordPointer"))
# define glTranslatef(...)           (_INVALID_GL_CALL("glTranslatef"))
# define glVertexPointer(...)        (_INVALID_GL_CALL("glVertexPointer"))
#endif
#ifdef SIL_OPENGL_ES
# define glGetBufferSubData(...)     (_INVALID_GL_CALL("glGetBufferSubData"))
# define glGetStringi(...)           (_INVALID_GL_CALL("glGetStringi"))
# define glGetTexImage(...)          (_INVALID_GL_CALL("glGetTexImage"))
# define glPointSize(...)            (_INVALID_GL_CALL("glPointSize"))
# ifndef SIL_PLATFORM_IOS
#  define glBindVertexArray(...)     (_INVALID_GL_CALL("glBindVertexArray"))
#  define glDeleteVertexArrays(...)  (_INVALID_GL_CALL("glDeleteVertexArrays"))
#  define glGenVertexArrays(...)     (_INVALID_GL_CALL("glGenVertexArrays"))
# endif
#endif

#if 0  // Enable this block to log all OpenGL calls (requires GCC extensions).
#ifndef SUPPRESS_GLLOG  // Used to avoid problems in sdl/dyngl.c.
# define GLLOG(call)  __extension__({DLOG(#call); (call);})
# define glAttachShader(...)         GLLOG(glAttachShader(__VA_ARGS__))
# define glBindAttribLocation(...)   GLLOG(glBindAttribLocation(__VA_ARGS__))
# define glBindFragDataLocation(...) GLLOG(glBindFragDataLocation(__VA_ARGS__))
# define glBindBuffer(...)           GLLOG(glBindBuffer(__VA_ARGS__))
# define glBindFramebuffer(...)      GLLOG(glBindFramebuffer(__VA_ARGS__))
# define glBindRenderbuffer(...)     GLLOG(glBindRenderbuffer(__VA_ARGS__))
# define glBindTexture(...)          GLLOG(glBindTexture(__VA_ARGS__))
# define glBindVertexArray(...)      GLLOG(glBindVertexArray(__VA_ARGS__))
# define glBlendFunc(...)            GLLOG(glBlendFunc(__VA_ARGS__))
# define glBufferData(...)           GLLOG(glBufferData(__VA_ARGS__))
# define glBufferSubData(...)        GLLOG(glBufferSubData(__VA_ARGS__))
# define glCheckFramebufferStatus(...) GLLOG(glCheckFramebufferStatus(__VA_ARGS__))
# define glClear(...)                GLLOG(glClear(__VA_ARGS__))
# define glClearColor(...)           GLLOG(glClearColor(__VA_ARGS__))
# define glCompileShader(...)        GLLOG(glCompileShader(__VA_ARGS__))
# define glCompressedTexImage2D(...) GLLOG(glCompressedTexImage2D(__VA_ARGS__))
# define glCopyTexImage2D(...)       GLLOG(glCopyTexImage2D(__VA_ARGS__))
# define glCopyTexSubImage2D(...)    GLLOG(glCopyTexSubImage2D(__VA_ARGS__))
# define glCreateProgram(...)        GLLOG(glCreateProgram(__VA_ARGS__))
# define glCreateShader(...)         GLLOG(glCreateShader(__VA_ARGS__))
# define glCullFace(...)             GLLOG(glCullFace(__VA_ARGS__))
# define glDeleteBuffers(...)        GLLOG(glDeleteBuffers(__VA_ARGS__))
# define glDeleteFramebuffers(...)   GLLOG(glDeleteFramebuffers(__VA_ARGS__))
# define glDeleteProgram(...)        GLLOG(glDeleteProgram(__VA_ARGS__))
# define glDeleteRenderbuffers(...)  GLLOG(glDeleteRenderbuffers(__VA_ARGS__))
# define glDeleteShader(...)         GLLOG(glDeleteShader(__VA_ARGS__))
# define glDeleteVertexArrays(...)   GLLOG(glDeleteVertexArrays(__VA_ARGS__))
# define glDeleteTextures(...)       GLLOG(glDeleteTextures(__VA_ARGS__))
# define glDepthFunc(...)            GLLOG(glDepthFunc(__VA_ARGS__))
# define glDepthMask(...)            GLLOG(glDepthMask(__VA_ARGS__))
# define glDisable(...)              GLLOG(glDisable(__VA_ARGS__))
# define glDisableVertexAttribArray(...) GLLOG(glDisableVertexAttribArray(__VA_ARGS__))
# define glDrawArrays(...)           GLLOG(glDrawArrays(__VA_ARGS__))
# define glDrawElements(...)         GLLOG(glDrawElements(__VA_ARGS__))
# define glEnable(...)               GLLOG(glEnable(__VA_ARGS__))
# define glEnableVertexAttribArray(...) GLLOG(glEnableVertexAttribArray(__VA_ARGS__))
# define glFinish(...)               GLLOG(glFinish(__VA_ARGS__))
# define glFramebufferRenderbuffer(...) GLLOG(glFramebufferRenderbuffer(__VA_ARGS__))
# define glFramebufferTexture2D(...) GLLOG(glFramebufferTexture2D(__VA_ARGS__))
# define glFrontFace(...)            GLLOG(glFrontFace(__VA_ARGS__))
# define glGenBuffers(...)           GLLOG(glGenBuffers(__VA_ARGS__))
# define glGenFramebuffers(...)      GLLOG(glGenFramebuffers(__VA_ARGS__))
# define glGenRenderbuffers(...)     GLLOG(glGenRenderbuffers(__VA_ARGS__))
# define glGenTextures(...)          GLLOG(glGenTextures(__VA_ARGS__))
# define glGenVertexArrays(...)      GLLOG(glGenVertexArrays(__VA_ARGS__))
# define glGenerateMipmap(...)       GLLOG(glGenerateMipmap(__VA_ARGS__))
# define glGetError(...)             GLLOG(glGetError(__VA_ARGS__))
# define glGetProgramInfoLog(...)    GLLOG(glGetProgramInfoLog(__VA_ARGS__))
# define glGetProgramiv(...)         GLLOG(glGetProgramiv(__VA_ARGS__))
# define glGetShaderInfoLog(...)     GLLOG(glGetShaderInfoLog(__VA_ARGS__))
# define glGetShaderSource(...)      GLLOG(glGetShaderSource(__VA_ARGS__))
# define glGetShaderiv(...)          GLLOG(glGetShaderiv(__VA_ARGS__))
# define glGetUniformLocation(...)   GLLOG(glGetUniformLocation(__VA_ARGS__))
# define glHint(...)                 GLLOG(glHint(__VA_ARGS__))
# define glLineWidth(...)            GLLOG(glLineWidth(__VA_ARGS__))
# define glLinkProgram(...)          GLLOG(glLinkProgram(__VA_ARGS__))
# define glPixelStorei(...)          GLLOG(glPixelStorei(__VA_ARGS__))
# define glReadPixels(...)           GLLOG(glReadPixels(__VA_ARGS__))
# define glRenderbufferStorage(...)  GLLOG(glRenderbufferStorage(__VA_ARGS__))
# define glShaderSource(...)         GLLOG(glShaderSource(__VA_ARGS__))
# define glTexImage2D(...)           GLLOG(glTexImage2D(__VA_ARGS__))
# define glTexParameteri(...)        GLLOG(glTexParameteri(__VA_ARGS__))
# define glTexSubImage2D(...)        GLLOG(glTexSubImage2D(__VA_ARGS__))
# define glUniform1fv(...)           GLLOG(glUniform1fv(__VA_ARGS__))
# define glUniform2fv(...)           GLLOG(glUniform2fv(__VA_ARGS__))
# define glUniform3fv(...)           GLLOG(glUniform3fv(__VA_ARGS__))
# define glUniform4fv(...)           GLLOG(glUniform4fv(__VA_ARGS__))
# define glUniformMatrix4fv(...)     GLLOG(glUniformMatrix4fv(__VA_ARGS__))
# define glUseProgram(...)           GLLOG(glUseProgram(__VA_ARGS__))
# define glValidateProgram(...)      GLLOG(glValidateProgram(__VA_ARGS__))
# define glVertexAttribPointer(...)  GLLOG(glVertexAttribPointer(__VA_ARGS__))
# define glViewport(...)             GLLOG(glViewport(__VA_ARGS__))
# ifndef SIL_OPENGL_SUPPRESS_FFP_FUNCTIONS
#  define glAlphaFunc(...)           GLLOG(glAlphaFunc(__VA_ARGS__))
#  define glColor4f(...)             GLLOG(glColor4f(__VA_ARGS__))
#  define glColorPointer(...)        GLLOG(glColorPointer(__VA_ARGS__))
#  define glDisableClientState(...)  GLLOG(glDisableClientState(__VA_ARGS__))
#  define glEnableClientState(...)   GLLOG(glEnableClientState(__VA_ARGS__))
#  define glFogf(...)                GLLOG(glFogf(__VA_ARGS__))
#  define glFogfv(...)               GLLOG(glFogfv(__VA_ARGS__))
#  define glFogi(...)                GLLOG(glFogi(__VA_ARGS__))
#  define glFogx(...)                GLLOG(glFogx(__VA_ARGS__))
#  define glLoadIdentity(...)        GLLOG(glLoadIdentity(__VA_ARGS__))
#  define glLoadMatrixf(...)         GLLOG(glLoadMatrixf(__VA_ARGS__))
#  define glMatrixMode(...)          GLLOG(glMatrixMode(__VA_ARGS__))
#  define glMultMatrixf(...)         GLLOG(glMultMatrixf(__VA_ARGS__))
#  define glTexCoordPointer(...)     GLLOG(glTexCoordPointer(__VA_ARGS__))
#  define glTranslatef(...)          GLLOG(glTranslatef(__VA_ARGS__))
#  define glVertexPointer(...)       GLLOG(glVertexPointer(__VA_ARGS__))
# endif
# ifndef SIL_OPENGL_ES
#  define glDepthRange(...)          GLLOG(glDepthRange(__VA_ARGS__))
#  define glGetBufferSubData(...)    GLLOG(glGetBufferSubData(__VA_ARGS__))
#  define glGetStringi(...)          GLLOG(glGetStringi(__VA_ARGS__))
#  define glGetTexImage(...)         GLLOG(glGetTexImage(__VA_ARGS__))
# endif
# ifdef SIL_OPENGL_ES
#  define glDepthRangef(...)         GLLOG(glDepthRangef(__VA_ARGS__))
# endif
#endif  // !SUPPRESS_GLLOG
#endif  // 0/1

/*-----------------------------------------------------------------------*/

/* Make sure optional constants are defined to _something_ to avoid
 * compilation errors.  (The constants aren't used if the respective
 * features are unavailable.)  We can't detect the presence or absence of
 * functions in this way, so system-specific code is responsible for
 * ensuring that all functions are available (even if no-ops) or defined
 * as macros. */

/* General constants. */
#ifndef GL_ALPHA_TEST
# define GL_ALPHA_TEST  GL_INVALID_ENUM
#endif
#ifndef GL_COLOR_ARRAY
# define GL_COLOR_ARRAY  GL_INVALID_ENUM
#endif
#ifndef GL_FOG
# define GL_FOG  GL_INVALID_ENUM
#endif
#ifndef GL_FOG_MODE
# define GL_FOG_MODE  GL_INVALID_ENUM
#endif
#ifndef GL_MODELVIEW
# define GL_MODELVIEW  GL_INVALID_ENUM
#endif
#ifndef GL_NUM_EXTENSIONS
# define GL_NUM_EXTENSIONS  GL_INVALID_ENUM
#endif
#ifndef GL_PROJECTION
# define GL_PROJECTION  GL_INVALID_ENUM
#endif
#ifndef GL_QUAD_STRIP
# define GL_QUAD_STRIP  GL_INVALID_ENUM
#endif
#ifndef GL_QUADS
# define GL_QUADS  GL_INVALID_ENUM
#endif
#if !defined(GL_TEXTURE_3D) && defined(GL_TEXTURE_3D_OES)
# define GL_TEXTURE_3D  GL_TEXTURE_3D_OES
#endif
#ifndef GL_TEXTURE_COORD_ARRAY
# define GL_TEXTURE_COORD_ARRAY  GL_INVALID_ENUM
#endif
#ifndef GL_VERTEX_ARRAY
# define GL_VERTEX_ARRAY  GL_INVALID_ENUM
#endif

/* Constants for pixel data types. */
#ifndef GL_BGRA
# ifdef GL_BGRA_EXT
#  define GL_BGRA  GL_BGRA_EXT
# else
#  define GL_BGRA  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_UNSIGNED_SHORT_5_6_5_REV
# ifdef GL_UNSIGNED_SHORT_5_6_5_REV_EXT
#  define GL_UNSIGNED_SHORT_5_6_5_REV  GL_UNSIGNED_SHORT_5_6_5_REV_EXT
# else
#  define GL_UNSIGNED_SHORT_5_6_5_REV  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_UNSIGNED_SHORT_1_5_5_5_REV
# ifdef GL_UNSIGNED_SHORT_1_5_5_5_REV_EXT
#  define GL_UNSIGNED_SHORT_1_5_5_5_REV  GL_UNSIGNED_SHORT_1_5_5_5_REV_EXT
# else
#  define GL_UNSIGNED_SHORT_1_5_5_5_REV  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_UNSIGNED_SHORT_4_4_4_4_REV
# ifdef GL_UNSIGNED_SHORT_4_4_4_4_REV_EXT
#  define GL_UNSIGNED_SHORT_4_4_4_4_REV  GL_UNSIGNED_SHORT_4_4_4_4_REV_EXT
# else
#  define GL_UNSIGNED_SHORT_4_4_4_4_REV  GL_INVALID_ENUM
# endif
#endif

/* Constants for texture formats.  Some of these are missing from system
 * headers even when they should be present, so insert the real values. */
#ifndef GL_COMPRESSED_RGB_S3TC_DXT1_EXT
# define GL_COMPRESSED_RGB_S3TC_DXT1_EXT  0x83F0
#endif
#ifndef GL_COMPRESSED_RGBA_S3TC_DXT3_EXT
# define GL_COMPRESSED_RGBA_S3TC_DXT3_EXT  0x83F2
#endif
#ifndef GL_COMPRESSED_RGBA_S3TC_DXT5_EXT
# define GL_COMPRESSED_RGBA_S3TC_DXT5_EXT  0x83F3
#endif
#ifndef GL_COMPRESSED_RGB_PVRTC_2BPPV1_IMG
# define GL_COMPRESSED_RGB_PVRTC_2BPPV1_IMG  0x8C01
#endif
#ifndef GL_COMPRESSED_RGBA_PVRTC_2BPPV1_IMG
# define GL_COMPRESSED_RGBA_PVRTC_2BPPV1_IMG  0x8C03
#endif
#ifndef GL_COMPRESSED_RGB_PVRTC_4BPPV1_IMG
# define GL_COMPRESSED_RGB_PVRTC_4BPPV1_IMG  0x8C00
#endif
#ifndef GL_COMPRESSED_RGBA_PVRTC_4BPPV1_IMG
# define GL_COMPRESSED_RGBA_PVRTC_4BPPV1_IMG  0x8C02
#endif

/* Constants for textures with external image buffers (OpenGL ES only). */
#ifndef GL_TEXTURE_EXTERNAL_OES
# define GL_TEXTURE_EXTERNAL_OES  GL_INVALID_ENUM
#endif
#ifndef GL_SAMPLER_EXTERNAL_OES
# define GL_SAMPLER_EXTERNAL_OES  GL_INVALID_ENUM
#endif
#ifndef GL_TEXTURE_BINDING_EXTERNAL_OES
# define GL_TEXTURE_BINDING_EXTERNAL_OES  GL_INVALID_ENUM
#endif
#ifndef GL_REQUIRED_TEXTURE_IMAGE_UNITS_OES
# define GL_REQUIRED_TEXTURE_IMAGE_UNITS_OES  GL_INVALID_ENUM
#endif

/* Constants for blend types. */
#ifndef GL_CONSTANT_COLOR
# define GL_CONSTANT_COLOR  GL_INVALID_ENUM
#endif
#ifndef GL_ONE_MINUS_CONSTANT_COLOR
# define GL_ONE_MINUS_CONSTANT_COLOR  GL_INVALID_ENUM
#endif

/* Constants related to vertex/index buffers. */
#ifndef GL_ARRAY_BUFFER
# ifdef GL_ARRAY_BUFFER_ARB
#  define GL_ARRAY_BUFFER  GL_ARRAY_BUFFER_ARB
# else
#  define GL_ARRAY_BUFFER  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_ELEMENT_ARRAY_BUFFER
# ifdef GL_ELEMENT_ARRAY_BUFFER_ARB
#  define GL_ELEMENT_ARRAY_BUFFER  GL_ELEMENT_ARRAY_BUFFER_ARB
# else
#  define GL_ELEMENT_ARRAY_BUFFER  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_DYNAMIC_DRAW
# ifdef GL_DYNAMIC_DRAW_ARB
#  define GL_DYNAMIC_DRAW  GL_DYNAMIC_DRAW_ARB
# else
#  define GL_DYNAMIC_DRAW  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_STATIC_DRAW
# ifdef GL_STATIC_DRAW_ARB
#  define GL_STATIC_DRAW  GL_STATIC_DRAW_ARB
# else
#  define GL_STATIC_DRAW  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_WRITE_ONLY
# ifndef GL_WRITE_ONLY_ARB
#  define GL_WRITE_ONLY  GL_WRITE_ONLY_ARB
# else
#  define GL_WRITE_ONLY  GL_INVALID_ENUM
# endif
#endif

/* Constants related to vertex aray objects. */
#ifndef GL_VERTEX_ARRAY_BINDING
# ifdef GL_VERTEX_ARRAY_BINDING_ARB
#  define GL_VERTEX_ARRAY_BINDING  GL_VERTEX_ARRAY_BINDING_ARB
# elif defined(GL_VERTEX_ARRAY_BINDING_OES)
#  define GL_VERTEX_ARRAY_BINDING  GL_VERTEX_ARRAY_BINDING_OES
# else
#  define GL_VERTEX_ARRAY_BINDING  GL_INVALID_ENUM
# endif
#endif

/* Constants related to shaders. */
#ifndef GL_COMPILE_STATUS
# define GL_COMPILE_STATUS  GL_INVALID_ENUM
#endif
#ifndef GL_FRAGMENT_SHADER
# define GL_FRAGMENT_SHADER  GL_INVALID_ENUM
#endif
#ifndef GL_INFO_LOG_LENGTH
# define GL_INFO_LOG_LENGTH  GL_INVALID_ENUM
#endif
#ifndef GL_LINK_STATUS
# define GL_LINK_STATUS  GL_INVALID_ENUM
#endif
#ifndef GL_PROGRAM_POINT_SIZE
# ifdef GL_PROGRAM_POINT_SIZE_EXT
#  define GL_PROGRAM_POINT_SIZE  GL_PROGRAM_POINT_SIZE_EXT
# else
#  define GL_PROGRAM_POINT_SIZE  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_SHADER_SOURCE_LENGTH
# define GL_SHADER_SOURCE_LENGTH  GL_INVALID_ENUM
#endif
#ifndef GL_SHADING_LANGUAGE_VERSION
# define GL_SHADING_LANGUAGE_VERSION  GL_INVALID_ENUM
#endif
#ifndef GL_VALIDATE_STATUS
# define GL_VALIDATE_STATUS  GL_INVALID_ENUM
#endif
#ifndef GL_VERTEX_SHADER
# define GL_VERTEX_SHADER  GL_INVALID_ENUM
#endif

/* Constants related to framebuffers. */
#ifndef GL_COLOR_ATTACHMENT0
# ifdef SIL_OPENGL_ES
#  define GL_COLOR_ATTACHMENT0  GL_COLOR_ATTACHMENT0_OES
# elif defined(GL_COLOR_ATTACHMENT0_EXT)
#  define GL_COLOR_ATTACHMENT0  GL_COLOR_ATTACHMENT0_EXT
# else
#  define GL_COLOR_ATTACHMENT0  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_DEPTH_ATTACHMENT
# ifdef SIL_OPENGL_ES
#  define GL_DEPTH_ATTACHMENT  GL_DEPTH_ATTACHMENT_OES
# elif defined(GL_DEPTH_ATTACHMENT_EXT)
#  define GL_DEPTH_ATTACHMENT  GL_DEPTH_ATTACHMENT_EXT
# else
#  define GL_DEPTH_ATTACHMENT  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_DEPTH_COMPONENT16
# ifdef SIL_OPENGL_ES
#  define GL_DEPTH_COMPONENT16  GL_DEPTH_COMPONENT16_OES
# elif defined(GL_DEPTH_COMPONENT16_EXT)
#  define GL_DEPTH_COMPONENT16  GL_DEPTH_COMPONENT16_EXT
# else
#  define GL_DEPTH_COMPONENT16  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_DEPTH_COMPONENT24
# ifdef SIL_OPENGL_ES
#  define GL_DEPTH_COMPONENT24  GL_DEPTH_COMPONENT24_OES
# elif defined(GL_DEPTH_COMPONENT24_EXT)
#  define GL_DEPTH_COMPONENT24  GL_DEPTH_COMPONENT24_EXT
# else
#  define GL_DEPTH_COMPONENT24  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_DEPTH_COMPONENT32
# ifdef SIL_OPENGL_ES
#  define GL_DEPTH_COMPONENT32  GL_DEPTH_COMPONENT32_OES
# elif defined(GL_DEPTH_COMPONENT32_EXT)
#  define GL_DEPTH_COMPONENT32  GL_DEPTH_COMPONENT32_EXT
# else
#  define GL_DEPTH_COMPONENT32  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_DEPTH24_STENCIL8
# ifdef SIL_OPENGL_ES
#  define GL_DEPTH24_STENCIL8  GL_DEPTH24_STENCIL8_OES
# elif defined(GL_DEPTH24_STENCIL8_EXT)
#  define GL_DEPTH24_STENCIL8  GL_DEPTH24_STENCIL8_EXT
# else
#  define GL_DEPTH24_STENCIL8  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_DEPTH32F_STENCIL8
# ifdef GL_DEPTH32F_STENCIL8_EXT
#  define GL_DEPTH32F_STENCIL8  GL_DEPTH32F_STENCIL8_EXT
# else
#  define GL_DEPTH32F_STENCIL8  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_FRAMEBUFFER
# ifdef SIL_OPENGL_ES
#  define GL_FRAMEBUFFER  GL_FRAMEBUFFER_OES
# elif defined(GL_FRAMEBUFFER_EXT)
#  define GL_FRAMEBUFFER  GL_FRAMEBUFFER_EXT
# else
#  define GL_FRAMEBUFFER  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_FRAMEBUFFER_BINDING
# ifdef SIL_OPENGL_ES
#  define GL_FRAMEBUFFER_BINDING  GL_FRAMEBUFFER_BINDING_OES
# elif defined(GL_FRAMEBUFFER_BINDING_EXT)
#  define GL_FRAMEBUFFER_BINDING  GL_FRAMEBUFFER_BINDING_EXT
# else
#  define GL_FRAMEBUFFER_BINDING  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_FRAMEBUFFER_COMPLETE
# ifdef SIL_OPENGL_ES
#  define GL_FRAMEBUFFER_COMPLETE  GL_FRAMEBUFFER_COMPLETE_OES
# elif defined(GL_FRAMEBUFFER_COMPLETE_EXT)
#  define GL_FRAMEBUFFER_COMPLETE  GL_FRAMEBUFFER_COMPLETE_EXT
# else
#  define GL_FRAMEBUFFER_COMPLETE  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT
# ifdef SIL_OPENGL_ES
#  define GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT  GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_OES
# elif defined(GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT)
#  define GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT  GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT
# else
#  define GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_FRAMEBUFFER_UNSUPPORTED
# ifdef SIL_OPENGL_ES
#  define GL_FRAMEBUFFER_UNSUPPORTED  GL_FRAMEBUFFER_UNSUPPORTED_OES
# elif defined(GL_FRAMEBUFFER_UNSUPPORTED_EXT)
#  define GL_FRAMEBUFFER_UNSUPPORTED  GL_FRAMEBUFFER_UNSUPPORTED_EXT
# else
#  define GL_FRAMEBUFFER_UNSUPPORTED  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_RENDERBUFFER
# ifdef SIL_OPENGL_ES
#  define GL_RENDERBUFFER  GL_RENDERBUFFER_OES
# elif defined(GL_RENDERBUFFER_EXT)
#  define GL_RENDERBUFFER  GL_RENDERBUFFER_EXT
# else
#  define GL_RENDERBUFFER  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_RENDERBUFFER_BINDING
# if defined(GL_RENDERBUFFER_BINDING_EXT)
#  define GL_RENDERBUFFER_BINDING  GL_RENDERBUFFER_BINDING_EXT
# else
#  define GL_RENDERBUFFER_BINDING  GL_INVALID_ENUM
# endif
#endif

/* Constants related to multisampling. */
#ifndef GL_SAMPLE_BUFFERS
# if defined(GL_SAMPLE_BUFFERS_ARB)
#  define GL_SAMPLE_BUFFERS  GL_SAMPLE_BUFFERS_ARB
# else
#  define GL_SAMPLE_BUFFERS  GL_INVALID_ENUM
# endif
#endif
#ifndef GL_SAMPLES
# if defined(GL_SAMPLES_ARB)
#  define GL_SAMPLES  GL_SAMPLES_ARB
# else
#  define GL_SAMPLES  GL_INVALID_ENUM
# endif
#endif

/*-----------------------------------------------------------------------*/

/* Disable glGetError() if requested by the system configuration. */

#ifdef SIL_OPENGL_DISABLE_GETERROR
# undef glGetError
# define glGetError() GL_NO_ERROR
#endif

/*************************************************************************/
/*************************************************************************/

#endif  // SIL_SRC_SYSDEP_OPENGL_GL_HEADERS_H

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
