/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/opengl/graphics.c: Graphics and rendering functionality for
 * OpenGL-based ports.
 */

/*
 * This file (along with the other sources in this directory) provides the
 * definitions for all sys_framebuffer_*(), sys_graphics_*(), and
 * sys_texture_*() functions _except_ for the following:
 *
 *    sys_graphics_init()
 *    sys_graphics_cleanup()
 *    sys_graphics_device_width()
 *    sys_graphics_device_height()
 *    sys_graphics_set_display_attr()
 *    sys_graphics_set_display_mode()
 *    sys_graphics_set_window_title()
 *    sys_graphics_set_window_icon()
 *    sys_graphics_show_mouse_pointer()
 *    sys_graphics_get_mouse_pointer_state()
 *    sys_graphics_get_frame_period()
 *    sys_graphics_has_focus()
 *    sys_graphics_start_frame()
 *    sys_graphics_finish_frame()
 *    sys_graphics_sync()
 */

#define IN_SYSDEP

#include "src/base.h"
#include "src/graphics.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/sysdep/opengl/internal.h"
#include "src/sysdep/opengl/opengl.h"
#include "src/sysdep/opengl/shader-table.h"

/*************************************************************************/
/****************** Global data (only used for testing) ******************/
/*************************************************************************/

#ifdef INCLUDE_TESTS
int TEST_opengl_force_feature_flags = 0;
int TEST_opengl_force_feature_mask = 0;
uint8_t TEST_opengl_always_convert_texture_data = 0;
#endif

/*************************************************************************/
/************** Exported data (local to the OpenGL library) **************/
/*************************************************************************/

int opengl_window_width, opengl_window_height;

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Major and minor OpenGL version. */

static int major_version;
static int minor_version;


/* Feature and supported-format flags for OpenGL extensions. */

static int features_avail;
static int formats_supported;


/* Data for and array of objects to delete in DELAYED_DELETE mode. */

typedef enum DeleteType DeleteType;
enum DeleteType {
    BUFFER = 1,
    FRAMEBUFFER,
    PROGRAM,
    RENDERBUFFER,
    SHADER,
    TEXTURE,
    VERTEX_ARRAY,
};

typedef struct DeleteInfo DeleteInfo;
struct DeleteInfo {
    DeleteType type;
    GLuint object;
};

static DeleteInfo *delete_info;        // Dynamically-allocated array.
static int delete_info_size;           // Array size (never shrinks).
static int delete_info_len;            // Number of actual entries.
static uint8_t delete_info_fixed_size; // Is the array buffer size fixed?

/*-----------------------------------------------------------------------*/

/* Local function declarations. */

/**
 * delayed_delete_one_resource:  Add the given GL resource to the array of
 * resources to delete at the next opengl_free_dead_resources() call.
 *
 * [Parameters]
 *     type: Resource type (as for DeleteInfo.type).
 *     object: GL object ID.
 */
static void delayed_delete_one_resource(DeleteType type, GLuint object);

/**
 * delete_one_resource:  Delete the given GL resource.
 *
 * [Parameters]
 *     type: Resources type (as for DeleteInfo.type).
 *     object: GL object ID.
 */
static void delete_one_resource(DeleteType type, GLuint object);

/*************************************************************************/
/********* Interface functions common to all OpenGL environments *********/
/*************************************************************************/

const char *sys_graphics_renderer_info(void)
{
    static char buf[256];
    strformat(buf, sizeof(buf), "OpenGL version %d.%d (GL_VERSION: %s)",
              major_version, minor_version,
              (const char *)glGetString(GL_VERSION));
    return buf;
}

/*-----------------------------------------------------------------------*/

void sys_graphics_clear(float r, float g, float b)
{
    glClearColor(r, g, b, 0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
}

/*-----------------------------------------------------------------------*/

int sys_graphics_read_pixels(int x, int y, int w, int h, int stride,
                             void *buffer)
{
    opengl_clear_error();
    if (stride == w) {
        glReadPixels(x, y, w, h, GL_RGBA, GL_UNSIGNED_BYTE, buffer);
    } else {
#ifdef SIL_OPENGL_ES
        uint32_t *tempbuf = mem_alloc(w*h*4, 0, MEM_ALLOC_TEMP);
        if (UNLIKELY(!tempbuf)) {
            DLOG("No memory for temporary buffer (%dx%d)", w, h);
            return 0;
        }
        glReadPixels(x, y, w, h, GL_RGBA, GL_UNSIGNED_BYTE, tempbuf);
        uint32_t *buffer32 = buffer;
        for (int j = 0; j < h; j++) {
            memcpy(buffer32 + j*stride, tempbuf + j*w, 4*w);
        }
        mem_free(tempbuf);
#else
        glPixelStorei(GL_PACK_ROW_LENGTH, stride);
        glReadPixels(x, y, w, h, GL_RGBA, GL_UNSIGNED_BYTE, buffer);
        glPixelStorei(GL_PACK_ROW_LENGTH, 0);
#endif
    }
    const int error = glGetError();
    if (error != GL_NO_ERROR) {
        DLOG("Failed to read pixels: 0x%04X", error);
        return 0;
    }

    uint32_t *buffer32 = buffer;
    for (int j = 0; j < h; j++, buffer32 += stride) {
        for (int i = 0; i < w; i++) {
#ifdef IS_LITTLE_ENDIAN
            buffer32[i] |= 0xFF000000;
#else
            buffer32[i] |= 0x000000FF;
#endif
        }
    }

    return 1;
}

/*************************************************************************/
/*********************** Exported utility routines ***********************/
/*************************************************************************/

int opengl_init(int width, int height, int features)
{
#ifdef INCLUDE_TESTS
    ASSERT((TEST_opengl_force_feature_flags
            & ~TEST_opengl_force_feature_mask) == 0);
    ASSERT(((features & TEST_opengl_force_feature_mask)
            ^ TEST_opengl_force_feature_flags)
           == TEST_opengl_force_feature_mask);
    if (TEST_opengl_force_feature_mask) {
        DLOG("Flipping feature flags: 0x%X", TEST_opengl_force_feature_mask);
        features = ((features & ~TEST_opengl_force_feature_mask)
                    | TEST_opengl_force_feature_flags);
    }
#endif

    const char *gl_version = (const char *)glGetString(GL_VERSION);
    ASSERT(gl_version, gl_version = "");
#ifdef SIL_OPENGL_ES
    if (strncmp(gl_version, "OpenGL ES ", 10) == 0) {
        gl_version += 10;
    } else if (strncmp(gl_version, "OpenGL ES-CM ", 13) == 0
            || strncmp(gl_version, "OpenGL ES-CL ", 13) == 0) {
        gl_version += 13;
    } else {
        DLOG("WARNING: Invalid OpenGL ES version string: [%s]", gl_version);
        gl_version = "2.0";
    }
#endif
    const char *minor_str;
    if ((major_version = strtoul(gl_version, (char **)&minor_str, 10)) < 1
     || *minor_str != '.'
     || (minor_str[1] < '0' || minor_str[1] > '9')) {
        DLOG("WARNING: Invalid OpenGL version number: [%s]", gl_version);
        // Assume the minimum supported version.
#ifdef SIL_OPENGL_ES
        major_version = 2;
        minor_version = 0;
#else
        major_version = 1;
        minor_version = 1;
#endif
    } else {
        minor_version = strtoul(minor_str + 1, NULL, 10);
    }

    opengl_window_width = width;
    opengl_window_height = height;
    features_avail = features & ~OPENGL_AUTOCONFIG_FEATURE_MASK;
#ifndef SIL_OPENGL_SUPPRESS_FFP_FUNCTIONS
    features_avail |= OPENGL_FEATURE_FFP_PRIMITIVES;
#endif
#ifdef SIL_OPENGL_ES
    features_avail |= OPENGL_FEATURE_COMPRESSEDTEXIMAGE;
    if (opengl_has_extension("GL_EXT_discard_framebuffer")) {
        features_avail |= OPENGL_FEATURE_DISCARD_FRAMEBUFFER;
    }
    formats_supported = OPENGL_FORMAT_ALPHA;
    if (opengl_has_extension("GL_OES_element_index_uint")) {
        formats_supported |= OPENGL_FORMAT_INDEX32;
    }
    if (opengl_has_extension("GL_EXT_texture_format_BGRA8888")
        /* Apple apparently doesn't believe in following standards... */
     || opengl_has_extension("GL_APPLE_texture_format_BGRA8888")) {
        /* Some GLES implementations also advertise GL_EXT_bgra, but
         * apparently that extension name is only valid for non-ES OpenGL. */
        formats_supported |= OPENGL_FORMAT_BGRA;
    }
#else  // !SIL_OPENGL_ES
    features_avail |= OPENGL_FEATURE_GETTEXIMAGE;
    if (major_version >= 2 || (major_version == 1 && minor_version >= 3)
     || opengl_has_extension("GL_ARB_texture_compression")) {
        features_avail |= OPENGL_FEATURE_COMPRESSEDTEXIMAGE;
    }
    formats_supported = OPENGL_FORMAT_BITREV | OPENGL_FORMAT_INDEX32;
    if (major_version >= 4
     || (major_version == 3 && minor_version >= 2)
     || opengl_has_extension("GL_EXT_bgra")) {
        formats_supported |= OPENGL_FORMAT_BGRA;
    }
    if (major_version < 3) {
        formats_supported |= OPENGL_FORMAT_ALPHA;
    }
#endif
    if (features_avail & OPENGL_FEATURE_COMPRESSEDTEXIMAGE) {
        if (opengl_has_extension("GL_EXT_texture_compression_s3tc")) {
            formats_supported |= OPENGL_FORMAT_S3TC;
        }
        if (opengl_has_extension("GL_IMG_texture_compression_pvrtc")) {
            formats_supported |= OPENGL_FORMAT_PVRTC;
        }
    }

#ifdef DEBUG
    const char *glsl_version = "";
    if (major_version >= 2) {
        glsl_version = (const char *)glGetString(GL_SHADING_LANGUAGE_VERSION);
        ASSERT(glsl_version, glsl_version = "");
# ifdef SIL_OPENGL_ES
        if (strncmp(glsl_version, "OpenGL ES GLSL ES ", 18) == 0) {
            glsl_version += 18;
        } else {
            DLOG("WARNING: Invalid OpenGL ES shader version string: [%s]",
                 glsl_version);
            glsl_version = "1.00";
        }
# endif
    }
    const char *gl_vendor = (const char *)glGetString(GL_VENDOR);
    ASSERT(gl_vendor, gl_vendor = "");
    const char *gl_renderer = (const char *)glGetString(GL_RENDERER);
    ASSERT(gl_renderer, gl_renderer = "");
    DLOG("OpenGL%s version: %s",
# ifdef SIL_OPENGL_ES
         " ES",
# else
         "",
# endif
         *gl_version ? gl_version : "(unknown)");
    if (major_version >= 2) {
        DLOG("GLSL version: %s", *glsl_version ? glsl_version : "(unknown)");
    }
    DLOG("OpenGL vendor: %s", *gl_vendor ? gl_vendor : "(unknown)");
    DLOG("OpenGL renderer: %s", *gl_renderer ? gl_renderer : "(unknown)");
#ifndef SIL_OPENGL_ES
    if (major_version >= 3) {
        GLint num_extensions = 0;
        glGetIntegerv(GL_NUM_EXTENSIONS, &num_extensions);
        for (GLint i = 0; i < num_extensions; i++) {
            DLOG("OpenGL extension %d: %s", i, glGetStringi(GL_EXTENSIONS, i));
        }
    } else
#endif  // !SIL_OPENGL_ES
    {
        const char *gl_extensions = (const char *)glGetString(GL_EXTENSIONS);
        ASSERT(gl_extensions, gl_extensions = "");
        DLOG("OpenGL extensions: %s",
             *gl_extensions ? gl_extensions : "(none)");
    }

#define DEFINE_FLAG(flag) {flag, #flag}
    static const struct {int flag; const char *name;} flag_names[] = {
        DEFINE_FLAG(OPENGL_FEATURE_DELAYED_DELETE),
        DEFINE_FLAG(OPENGL_FEATURE_FAST_DYNAMIC_VBO),
        DEFINE_FLAG(OPENGL_FEATURE_FAST_STATIC_VBO),
        DEFINE_FLAG(OPENGL_FEATURE_FRAMEBUFFERS),
        DEFINE_FLAG(OPENGL_FEATURE_GENERATEMIPMAP),
        DEFINE_FLAG(OPENGL_FEATURE_NATIVE_QUADS),
        DEFINE_FLAG(OPENGL_FEATURE_SHADERS),
        DEFINE_FLAG(OPENGL_FEATURE_USE_STATIC_VAO),
        DEFINE_FLAG(OPENGL_FEATURE_COMPRESSEDTEXIMAGE),
        DEFINE_FLAG(OPENGL_FEATURE_DISCARD_FRAMEBUFFER),
        DEFINE_FLAG(OPENGL_FEATURE_FFP_PRIMITIVES),
        DEFINE_FLAG(OPENGL_FEATURE_GETTEXIMAGE),
    }, format_names[] = {
        DEFINE_FLAG(OPENGL_FORMAT_BGRA),
        DEFINE_FLAG(OPENGL_FORMAT_BITREV),
        DEFINE_FLAG(OPENGL_FORMAT_ALPHA),
        DEFINE_FLAG(OPENGL_FORMAT_S3TC),
        DEFINE_FLAG(OPENGL_FORMAT_PVRTC),
        DEFINE_FLAG(OPENGL_FORMAT_INDEX32),
    };
#undef DEFINE_FLAG
    DLOG("OpenGL features enabled: 0x%X", features_avail);
    for (int i = 0; i < lenof(flag_names); i++) {
        if (opengl_has_features(flag_names[i].flag)) {
            DLOG("   %s", flag_names[i].name);
        }
    }
    DLOG("OpenGL texture formats supported: 0x%X", formats_supported);
    for (int i = 0; i < lenof(format_names); i++) {
        if (opengl_has_formats(format_names[i].flag)) {
            DLOG("   %s", format_names[i].name);
        }
    }
#endif  // DEBUG

    /* Delay these checks until after the log output to help debugging. */
#ifdef SIL_OPENGL_ES
    if (major_version < 2) {
        /* The program shouldn't even compile with GLES 1.1 libraries, but
         * check just in case. */
        DLOG("OpenGL ES version (%d.%d) is too old!  Need at least OpenGL"
             " ES 2.0.", major_version, minor_version);
        return 0;
    }
#else
    if (major_version < 1 || (major_version == 1 && minor_version < 1)) {
        DLOG("OpenGL version (%d.%d) is too old!  Need at least OpenGL 1.1.",
             major_version, minor_version);
        return 0;
    }
    if (major_version == 1) {
        if (!opengl_has_extension("GL_ARB_texture_non_power_of_two")
         || !opengl_has_extension("GL_EXT_blend_color")
         || !opengl_has_extension("GL_EXT_blend_subtract")) {
            DLOG("OpenGL does not support required extensions!"
                 " (missing:%s%s%s)",
                 opengl_has_extension("GL_ARB_texture_non_power_of_two")
                     ? "" : " ARB_texture_non_power_of_two",
                 opengl_has_extension("GL_EXT_blend_color")
                     ? "" : " EXT_blend_color",
                 opengl_has_extension("GL_EXT_blend_subtract")
                     ? "" : " EXT_blend_subtract");
            return 0;
        }
    }
#endif

    opengl_state_init();
    opengl_set_shader_callbacks(NULL, NULL, 0, 0);
    sys_framebuffer_bind(NULL);
    sys_texture_apply(NULL);

    return 1;
}

/*-----------------------------------------------------------------------*/

void opengl_cleanup(void)
{
    opengl_primitive_cleanup();
    opengl_shader_clear_all();
    shader_table_init(0, 1);  // Ensure the table memory is also freed.
    opengl_free_dead_resources(1);
    opengl_sync();
}

/*-----------------------------------------------------------------------*/

int opengl_major_version(void)
{
    return major_version;
}

/*-----------------------------------------------------------------------*/

int opengl_minor_version(void)
{
    return minor_version;
}

/*-----------------------------------------------------------------------*/

int opengl_has_extension(const char *name)
{
    PRECOND(name != NULL, return 0);

    if (!*name) {
        return 0;
    }

#ifndef SIL_OPENGL_ES
    if (major_version >= 3) {
        GLint num_extensions = 0;
        glGetIntegerv(GL_NUM_EXTENSIONS, &num_extensions);
        for (GLint i = 0; i < num_extensions; i++) {
            const char *extension =
                (const char *)glGetStringi(GL_EXTENSIONS, i);
            if (strcmp(extension, name) == 0) {
                return 1;
            }
        }
        return 0;
    } else
#endif  // !SIL_OPENGL_ES
    {
        const char *extensions = (const char *)glGetString(GL_EXTENSIONS);
        ASSERT(extensions, extensions = "");
        const int namelen = strlen(name);
        const char *s = extensions;
        while ((s = strstr(s, name)) != NULL) {
            if ((s == extensions || s[-1] == ' ')
             && (s[namelen] == 0 || s[namelen] == ' ')) {
                return 1;
            }
            s += namelen;
            s += strcspn(s, " ");
        }
        return 0;
    }
}

/*-----------------------------------------------------------------------*/

int opengl_has_features(int features)
{
    return (features_avail & features) == features;
}

/*-----------------------------------------------------------------------*/

int opengl_has_formats(int formats)
{
    return (formats_supported & formats) == formats;
}

/*-----------------------------------------------------------------------*/

int opengl_set_delete_buffer_size(int size)
{
    delete_info_fixed_size = 0;
    opengl_free_dead_resources(1);

    if (opengl_has_features(OPENGL_FEATURE_DELAYED_DELETE) && size > 0) {
        delete_info = mem_alloc(size * sizeof(*delete_info), 0, 0);
        if (UNLIKELY(!delete_info)) {
            DLOG("No memory for %d delete entries", size);
            return 0;
        }
        delete_info_size = size;
        delete_info_fixed_size = 1;
    }
    return 1;
}

/*-----------------------------------------------------------------------*/

void opengl_start_frame(void)
{
    opengl_primitive_reset_bindings();
}

/*-----------------------------------------------------------------------*/

void opengl_sync(void)
{
    glFinish();
}

/*-----------------------------------------------------------------------*/

void opengl_free_dead_resources(int also_array)
{
    for (int i = 0; i < delete_info_len; i++) {
        delete_one_resource(delete_info[i].type, delete_info[i].object);
    }
    delete_info_len = 0;
    if (also_array && !delete_info_fixed_size) {
        mem_free(delete_info);
        delete_info = NULL;
        delete_info_size = 0;
    }
}

/*-----------------------------------------------------------------------*/

void opengl_save_state(void)
{
    opengl_texture_save_all();
    opengl_shader_clear_all();
    opengl_primitive_save_all();
    opengl_primitive_cleanup();
    opengl_framebuffer_clear_all();
    opengl_free_dead_resources(1);
}

/*-----------------------------------------------------------------------*/

void opengl_restore_state(void)
{
    opengl_framebuffer_restore_all();
    opengl_primitive_restore_all();
    opengl_state_reapply();
    opengl_texture_restore_all();
}

/*************************************************************************/
/******************* Library-internal utility routines *******************/
/*************************************************************************/

void opengl_delete_buffer(GLuint object)
{
    if (opengl_has_features(OPENGL_FEATURE_DELAYED_DELETE)) {
        delayed_delete_one_resource(BUFFER, object);
    } else {
        delete_one_resource(BUFFER, object);
    }
}

/*-----------------------------------------------------------------------*/

void opengl_delete_framebuffer(GLuint object)
{
    if (opengl_has_features(OPENGL_FEATURE_DELAYED_DELETE)) {
        delayed_delete_one_resource(FRAMEBUFFER, object);
    } else {
        delete_one_resource(FRAMEBUFFER, object);
    }
}

/*-----------------------------------------------------------------------*/

void opengl_delete_program(GLuint object)
{
    if (opengl_has_features(OPENGL_FEATURE_DELAYED_DELETE)) {
        delayed_delete_one_resource(PROGRAM, object);
    } else {
        delete_one_resource(PROGRAM, object);
    }
}

/*-----------------------------------------------------------------------*/

void opengl_delete_renderbuffer(GLuint object)
{
    if (opengl_has_features(OPENGL_FEATURE_DELAYED_DELETE)) {
        delayed_delete_one_resource(RENDERBUFFER, object);
    } else {
        delete_one_resource(RENDERBUFFER, object);
    }
}

/*-----------------------------------------------------------------------*/

void opengl_delete_shader(GLuint object)
{
    if (opengl_has_features(OPENGL_FEATURE_DELAYED_DELETE)) {
        delayed_delete_one_resource(SHADER, object);
    } else {
        delete_one_resource(SHADER, object);
    }
}

/*-----------------------------------------------------------------------*/

void opengl_delete_texture(GLuint object)
{
    if (object == opengl_current_texture_id) {
        glBindTexture(GL_TEXTURE_2D, 0);
        opengl_current_texture_id = 0;
    }
    if (opengl_has_features(OPENGL_FEATURE_DELAYED_DELETE)) {
        delayed_delete_one_resource(TEXTURE, object);
    } else {
        delete_one_resource(TEXTURE, object);
    }
}

/*-----------------------------------------------------------------------*/

void opengl_delete_vertex_array(GLuint object)
{
    if (opengl_has_features(OPENGL_FEATURE_DELAYED_DELETE)) {
        delayed_delete_one_resource(VERTEX_ARRAY, object);
    } else {
        delete_one_resource(VERTEX_ARRAY, object);
    }
}

/*************************************************************************/
/**************************** Local routines *****************************/
/*************************************************************************/

static void delayed_delete_one_resource(DeleteType type, GLuint object)
{
    if (delete_info_len >= delete_info_size) {
        if (delete_info_fixed_size) {
            ASSERT(delete_info_size > 0,
                   delete_one_resource(type, object); return);
            DLOG("Delete array full for %d/%u, flushing objects",
                 type, object);
            opengl_free_dead_resources(0);
        } else {
            int new_size = delete_info_len + OPENGL_DELETE_INFO_EXPAND;
            DeleteInfo *new_array = mem_realloc(
                delete_info, new_size * sizeof(*delete_info), MEM_ALLOC_TOP);
            if (new_array) {
                delete_info = new_array;
                delete_info_size = new_size;
            } else if (delete_info_size > 0) {
                DLOG("Failed to expand array for %d/%u, flushing objects",
                     type, object);
                opengl_free_dead_resources(0);
                ASSERT(delete_info_len < delete_info_size,
                       delete_one_resource(type, object); return);
            } else {
                DLOG("Failed to allocate array for %d/%u, deleting"
                     " immediately", type, object);
                delete_one_resource(type, object);
                return;
            }
        }
    }

    delete_info[delete_info_len].type = type;
    delete_info[delete_info_len].object = object;
    delete_info_len++;
}

/*-----------------------------------------------------------------------*/

static void delete_one_resource(DeleteType type, GLuint object)
{
    switch (type) {
        case BUFFER:       glDeleteBuffers(1, &object);       break;
        case FRAMEBUFFER:  glDeleteFramebuffers(1, &object);  break;
        case PROGRAM:      glDeleteProgram(object);           break;
        case RENDERBUFFER: glDeleteRenderbuffers(1, &object); break;
        case SHADER:       glDeleteShader(object);            break;
        case TEXTURE:      glDeleteTextures(1, &object);      break;
        case VERTEX_ARRAY: glDeleteVertexArrays(1, &object);  break;
    }
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
