/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/sysdep/sdl/dyngl.h: OpenGL dynamic loading support header.
 */

#ifndef SIL_SRC_SYSDEP_SDL_DYNGL_H
#define SIL_SRC_SYSDEP_SDL_DYNGL_H

/*************************************************************************/
/*************************************************************************/

/**
 * dyngl_init:  Look up the addresses of all OpenGL symbols.
 *
 * This function will fail if the OpenGL library does not export symbols
 * mandated by the OpenGL version which the library reports (or if the
 * library does not return a valid version string).
 *
 * [Return value]
 *     True on success, false on failure.
 */
extern int dyngl_init(void);

/**
 * dyngl_cleanup:  Restore internal state to uninitialized after the OpenGL
 * library has been closed (via SDL_QuitSubSystem()).
 */
extern void dyngl_cleanup(void);

/**
 * dyngl_has_*:  Return whether all functions associated with the specified
 * OpenGL feature are available:
 *    - dyngl_has_compressed_textures() ==> ARB_texture_compression
 *    - dyngl_has_blend_full()          ==> OpenGL v1.4 blend features
 *    - dyngl_has_vertex_buffers()      ==> ARB_vertex_buffer_object
 *    - dyngl_has_shaders()             ==> ARB_shader_objects + related ext'ns
 *    - dyngl_has_framebuffers()        ==> EXT_framebuffer_object
 *    - dyngl_has_opengl3()             ==> OpenGL v3 functions
 *
 * [Return value]
 *     True if all associated functions are available, false otherwise.
 */
extern int dyngl_has_compressed_textures(void);
extern int dyngl_has_blend_full(void);
extern int dyngl_has_vertex_buffers(void);
extern int dyngl_has_shaders(void);
extern int dyngl_has_framebuffers(void);
extern int dyngl_has_opengl3(void);

/*-----------------------------------------------------------------------*/

#ifdef INCLUDE_TESTS

/**
 * TEST_dyngl_override_glGetString:  Override the dynamically-loaded
 * glGetString() function with the given function.
 *
 * This function must be called before dyngl_init().
 *
 * [Parameters]
 *     function: Function to call in place of glGetString().  This function
 *         receives a pointer to the original glGetString() function in
 *         addition to the usual "name" parameter.
 */
extern void TEST_dyngl_override_glGetString(
    const GLubyte * GLAPIENTRY (*function)(
        GLenum name,
        const GLubyte * GLAPIENTRY (*original_glGetString)(GLenum)));

#endif  // INCLUDE_TESTS

/*************************************************************************/
/*************************************************************************/

#endif  // SIL_SRC_SYSDEP_SDL_DYNGL_H

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
