/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/math/matrix.c: Tests for matrix functions.
 */

#ifdef INCLUDE_TESTS  // To the end of the file.

#include "src/base.h"
#include "src/math.h"
#include "src/test/base.h"
#include "src/test/math/internal.h"

/*************************************************************************/
/*************************************************************************/

DEFINE_MATH_TEST_RUNNER(test_math_matrix)

/*-----------------------------------------------------------------------*/

/* Continue through failures.  (This must come after the test runner is
 * defined.) */
#undef FAIL_ACTION
#define FAIL_ACTION  failed = 1

/*-----------------------------------------------------------------------*/

TEST(test_identity)
{
    int failed = 0;

    CHECK_FLOATEQUAL(mat4_identity._11, 1);
    CHECK_FLOATEQUAL(mat4_identity._12, 0);
    CHECK_FLOATEQUAL(mat4_identity._13, 0);
    CHECK_FLOATEQUAL(mat4_identity._14, 0);
    CHECK_FLOATEQUAL(mat4_identity._21, 0);
    CHECK_FLOATEQUAL(mat4_identity._22, 1);
    CHECK_FLOATEQUAL(mat4_identity._23, 0);
    CHECK_FLOATEQUAL(mat4_identity._24, 0);
    CHECK_FLOATEQUAL(mat4_identity._31, 0);
    CHECK_FLOATEQUAL(mat4_identity._32, 0);
    CHECK_FLOATEQUAL(mat4_identity._33, 1);
    CHECK_FLOATEQUAL(mat4_identity._34, 0);
    CHECK_FLOATEQUAL(mat4_identity._41, 0);
    CHECK_FLOATEQUAL(mat4_identity._42, 0);
    CHECK_FLOATEQUAL(mat4_identity._43, 0);
    CHECK_FLOATEQUAL(mat4_identity._44, 1);

    return !failed;
}

/*-----------------------------------------------------------------------*/

TEST(test_basic)
{
    int failed = 0;

    for (int i = 0; i < lenof_matrix_tests(); i++) {
        if (matrix_tests[i].type == PLUS || matrix_tests[i].type == NEG) {
            continue;  // Skip C++-specific tests.
        }

        if (matrix_tests[i].size == 4) {
            Matrix4f res;

            /* Handle noninvertible matrices specially. */
            if (matrix_tests[i].type == INV && matrix_tests[i].b._11 != 0) {
                if (mat4_inv(&res, &matrix_tests[i].a)) {
                    FAIL("test %u: inverted a non-invertible matrix", i);
                }
                continue;  // Skip regular processing.
            }

            /* Check operation of the function itself. */
            mem_clear(&res, sizeof(res));
            switch (matrix_tests[i].type) {
                case ADD:   mat4_add(&res,
                                     &matrix_tests[i].a, &matrix_tests[i].b);
                            break;
                case SUB:   mat4_sub(&res,
                                     &matrix_tests[i].a, &matrix_tests[i].b);
                            break;
                case MUL:   mat4_mul(&res,
                                     &matrix_tests[i].a, &matrix_tests[i].b);
                            break;
                case TRANS: mat4_transpose(&res, &matrix_tests[i].a); break;
                case INV:   mat4_inv(&res, &matrix_tests[i].a); break;
                default:    FAIL("test %u: invalid type", i); break;
            }
            if (!check_matrix_error(&res, &matrix_tests[i].res,
                                    "test %u: result != expect", i)) {
                failed = 1;
                continue;  // Test failed already, so skip the dest==src tests.
            }

            /* Check operation with dest == src1. */
            res = matrix_tests[i].a;
            switch (matrix_tests[i].type) {
                case ADD:   mat4_add(&res, &res, &matrix_tests[i].b); break;
                case SUB:   mat4_sub(&res, &res, &matrix_tests[i].b); break;
                case MUL:   mat4_mul(&res, &res, &matrix_tests[i].b); break;
                case TRANS: mat4_transpose(&res, &res); break;
                case INV:   mat4_inv(&res, &res); break;
                default:    FAIL("test %u: invalid type", i); break;
            }
            if (!check_matrix_error(&res, &matrix_tests[i].res,
                                    "test %u: fail on dest == src1", i)) {
                failed = 1;
            }

            /* Check operation with dest == src2 (if appropriate). */
            res = matrix_tests[i].b;
            switch (matrix_tests[i].type) {
                case ADD:   mat4_add(&res, &matrix_tests[i].a, &res); break;
                case SUB:   mat4_sub(&res, &matrix_tests[i].a, &res); break;
                case MUL:   mat4_mul(&res, &matrix_tests[i].a, &res); break;
                case TRANS: goto skip_src2;  // Only one operand.
                case INV:   goto skip_src2;  // Only one operand.
                default:    FAIL("test %u: invalid type", i); break;
            }
            if (!check_matrix_error(&res, &matrix_tests[i].res,
                                    "test %u: fail on dest == src2", i)) {
                failed = 1;
            }
          skip_src2:;

        } else {
            FAIL("test %u: bad matrix size %d", i, matrix_tests[i].size);
        }
    }

    return !failed;
}

/*-----------------------------------------------------------------------*/

TEST(test_transform)
{
    int failed = 0;

    for (int i = 0; i < lenof_transform_tests(); i++) {
        Matrix4f m;
        m._11 = transform_tests[i].m1[0];
        m._12 = transform_tests[i].m1[1];
        m._13 = transform_tests[i].m1[2];
        m._14 = transform_tests[i].m1[3];
        m._21 = transform_tests[i].m2[0];
        m._22 = transform_tests[i].m2[1];
        m._23 = transform_tests[i].m2[2];
        m._24 = transform_tests[i].m2[3];
        m._31 = transform_tests[i].m3[0];
        m._32 = transform_tests[i].m3[1];
        m._33 = transform_tests[i].m3[2];
        m._34 = transform_tests[i].m3[3];
        m._41 = transform_tests[i].m4[0];
        m._42 = transform_tests[i].m4[1];
        m._43 = transform_tests[i].m4[2];
        m._44 = transform_tests[i].m4[3];

        if (transform_tests[i].size == 2) {
            Vector2f vec, res;
            vec.x = transform_tests[i].coord[0];
            vec.y = transform_tests[i].coord[1];
            mat4_vec2_transform(&res, &vec, &m);
            if (res.x != transform_tests[i].res[0]
             || res.y != transform_tests[i].res[1]) {
                FAIL("transform test %u: result <%f,%f> != expect <%f,%f>",
                     i, res.x, res.y,
                     transform_tests[i].res[0], transform_tests[i].res[1]);
            }
            res = vec;
            mat4_vec2_transform(&res, &res, &m);
            if (res.x != transform_tests[i].res[0]
             || res.y != transform_tests[i].res[1]) {
                FAIL("transform test %u: result <%f,%f> != expect <%f,%f>"
                     " for dest == src", i, res.x, res.y,
                     transform_tests[i].res[0], transform_tests[i].res[1]);
            }

        } else if (transform_tests[i].size == 3) {
            Vector3f vec, res;
            vec.x = transform_tests[i].coord[0];
            vec.y = transform_tests[i].coord[1];
            vec.z = transform_tests[i].coord[2];
            mat4_vec3_transform(&res, &vec, &m);
            if (res.x != transform_tests[i].res[0]
             || res.y != transform_tests[i].res[1]
             || res.z != transform_tests[i].res[2]) {
                FAIL("transform test %u: result <%f,%f,%f> != expect"
                     " <%f,%f,%f>", i, res.x, res.y, res.z,
                     transform_tests[i].res[0], transform_tests[i].res[1],
                     transform_tests[i].res[2]);
            }
            res = vec;
            mat4_vec3_transform(&res, &res, &m);
            if (res.x != transform_tests[i].res[0]
             || res.y != transform_tests[i].res[1]
             || res.z != transform_tests[i].res[2]) {
                FAIL("transform test %u: result <%f,%f,%f> != expect"
                     " <%f,%f,%f> for dest == src", i, res.x, res.y, res.z,
                     transform_tests[i].res[0], transform_tests[i].res[1],
                     transform_tests[i].res[2]);
            }

        } else if (transform_tests[i].size == 4) {
            Vector4f vec, res;
            vec.x = transform_tests[i].coord[0];
            vec.y = transform_tests[i].coord[1];
            vec.z = transform_tests[i].coord[2];
            vec.w = transform_tests[i].coord[3];
            mat4_vec4_transform(&res, &vec, &m);
            if (res.x != transform_tests[i].res[0]
             || res.y != transform_tests[i].res[1]
             || res.z != transform_tests[i].res[2]
             || res.w != transform_tests[i].res[3]) {
                FAIL("transform test %u: result <%f,%f,%f,%f> != expect"
                     " <%f,%f,%f,%f>", i, res.x, res.y, res.z, res.w,
                     transform_tests[i].res[0], transform_tests[i].res[1],
                     transform_tests[i].res[2], transform_tests[i].res[3]);
            }
            res = vec;
            mat4_vec4_transform(&res, &res, &m);
            if (res.x != transform_tests[i].res[0]
             || res.y != transform_tests[i].res[1]
             || res.z != transform_tests[i].res[2]
             || res.w != transform_tests[i].res[3]) {
                FAIL("transform test %u: result <%f,%f,%f,%f> != expect"
                     " <%f,%f,%f,%f> for dest == src",
                     i, res.x, res.y, res.z, res.w,
                     transform_tests[i].res[0], transform_tests[i].res[1],
                     transform_tests[i].res[2], transform_tests[i].res[3]);
            }

        } else {
            FAIL("test %u: bad vector size %d", i, transform_tests[i].size);
        }
    }

    return !failed;
}

/*************************************************************************/
/*************************************************************************/

#endif  // INCLUDE_TESTS

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
