/*
 * System Interface Library for games
 * Copyright (c) 2011 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/sysdep/psp/texture.c: PSP-specific texture tests.
 */

#ifdef INCLUDE_TESTS  // To the end of the file.

#include "src/base.h"
#include "src/graphics.h"
#include "src/math.h"
#include "src/memory.h"
#include "src/sysdep.h"
#include "src/test/base.h"
#include "src/test/graphics/internal.h"
#include "src/texture.h"

/*************************************************************************/
/***************** Test runner and init/cleanup routines *****************/
/*************************************************************************/

static int do_test_psp_texture(void);
int test_psp_texture(void)
{
    CHECK_TRUE(graphics_init());
    const int result = do_test_psp_texture();
    graphics_cleanup();
    return result;
}

DEFINE_GENERIC_TEST_RUNNER(do_test_psp_texture)

/*-----------------------------------------------------------------------*/

TEST_INIT(init)
{
    graphics_start_frame();
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST_CLEANUP(cleanup)
{
    graphics_finish_frame();
    graphics_flush_resources();
    return 1;
}

/*************************************************************************/
/***************************** Test routines *****************************/
/*************************************************************************/

TEST(test_delayed_delete)
{
    const int64_t bytes_used = mem_debug_bytes_allocated();

    int texture;
    CHECK_TRUE(texture = texture_create(4, 2, 0, 0));

    texture_destroy(texture);
    /* The memory should still be allocated so that the GE could draw from
     * it if the texture had been used in a drawing operation. */
    CHECK_TRUE(mem_debug_bytes_allocated() >= bytes_used + 4*2*4);

    /* A subsequent flush operation should free the texture. */
    graphics_flush_resources();
    CHECK_INTEQUAL(mem_debug_bytes_allocated(), bytes_used);

    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_tall_texture)
{
    int texture;
    CHECK_TRUE(texture = texture_create(1, 1024, 0, 0));
    uint8_t *pixels;
    CHECK_TRUE(pixels = texture_lock(texture));
    for (int y = 0; y < 512; y++) {
        pixels[y*4+0] = 0;
        pixels[y*4+1] = lbound(y-256, 0);
        pixels[y*4+2] = ubound(y, 255);
        pixels[y*4+3] = 255;
        pixels[(512+y)*4+0] = 255;
        pixels[(512+y)*4+1] = lbound(y-256, 0);
        pixels[(512+y)*4+2] = ubound(y, 255);
        pixels[(512+y)*4+3] = 255;
    }
    texture_unlock(texture);
    texture_set_repeat(texture, 0, 0);
    texture_set_antialias(texture, 0);
    texture_apply(texture);

    uint8_t readbuf[4*4*4];
    graphics_set_viewport(0, 0, 4, 4);

    graphics_clear(0, 1, 0);
    graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS);
    graphics_add_vertex(
        &(Vector3f){-1,-1,0}, &(Vector2f){0, 508/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){+1,-1,0}, &(Vector2f){1, 508/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){+1,+1,0}, &(Vector2f){1, 512/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){-1,+1,0}, &(Vector2f){0, 512/1024.0f}, NULL);
    CHECK_TRUE(graphics_end_and_draw_primitive());
    CHECK_TRUE(graphics_read_pixels(0, 0, 4, 4, readbuf));
    for (int i = 0; i < lenof(readbuf); i += 4) {
        const int x = (i/4)%4, y = (i/4)/4;
        CHECK_PIXEL(&readbuf[i], 0,252+y,255,255, x,y);
    }

    graphics_clear(0, 1, 0);
    graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS);
    graphics_add_vertex(
        &(Vector3f){-1,-1,0}, &(Vector2f){0, 512/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){+1,-1,0}, &(Vector2f){1, 512/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){+1,+1,0}, &(Vector2f){1, 516/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){-1,+1,0}, &(Vector2f){0, 516/1024.0f}, NULL);
    CHECK_TRUE(graphics_end_and_draw_primitive());
    CHECK_TRUE(graphics_read_pixels(0, 0, 4, 4, readbuf));
    for (int i = 0; i < lenof(readbuf); i += 4) {
        const int x = (i/4)%4, y = (i/4)/4;
        CHECK_PIXEL(&readbuf[i], 255,0,y,255, x,y);
    }

    graphics_clear(0, 1, 0);
    graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS);
    graphics_add_vertex(
        &(Vector3f){-1,-1,0}, &(Vector2f){0, 508/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){ 0,-1,0}, &(Vector2f){1, 508/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){ 0,+1,0}, &(Vector2f){1, 512/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){-1,+1,0}, &(Vector2f){0, 512/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){ 0,-1,0}, &(Vector2f){0, 512/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){+1,-1,0}, &(Vector2f){1, 512/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){+1,+1,0}, &(Vector2f){1, 516/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){ 0,+1,0}, &(Vector2f){0, 516/1024.0f}, NULL);
    CHECK_TRUE(graphics_end_and_draw_primitive());
    CHECK_TRUE(graphics_read_pixels(0, 0, 4, 4, readbuf));
    for (int i = 0; i < lenof(readbuf); i += 4) {
        const int x = (i/4)%4, y = (i/4)/4;
        if (x < 2) {
            CHECK_PIXEL(&readbuf[i], 0,252+y,255,255, x,y);
        } else {
            CHECK_PIXEL(&readbuf[i], 255,0,y,255, x,y);
        }
    }

    texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_tall_texture_indexed)
{
    int texture;
    CHECK_TRUE(texture = texture_create(512, 1024, 0, 0));
    uint8_t *pixels;
    CHECK_TRUE(pixels = texture_lock(texture));
    mem_fill32(pixels,           0xFFFF0000, 512*512*4);
    mem_fill32(pixels+512*512*4, 0xFF00FF00, 512*512*4);
    texture_unlock(texture);
    texture_set_repeat(texture, 0, 0);
    texture_set_antialias(texture, 0);
    texture_apply(texture);

    uint8_t readbuf[4*4*4];
    graphics_clear(0, 0, 0);
    graphics_set_viewport(0, 0, 4, 4);
    static const struct {Vector2f pos; Vector2f tex;} vertices[] = {
        {{-1,-1}, {0/512.0f, 508/1024.0f}},
        {{-1,-1}, {0/512.0f, 512/1024.0f}},
        {{+1,-1}, {4/512.0f, 508/1024.0f}},
        {{+1,-1}, {4/512.0f, 512/1024.0f}},
        {{+1,+1}, {4/512.0f, 516/1024.0f}},
        {{+1,+1}, {4/512.0f, 512/1024.0f}},
        {{-1,+1}, {0/512.0f, 516/1024.0f}},
        {{-1,+1}, {0/512.0f, 512/1024.0f}},
    };
    static const uint32_t format[] = {
        GRAPHICS_VERTEX_FORMAT(POSITION_2F, 0),
        GRAPHICS_VERTEX_FORMAT(TEXCOORD_2F, sizeof(Vector2f)),
        0
    };

    static const uint8_t indices8[] = {1,3,4,6};
    graphics_draw_indexed_vertices(
        GRAPHICS_PRIMITIVE_QUADS,
        vertices, format, sizeof(*vertices), lenof(vertices),
        indices8, sizeof(*indices8), lenof(indices8));
    CHECK_TRUE(graphics_read_pixels(0, 0, 4, 4, readbuf));
    for (int i = 0; i < lenof(readbuf); i += 4) {
        const int x = (i/4)%4, y = (i/4)/4;
        CHECK_PIXEL(&readbuf[i], 0,255,0,255, x,y);
    }

    static const uint16_t indices16[] = {0,2,5,7};
    graphics_draw_indexed_vertices(
        GRAPHICS_PRIMITIVE_QUADS,
        vertices, format, sizeof(*vertices), lenof(vertices),
        indices16, sizeof(*indices16), lenof(indices16));
    CHECK_TRUE(graphics_read_pixels(0, 0, 4, 4, readbuf));
    for (int i = 0; i < lenof(readbuf); i += 4) {
        const int x = (i/4)%4, y = (i/4)/4;
        CHECK_PIXEL(&readbuf[i], 0,0,255,255, x,y);
    }

    texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_tall_texture_16bpp)
{
    uint8_t *pixels = mem_alloc(512*1024*2, 0, 0);
    mem_fill32(pixels,           0xF800F800, 512*512*2);
    mem_fill32(pixels+512*512*2, 0x07E007E0, 512*512*2);
    int texture;
    CHECK_TRUE(texture = texture_create_with_data(
                   512, 1024, pixels, TEX_FORMAT_RGB565, 512, 0, 0));
    mem_free(pixels);
    texture_set_repeat(texture, 0, 0);
    texture_set_antialias(texture, 0);
    texture_apply(texture);

    uint8_t readbuf[4*4*4];
    graphics_clear(0, 0, 0);
    graphics_set_viewport(0, 0, 4, 4);

    graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS);
    graphics_add_vertex(
        &(Vector3f){-1,-1,0}, &(Vector2f){0/512.0f, 508/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){ 0,-1,0}, &(Vector2f){2/512.0f, 508/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){ 0,+1,0}, &(Vector2f){2/512.0f, 512/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){-1,+1,0}, &(Vector2f){0/512.0f, 512/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){ 0,-1,0}, &(Vector2f){0/512.0f, 512/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){+1,-1,0}, &(Vector2f){2/512.0f, 512/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){+1,+1,0}, &(Vector2f){2/512.0f, 516/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){ 0,+1,0}, &(Vector2f){0/512.0f, 516/1024.0f}, NULL);
    CHECK_TRUE(graphics_end_and_draw_primitive());
    CHECK_TRUE(graphics_read_pixels(0, 0, 4, 4, readbuf));
    for (int i = 0; i < lenof(readbuf); i += 4) {
        const int x = (i/4)%4, y = (i/4)/4;
        if (x < 2) {
            CHECK_PIXEL(&readbuf[i], 0,0,255,255, x,y);
        } else {
            CHECK_PIXEL(&readbuf[i], 0,255,0,255, x,y);
        }
    }

    texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_tall_texture_8bpp)
{
    uint8_t *pixels = mem_alloc(256*4 + 512*1024, 0, 0);
    memset(pixels, 0, 256*4);
    memcpy(pixels, "\0\0\xFF\xFF\0\xFF\0\xFF", 8);
    memset(pixels + 256*4, 0, 512*512);
    memset(pixels + 256*4 + 512*512, 1, 512*512);
    int texture;
    CHECK_TRUE(texture = texture_create_with_data(
                   512, 1024, pixels, TEX_FORMAT_PALETTE8_RGBA8888, 512, 0, 0));
    mem_free(pixels);
    texture_set_repeat(texture, 0, 0);
    texture_set_antialias(texture, 0);
    texture_apply(texture);

    uint8_t readbuf[4*4*4];
    graphics_clear(0, 0, 0);
    graphics_set_viewport(0, 0, 4, 4);

    graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS);
    graphics_add_vertex(
        &(Vector3f){-1,-1,0}, &(Vector2f){0/512.0f, 508/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){ 0,-1,0}, &(Vector2f){2/512.0f, 508/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){ 0,+1,0}, &(Vector2f){2/512.0f, 512/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){-1,+1,0}, &(Vector2f){0/512.0f, 512/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){ 0,-1,0}, &(Vector2f){0/512.0f, 512/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){+1,-1,0}, &(Vector2f){2/512.0f, 512/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){+1,+1,0}, &(Vector2f){2/512.0f, 516/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){ 0,+1,0}, &(Vector2f){0/512.0f, 516/1024.0f}, NULL);
    CHECK_TRUE(graphics_end_and_draw_primitive());
    CHECK_TRUE(graphics_read_pixels(0, 0, 4, 4, readbuf));
    for (int i = 0; i < lenof(readbuf); i += 4) {
        const int x = (i/4)%4, y = (i/4)/4;
        if (x < 2) {
            CHECK_PIXEL(&readbuf[i], 0,0,255,255, x,y);
        } else {
            CHECK_PIXEL(&readbuf[i], 0,255,0,255, x,y);
        }
    }

    texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_tall_texture_apply_0)
{
    int texture;
    CHECK_TRUE(texture = texture_create(512, 1024, 0, 0));
    uint8_t *pixels;
    CHECK_TRUE(pixels = texture_lock(texture));
    mem_fill32(pixels,           0xFFFF0000, 512*512*4);
    mem_fill32(pixels+512*512*4, 0xFF00FF00, 512*512*4);
    texture_unlock(texture);
    texture_set_repeat(texture, 0, 0);
    texture_set_antialias(texture, 0);
    texture_apply(texture);

    uint8_t readbuf[4*4*4];
    graphics_set_viewport(0, 0, 4, 4);

    graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS);
    graphics_add_vertex(
        &(Vector3f){-1,-1,0}, &(Vector2f){0/512.0f, 508/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){+1,-1,0}, &(Vector2f){4/512.0f, 508/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){+1,+1,0}, &(Vector2f){4/512.0f, 512/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){-1,+1,0}, &(Vector2f){0/512.0f, 512/1024.0f}, NULL);
    int primitive;
    CHECK_TRUE(primitive = graphics_end_primitive());

    /* Draw it once to do the texture coordinate adjustment. */
    graphics_clear(0, 0, 0);
    graphics_draw_primitive(primitive);

    /* Make sure a subsequent draw doesn't crash trying to choose
     * subtextures. */
    texture_apply(0);
    graphics_clear(0, 0, 0);
    graphics_draw_primitive(primitive);
    CHECK_TRUE(graphics_read_pixels(0, 0, 4, 4, readbuf));
    for (int i = 0; i < lenof(readbuf); i += 4) {
        const int x = (i/4)%4, y = (i/4)/4;
        CHECK_PIXEL(&readbuf[i], 255,255,255,255, x,y);
    }

    graphics_destroy_primitive(primitive);
    texture_destroy(texture);
    return 1;
}

/*-----------------------------------------------------------------------*/

TEST(test_tall_texture_apply_short)
{
    int texture;
    CHECK_TRUE(texture = texture_create(512, 1024, 0, 0));
    /* For this test, we won't actually try to read back any data, so we
     * don't need to initialize the texture. */

    int texture2;
    CHECK_TRUE(texture2 = texture_create(512, 1, 0, 0));

    graphics_set_viewport(0, 0, 4, 4);

    graphics_begin_primitive(GRAPHICS_PRIMITIVE_QUADS);
    graphics_add_vertex(
        &(Vector3f){-1,-1,0}, &(Vector2f){0/512.0f, 512/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){+1,-1,0}, &(Vector2f){4/512.0f, 512/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){+1,+1,0}, &(Vector2f){4/512.0f, 516/1024.0f}, NULL);
    graphics_add_vertex(
        &(Vector3f){-1,+1,0}, &(Vector2f){0/512.0f, 516/1024.0f}, NULL);
    int primitive;
    CHECK_TRUE(primitive = graphics_end_primitive());

    /* Draw it once to do the texture coordinate adjustment. */
    graphics_clear(0, 0, 0);
    texture_apply(texture);
    graphics_draw_primitive(primitive);

    /* Make sure a subsequent draw with a short (height <= 512) texture
     * doesn't cause a crash, even though it violates the documented
     * requirements for tall texture rendering. */
    graphics_clear(0, 0, 0);
    texture_apply(texture2);
    graphics_draw_primitive(primitive);

    graphics_destroy_primitive(primitive);
    texture_destroy(texture2);
    texture_destroy(texture);
    return 1;
}

/*************************************************************************/
/*************************************************************************/

#endif  // INCLUDE_TESTS

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
