/*
 * System Interface Library for games
 * Copyright (c) 2007-2017 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/utility/id-array.h: Header for ID array management routines.
 */

/*
 * These routines are used to implement the internal arrays used to assign
 * numeric IDs to objects such as fonts and textures, but client code is
 * free to make use of them as well.
 */

#ifndef SIL_SRC_UTILITY_ID_ARRAY_H
#define SIL_SRC_UTILITY_ID_ARRAY_H

EXTERN_C_BEGIN

/*************************************************************************/
/*************************************************************************/

/**
 * IDArray:  Data structure for an ID array.  Callers should treat these
 * fields as private!
 */
typedef struct IDArray IDArray;
struct IDArray {
    void **array;   // Array of pointers.
    int first_free; // Index of first unused element.
    int num_used;   // Number of elements currently in use.
    int size;       // Current size of array.
    int expand_by;  // Number of elements by which to expand array when needed.
};

/**
 * ID_ARRAY_INITIALIZER:  Macro to initialize a static IDArray.  Use this
 * macro instead of writing out the initializer manually; for example:
 *
 * static IDArray my_array = ID_ARRAY_INITIALIZER(100);
 *
 * [Parameters]
 *     expand_by: Number of slots to add when registering an object in a
 *         full array.
 */
#define ID_ARRAY_INITIALIZER(expand_by_) \
    {.array = NULL, .first_free = 1, .size = 0, .expand_by = (expand_by_)}


/**
 * ID_ARRAY_VALIDATE:  Validate an ID against the given ID array.  If the
 * ID is valid, assign the corresponding value to a variable; if the ID is
 * invalid, execute the specified failure actions.
 *
 * The variadic parameter is one or more statements which will be executed
 * on failure, typically ending with a "return" or "break" statement to
 * abort the enclosing function or block.  No semicolon is required at the
 * end, though if multiple statements are used, each non-final statement
 * should be followed by a semicolon as usual.  If referencing the value
 * of the "id" parameter, use "__id", which is an internal variable used
 * to avoid side effects from multiple references to the macro parameter.
 *
 * [Parameters]
 *     array: IDArray against which to validate.
 *     id: ID to validate.
 *     type: Type of data contained in array.
 *     var: Variable (or other lvalue) to modify on auccess.
 *     ...: Statement(s) to execute on failure.
 */
#define ID_ARRAY_VALIDATE(array,id,type,var,...)  do { \
    const int __id = (id);                      \
    (var) = (type)id_array_get((array), __id);  \
    if (UNLIKELY(!(var))) {                     \
        __VA_ARGS__;                            \
    }                                           \
} while (0)

/*-----------------------------------------------------------------------*/

/**
 * id_array_register:  Register the given object in the ID array and return
 * the ID allocated for it.
 *
 * [Parameters]
 *     array: IDArray in which to register the object.
 *     object: Object to register.
 * [Return value]
 *     Allocated ID (nonzero), or zero on error.
 */
extern int id_array_register(IDArray *array, void *object
#ifdef DEBUG
                             , const char *file, int line
#endif
);

/**
 * id_array_get:  Return the object corresponding to the given ID.
 *
 * [Parameters]
 *     array: IDArray from which to retrieve object.
 *     id: ID of object to retrieve.
 * [Return value]
 *     Object pointer associated with the given ID, or NULL if the ID is
 *     unused.
 */
static inline void *id_array_get(const IDArray *array, int id)
{
    PRECOND(array != NULL, return NULL);
    if (UNLIKELY(id <= 0) || UNLIKELY(id >= array->size)) {
        return NULL;
    }
    return array->array[id];
}

/**
 * id_array_set:  Assign a new object to the given ID.  Behavior is
 * undefined if the given ID is not an ID which was previously returned by
 * id_array_register() and has not yet been released with id_array_release().
 *
 * [Parameters]
 *     array: IDArray in which to assign object.
 *     id: ID to which to assign object.
 *     object: Object to assign.
 */
extern void id_array_set(IDArray *array, int id, void *object);

/**
 * id_array_release:  Release the given ID for reuse.  Does nothing if id
 * is zero.
 *
 * [Parameters]
 *     array: IDArray on which to operate.
 *     id: ID to release.
 */
extern void id_array_release(IDArray *array, int id);

/**
 * id_array_size:  Return the number of slots allocated for the array.
 * This value is always greater than the largest valid ID at that time.
 *
 * [Parameters]
 *     array: IDArray on which to operate.
 * [Return value]
 *     Number of slots allocated.
 */
static inline int id_array_size(const IDArray *array)
{
    PRECOND(array != NULL, return 0);
    return array->size;
}

/*-----------------------------------------------------------------------*/

/*
 * When debugging is enabled, we wrap allocating calls with these macros
 * which pass down the source file and line at which the call was made,
 * as in ../resource.h.
 */

#ifdef DEBUG
# define id_array_register(array,expand_by) \
    id_array_register((array), (expand_by), __FILE__, __LINE__)
#endif

/*************************************************************************/
/*************************************************************************/

EXTERN_C_END

#endif  // SIL_SRC_UTILITY_ID_ARRAY_H

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
